
; Simple 9600 N81 software uart specially for ericsson chatboard
; rx: IC4 detects startbit, OC5 times databits
; tx: OC4 times bits and toggles pin
;
; rx speedbuffer does dirty tricks with alignment and self-modification

	BIT = 109   ; 16.777MHz / 16 / 9600  ; really about 9620 baud

	include regs.inc

; IC1     PGP0 SSEL
; IC2     PGP1 RTSB
; IC3     PGP2 CTSA
; IC4/OC5 PGP7 DCDB RXD3  cpu/130  U2/19 -->o-- /24 --- D25/12 DCD2
; OC1     PGP3 CTSB
; OC2     PGP4 DSRA
; OC3     PGP5 DSRB
; OC4     PGP6 DCDA TXD3  cpu/131  U1/21 -->o-- /20 --- D25/8  DCD1

TXD3      equ  40h  ; PGP6 and PGP7
RXD3      equ  80h  ; there is a rolb down there, too

OC4_V     equ  8Eh
IC4_OC5_V equ  90h

I4O5F     equ  80h  ; TFLG1 and TMSK1
OC4F      equ  40h

OC5mode   equ 0C0h  ; TCTL1/2
OC4mode   equ  30h
OC4toggle equ  10h
IC4edges  equ 0C0h
IC4fedge  equ  80h

I4O5      equ  04h  ; PACTL selects which it is, 0=OC5 1=IC4

	;------------------------------------------------------------
	section .bankf
	align 1

tx_ptr     ds.w 1
bittime    ds.w 1
bittime1_5 ds.w 1
rx_shifter ds.b 2      ; only 1 byte really

	;------------------------------------------------------------
	section .rodata
	align 1

	; 0117  79 ( 79) 0x4f  O 1'0100'1111'0
	; 0113  75 ( 75) 0x4b  K 1'0100'1011'0
	; 0015  13 ( 13) 0x0d ^M 1'0000'1101'0
	; 0012  10 ( 10) 0x0a ^J 1'0000'1010'0

ack_pattern
	dc.w BIT,BIT+BIT+BIT+BIT,BIT+BIT,BIT,BIT,BIT + 20 ; stretched stopbits
	dc.w BIT,BIT+BIT,BIT,BIT,BIT+BIT,BIT,BIT,BIT + 20
	dc.w BIT,BIT,BIT,BIT+BIT,BIT+BIT+BIT+BIT,BIT + 20
	dc.w BIT+BIT,BIT,BIT,BIT,BIT+BIT+BIT+BIT,       0 ; end 0 is "forever"

	;------------------------------------------------------------
	; trickstery here, _cnt is "live text" inside a staa EXT

	global _chatboard_buf, _chatboard_cnt

	section .bsspgs         ; only for 'align 8' variables

_chatboard_buf ds.b 256     ; align 8 and size 256 a must

	;------------------------------------------------------------
	section .inthand
	align 1


	;------------------------------------------------------------
	; low-going edge from startbit on IC4 aka OC5 aka PGP7 aka RXD3,
	; IC4 active, or time to sample another _data_ bit, OC5 active

IC4_OC5_int_startbit

	pshm d,k
	ldab # 0FFh
	tbek

	; TI4O5 can be written to only when OC5 is selected
	; are there two separate registers ? try to play safe.

	ldd  bittime1_5        ; 1.5 bits onward lands at the middle of databit 0
	addd TI4O5             ; from the falling edge of startbit
	bclr PACTL, # I4O5     ; flip to OC5 function, bit clear
	std  TI4O5             ; to middle of data bit 0 (not checking noise)
	bclr TFLG1, # I4O5F

	ldd  # 8000h           ; A=marker B=segment
	staa rx_shifter        ; prime with fresh marker, segment F XXX
	tbek                   ; switch EXT

	ldd  # IC4_OC5_int_databits
	std  IC4_OC5_V

	pulm d,k
	rti

IC4_OC5_int_databits

	pshm d,k
	ldab # 0FFh
	tbek

	ldab PORTGP            ; sample rx bit XXX can be delayed too much
	ldaa rx_shifter        ; previous bits, segment F variable
	rolb                   ; put new bit PGP7 = RXD3 to carry
	rora                   ; another bit in byte
	bcs  rx_did_byte       ; if have all of them now
	staa rx_shifter        ; re-shelve

	ldd  bittime           ; center of next bit
	addd TI4O5
	std  TI4O5
	bclr TFLG1, # I4O5F

	pulm d,k
	rti

rx_did_byte

	; last databit has not yet ended, but the next low-going edge
	; can happen only after the stop bit. XXX again, noise not handled XXX

	bset PACTL, # I4O5     ; flip to IC4 function, bit set
	bclr TFLG1, # I4O5F

	clrb                   ; RAM
	tbek                   ; switch EK

_chatboard_cnt equ * + 3   ; LSByte in the following 'staa ext'

	staa _chatboard_buf    ; self-mofifying silly trickstery
	inc  _chatboard_cnt    ; modify the address, wrap after 255 bytes

	ldd  # IC4_OC5_int_startbit
	std  IC4_OC5_V

	pulm d,k
	rti


	;------------------------------------------------------------
	; TXD3 has just complemented automatically, just reload TOC4
	; for the next run of bits of the same value

OC4_int

	pshm d,x,k
	ldab # 0FFh
	tbek
	ldd  # __text_page
	tbxk

	ldx  tx_ptr            ; lives in F segment
	ldd  0,x               ; lives in text segment
	beq  tx_finished       ; array ends with zero
	aix  #+2
	stx  tx_ptr

	addd TOC4
	std  TOC4
	bclr TFLG1, # OC4F

	pulm d,x,k
	rti

tx_finished

	bclr TCTL1, # OC4mode  ; 00 = timer disconnected from output logic
	bclr TMSK1, # OC4F     ; mask OC4 interrupt

	pulm d,x,k
	rti

	;------------------------------------------------------------
	section .text
	align 1

	global _chatboard_OK

_chatboard_OK

	pshm k,ccr
	orp  # 000E0h

	ldab # 0FFh
	tbek                   ; switch EXT

	ldd  # ack_pattern     ; rewind the
	std  tx_ptr            ; pointer in F segment XXX

	ldd  bittime           ; from
	addd TCNT              ;  now
	std  TOC4              ;  to end of (quite useless) quiet period
	bclr TFLG1, # OC4F     ; forget stale compare
	bset TCTL1, # OC4toggle ; 01 = toggle output line when compare
	bset TMSK1, # OC4F     ; enable OC4 interrupt

	pulm k,ccr             ; restore IPL
	rts

	global _chatboard_init

_chatboard_init

	; formula would be 16.777MHz / 16 / baud

	ted                    ; ericsson, aplicom, 4800, 1200

	lde  # 109             ; 9600 baud about
	decb
	beq  L9

	lde  # 87              ; 12000 wierd baudrate
	decb
	beq  L9

	lde  # 218             ; 4800 baud about
	decb
	beq  L9

	lde  # 874             ; 1200 baud about
L9

	pshm k,ccr
	orp  # 000E0h

	ldd  # IC4_OC5_int_startbit
	std  IC4_OC5_V

	ldd  # OC4_int
	std  OC4_V

	ldab # 0FFh
	tbek                   ; switch EXT

	ste  bittime           ; one bit
	lsre                   ; half
	adde bittime           ; one and a half
	ste  bittime1_5        ; here.

	bclr DDRGP, # RXD3     ; input
	bset DDRGP, # TXD3     ; output

	bset PORTGP, # TXD3    ; idle TXD3

	; 00 = timer disconnected from output logic

	bclr TCTL1, # OC4mode + OC5mode

	bset PACTL, # I4O5     ; select IC4, bit set

	bclr TCTL2, # IC4edges ; 00 = capture disabled
	bset TCTL2, # IC4fedge ; 10 = low-going edge interrupts

	bclr TFLG1, # I4O5F    ; forget stale capture
	bset TMSK1, # I4O5F    ; enable IC4 interrupt

	pulm k,ccr             ; restore IPL
	rts

