
#include "standard.h"
#include "hamdr_cfg.h"
#include "fix.h"

extern int debug;

STATIC word byte_cksum;

static void
putbyte(word x)
{
	byte_cksum += x;

	SCI_putc(x);
}

static void
putstring(byte *s)
{
	do {
		putbyte(*s);
	} while (*++s);
}

/* SiRF -------------------------------------------------------------- */

/* motorola order suits us */

static void
sirf_word(word x)
{
	putbyte(((byte *) &x)[0]); /* MSByte */
	putbyte(((byte *) &x)[1]);
}

static void
SiRF_initialize()
{
	static byte msg[] = {
		0x81,       /* Message ID - Switch To NMEA */
		0x02,       /* Mode - 2 = NMEA debug neither enabled nor disabled */
		1, 1,       /* GGA interval and checksum flag - 1 Hz */
		 0, 1,      /* GGL interval and checksum flag */
		 0, 1,      /* GSA interval and checksum flag */
		 0, 1,      /* GSV interval and checksum flag */
		1, 1,       /* RMC interval and checksum flag - 1 Hz */
		 0, 1,      /* VTG interval and checksum flag */
		 0, 1,      /* MSS interval and checksum flag */
		 0, 1,      /*  future sentence */
		 0, 1,      /* ZDA interval and checksum flag */
		 0, 1,      /*  future sentence */
		0x12, 0xC0, /* Baud Rate - 4800 baud XXX see SCI_config below */
	};
	int i;

	SCI_config(38400, 0);

	/*
	 *  start sequence
	 *  payload length
	 *  payload
	 *  message checksum, MSbit clear, over payload only
	 *  end sequence
	 */
	sirf_word(0xA0A2);
	sirf_word(sizeof(msg));

	byte_cksum = 0;
	for (i = 0; i < sizeof(msg); ++i)
		putbyte(msg[i]);

	sirf_word(byte_cksum & 0x7FFF);
	sirf_word(0xB0B3);


	SCI_drain();             /* make sure the bytes get out safely */
	SCI_config(4800, 0);     /* XXX see Baud Rate in msg above */
}

/* EarthMate et al Zodiac speakers ----------------------------------- */

#if 0

STATIC word rockwell_to_NMEA[] = {
	0x81FF,                  /* sync */
	1331,                    /* id */
	3,                       /* num_words */
	0,                       /* flags */
	- 0x81FF - 1331 - 3 - 0, /* hdr_cksum */

	0,                       /* seq_num */
	0,                       /* reserved */
	1,                       /* NMEA = 1 */
	- 0 - 0 - 1,             /* data_cksum */
};

/* intel order, flip */

static void
zodiac_word(word x)
{
	putbyte(((byte *) &x)[1]); /* LSByte */
	putbyte(((byte *) &x)[0]);
}

#endif

/* only words and longs in structure, intel order */

STATIC struct {
	long set_time;             /* 6,7 UDI 10 msec ticks 0..4G */
	word seq_num;
	word sat_meas_seq_num;
	word solution_validity_bits; /* bit set = invalid, not all are fatal */
	word solution_type_bits;
	word solution_meas_count;
	word polar_navigation;
	word gps_week;
	long gps_seconds;
	long gps_nanoseconds;
	word utc_day;              /* 19    UI  day     1..31 */
	word utc_month;            /* 20    UI  month   1..12 */
	word utc_year;             /* 21    UI  year    1980..2079 */
	word utc_hours;            /* 22    UI  hours   0..23 */
	word utc_minutes;          /* 23    UI  minutes 0..59 */
	word utc_seconds;          /* 24    UI  seconds 0..59 */
	long utc_nanoseconds;
	long latitude;             /* 27,28 DI  radians +-pi/2  10^-8 */
	long longitude;            /* 29,30 DI  radians +-pi    10^-8 */
	long height;               /* 31,32 DI  meters  +-50000 10^-2 */
	word geoidal_separation;
	long ground_speed;         /* 34,35 UDI m/s     0..1000 10^-2 */
	word true_course;          /* 36    UI  radians 0..2pi  10^-3 */
	word magnetic_variation;
	word climb_rate;
	word map_datum;
	long hposerr;
	long vposerr;
	long timerr;
	word spderr;
	long clock_bias;
	long clock_bias_stddev;
	long clock_drift;
	long clock_drift_stddev;   /* 53,54 */
	word data_cksum;           /* 55    not included in num_words - oddity */
} msg1000;

/*
 *  unit of x is signed radian / 1e8
 *  convert it to signed milliminutes
 */
static void
zodiac_coord(long *px, long *mm)
{
	static long scale = 147650099;      /* 10800000 * 4G / 314159265 */

	long x = ld_long_intel(px);

	*mm = SCALE32_signed(&x, &scale);
}

static void
zodiac_gps_update()
{
	struct wallclock wc;
	struct vector v;

	/* not enough satellites or exceeded maximum E?PE = leave */

	if (ld_word_intel(&msg1000.solution_validity_bits) & 0x1C)
		return;

	wc.year   = ld_word_intel(&msg1000.utc_year) % 100;
	wc.month  = ld_word_intel(&msg1000.utc_month);
	wc.day    = ld_word_intel(&msg1000.utc_day);

	wc.hour   = ld_word_intel(&msg1000.utc_hours);
	wc.minute = ld_word_intel(&msg1000.utc_minutes);
	wc.second = ld_word_intel(&msg1000.utc_seconds);


	/* latitude/longitude -- signed radian / 1e8 */

	zodiac_coord(&msg1000.latitude,  &v.lat);
	zodiac_coord(&msg1000.longitude, &v.lon);


	/*
	 *  ground_speed -- unsigned m/s / 1e2
	 *  0.01 m/s to knots: *= 0.019438445 ~ 9/463
	 *  calc overflows at MACH 2
	 *
	 *  true_course -- unsigned radian / 1e3
	 *  0.001 radians to degrees: *= 0.05729578 ~ 653/11397
	 */
	v.knots  = muldiv(ld_dword_intel(&msg1000.ground_speed), 9, 463);
	v.course = muldiv(ld_word_intel(&msg1000.true_course), 653, 11397);

	/* polar ? force abs(longitude)=90deg and course=south */

	gps_fix(&v, &wc);
}

static bool
zodiac_cksum(void *p, int n)
{
	word *wp = p;
	word sum = 0;

	do {
		sum += ld_word_intel(wp++);
	} while (--n);

	return (sum == 0);
}

/*
 *  sync               0x81FF
 *  id                 1000 (else ignore message)
 *  num_words          55 - 5  - 1 = 49
 *  flags              dont care
 *  hdr_cksum          such that sum(hdr) == 0
 *  [data data_cksum]  num_words + 1 iff any words
 */
static void
zodiac_gps_input(word c)
{
	static word i, id, num_bytes;
	static byte hdr[5 * 2];         /* 5 words in header */

	if (i < sizeof(hdr)) {
		hdr[i] = c;
		if (++i != sizeof(hdr))
			return;

		/* XXX catch "EARTHA" and reply to it */

		if (ld_word_intel(&hdr[0]) == 0x81FF
		 && zodiac_cksum(hdr, sizeof(hdr) / 2)) {

			word wc;

			id = ld_word_intel(&hdr[2]);
			wc = ld_word_intel(&hdr[4]);

			if (wc)
				num_bytes = (wc + 1) * 2;
			else
				i = 0;           /* no associated data */
			return;
		}
		/* resync until header is valid. should not happen often */

		bcopy(hdr + 1, hdr, --i);
		return;
	}

	/* now in data area. messages other than 1000 are discarded */

	if (id == 1000 && i < sizeof(hdr) + sizeof(msg1000))
		((byte *) &msg1000)[i - sizeof(hdr)] = c;

	if (++i < sizeof(hdr) + num_bytes)
		return;
	i = 0;

	if (id == 1000 && zodiac_cksum(&msg1000, sizeof(msg1000) / 2)) {

		gps_is_speaking();
		zodiac_gps_update();
	}
}

/* Aisin-Seiki ------------------------------------------------------- */

/* motorola order. the alignment does not cooperate though */

STATIC struct {
	word control_code;
	byte validity;
	byte latitude[4], longitude[4], height[2];
	byte error_ellipse[2];
	byte heading[2], heading_error;
	byte speed[2], speed_error;
	byte pad1[3];
	byte datetime[3+3];  /* yymmddHHMMSS as 6 bcd bytes */
	byte HDOP, VDOP;
	byte satellite_count, satellite_info[8];
	byte pad2;
	byte checksum, terminator;
} CACA;

/*  0.25 meter per second to knots */

static word
quarter_ms_to_knots(word qms)
{
	return (muldiv(qms, 121, 249));
}

/* course --- 360/1024 ~ 45/128 */

static word
aisin_heading_to_course(word heading)
{
	return (muldiv(heading, 45, 128));
}

/* unit of *px is 1/256 of a second. signed */

static void
aisin_coord(long *px, long *mm)
{
	static long scale = 279620266; /* second/256 to milliminute */

	*mm = SCALE32_signed(px, &scale);
}

static void
aisin_gps_update()
{
	struct wallclock wc;
	struct vector v;

	switch (CACA.validity & 0x0F) {
	default:
		return;                  /* no fix, bail out */

	case 3: /* 2D fix */
	case 4: /* 3D fix */
		;
	}

	/* unpack date and time from BCD bytes to binary - bit nasty */

	bcopy(CACA.datetime, &wc, 6);      /* XXX layout must stay same */
	repack_bcd_to_binary(&wc, 6);

	/* lat/lon --- unit is 1/256 of a second */

	aisin_coord((long *) CACA.latitude,  &v.lat);
	aisin_coord((long *) CACA.longitude, &v.lon);

	/*
	 * speed --- 1/4 m/s to knots ~ 0.48596112 ~ 121 / 249
	 * speed range is 0...333 or 83.3 m/s or 300 km/h or 162 knots.
	 * speed out of range is encoded as FFFF
	 */
	v.knots = *(word *) CACA.speed;
	if (v.knots < 333)
		v.knots = quarter_ms_to_knots(v.knots);
	else
		v.knots = 162; /* needle at max, 300 km/h */

	v.course = aisin_heading_to_course(*(word *) CACA.heading);

	gps_fix(&v, &wc);
}

/*
 * Data is TTL level 9600 baud 8 bits _plus_ even parity
 */
static void
aisin_gps_input(word c)
{
	static byte sum, i, buf[sizeof(CACA)];

	/*
	 *  remove oldest character from checksum.
	 *  replace oldest with new and update checksum.
	 *  bump and wrap index.
	 */
	sum -= buf[i];
	sum += buf[i] = c;
	if (++i == sizeof(CACA))
		i = 0;

	/*
	 *  character added was CR ?
	 *  sum was 0 before CR was added ?
	 *  two oldest characters are now CA and CA ?
	 *  align and save the halves of data.
	 *  tbd align on-the-fly in buf[] and delete CACA.
	 */
	if ((c & 0xFF) == 0x0D
	 && sum == 0x0D
	 && buf[i] == 0xCA
	 && buf[(i < sizeof(CACA) - 1) ? (i + 1) : 0] == 0xCA) {

		bcopy(buf + i, &CACA, sizeof(CACA) - i);
		bcopy(buf, (byte *) &CACA + i, i);

		gps_is_speaking();
		aisin_gps_update();
	}
}

/* SeaTalk ----------------------------------------------------------- */

static struct {
	struct seatalk_coord {
		byte SW;
		byte deg;
		word mm;
	} lat, lon;
	word course, knots;  /* integer */
	struct wallclock wc;
} seatalk;

static void
seatalk_coord(struct seatalk_coord *ord, long *mm)
{
	*mm = EMUL(_DEG_AS_mm, ord->deg) + ord->mm;
	if (ord->SW)
		*mm = lnegate(*mm);
}

static void
seatalk_gps_update()
{
	struct vector v;

	seatalk_coord(&seatalk.lat, &v.lat);
	seatalk_coord(&seatalk.lon, &v.lon);

	v.course = seatalk.course;
	v.knots  = seatalk.knots;

	gps_fix(&v, &seatalk.wc);
}

/*
 * 53 U0 VW      (U & 3) * 90 + (VW & 0x3F) * 2 + (U & 0xC) / 8 = course
 */
static word
seatalk_course(word U0, word VW)
{
	return ((U0 & 0x30) >> 3) * 45 + (VW & 0x3F) * 2; /* XXX huh ? */
}

static void
seatalk_message(byte *msg)
{
	word x;

	if (debug) {
		print_buf(msg, (msg[1] & 15) + 3);
		eol();
	}
	switch (msg[0]) {

	/*
	 * 50 A2 XX YYYY latitude,  XX deg, YYYY / 100 = min YYYY.15 = South
	 * 51 A2 XX YYYY longitude, XX deg, YYYY / 100 = min YYYY.15 = East
	 * 58 Z5 LA XXYY LO QQRR
	 *                 LA and LO deg, XXYY / 1000 and QQRR / 1000 = min
	 *                 Z.0 = South, Z.1 = East
	 */
	case 0x50:
		x = ld_word_intel(&msg[3]);
		seatalk.lat.deg = msg[2];
		seatalk.lat.mm  = (x & 0x7FFF) * 10;
		seatalk.lat.SW  = (x & 0x8000) ? TRUE : FALSE;
		break;

	case 0x51:
		x = ld_word_intel(&msg[3]);
		seatalk.lon.deg = msg[2];
		seatalk.lon.mm  = (x & 0x7FFF) * 10;
		seatalk.lon.SW  = (x & 0x8000) ? FALSE : TRUE;
		break;

	case 0x58:
		seatalk.lat.deg = msg[2];
		seatalk.lat.mm  = ld_word_intel(&msg[3]);
		seatalk.lat.SW  = (msg[1] & 1) ? TRUE : FALSE;

		seatalk.lon.deg = msg[5];
		seatalk.lon.mm  = ld_word_intel(&msg[6]);
		seatalk.lon.SW  = (msg[1] & 2) ? FALSE : TRUE;
		break;

	/*
	 * 52 01 XXXX    XXXX / 10 = knots
	 * 53 U0 VW      (U & 3) * 90 + (VW & 0x3F) * 2 + (U & 0xC) / 8 = course
	 */
	case 0x52:
		seatalk.knots = ld_word_intel(&msg[2]) / 10;
		break;

	case 0x53:
		seatalk.course = seatalk_course(msg[1], msg[2]); /* cc1.core */
		break;

	/*
	 * 54 t1 TT HH   HH hours, (TTt & 0xFC0) / 64 = minutes, TTt & 0x3F = secs
	 * 56 M1 DD YY   YY year, M month, DD day in month
	 */
	case 0x54:
		x = ld_word_intel(&msg[1]) >> 4;
		seatalk.wc.hour   = msg[3];
		seatalk.wc.minute = msg[2] >> 2;
		seatalk.wc.second = x & 0x3F;

		seatalk_gps_update(); /* XXX which msg ends "current solution" ? */
		break;

	case 0x56:
		seatalk.wc.year  = msg[3];
		seatalk.wc.month = msg[1] >> 4;
		seatalk.wc.day   = msg[2];
		break;

	/*
	 * 6E 07 00 00 00 00 00 00 00 00  MOB
	 * 36 00 01                       Cancel MOB condition
	 * A5 57 QQ HH ?? AA GGZZ YY DD   GPS and DGPS fix info
	 */
	case 0xA5:
		gps_is_speaking();
		break;
	}
}

static void
seatalk_gps_input(word c) /* XXX */
{
	static byte buf[3 + 15];
	static word i = sizeof(buf);  /* discard until fresh message */

	if (c > 0xFF)
		i = 0;

	if (i >= sizeof(buf))
		return;

	buf[i++] = c;

	if (i >= 3 && (buf[1] & 15) + 3 == i)
		seatalk_message(buf);
}

/* Oncore ------------------------------------------------------------ */

#if 0

static void
oncore_send(byte *p, word cc)
{
	byte cksum = 0;

	SCI_putc(0x40);
	SCI_putc(0x40);

	do {
		cksum ^= *p;
		SCI_putc(*p);
		++p;
	} while (--cc);

	SCI_putc(cksum);

	SCI_putc(0x0D); /* CR LF */
	SCI_putc(0x0A);
}

static void
oncore_init()
{
	SCI_drain();
	SCI_config(9600, 0);

	oncore_send("Cg\001", 3); /* terminate idle mode */
	oncore_send("Cg\001", 3); /* terminate idle mode */

	oncore_send("Ha\001", 3); /* 12 CH message every second */
	oncore_send("Hb\001", 3); /* 12 CH short position message every second */
	oncore_send("Ea\001", 3); /*  8 CH message every second */
	oncore_send("Ba\001", 3); /*  6 CH message every second */

	oncore_send("Ci\001", 3); /* switch to NMEA */
}
#endif

/*
 * @@Bamdyyhmsffffaaaaoooohhhhmmmmvvhhddtnt imsd*6 sC\r\n 68 bytes
 * @@Eamdyyhmsffffaaaaoooohhhhmmmmvvhhddtnt imsd*8 sC\r\n 76 bytes
 *
 * s at end, the receiver status byte:
 *		Bit 7: Position Propagate mode
 *		Bit 6: Poor Geometry (DOP over 20)
 *		Bit 5: 3D fix
 *		Bit 4: Altitude Hold (2D fix)
 *		Bit 3: Acquiring Satellites/Position Hold
 *		Bit 2: Differential
 *		Bit 1: Insufficient visible satellites (less than 3)
 *		Bit 0: Bad Almanac
 *
 * @@HbmdyyhmsffffaaaaoooohhhhmmmmVVvvhhddnt ssrrvvvvvvC\r\n 54 bytes
 * @@HamdyyhmsffffaaaaoooohhhhmmmmaaaaoooohhhhmmmmVVvvhhddnt imsidd*12
 *       ssrrccooooTTushmvvvvvvC\r\n 154 bytes
 *
 * ss, the receiver status word
 *		Bit 15-13: 111 = 3D Fix
 *				   110 = 2D Fix
 *				   101 = Propagate Mode
 *				   100 = Position Hold
 *				   011 = Acquiring Satellites
 *				   010 = Bad Geometry
 *				   001 = Reserved, 000 = Reserved
 *		Bit 12-11: Reserved
 *		Bit 10: Narrow band tracking mode (timing rx only)
 *		Bit 9: Fast Acquisition Position
 *		Bit 8: Filter Reset To Raw GPS Solution
 *		Bit 7: Cold Start
 *		Bit 6: Differential Fix
 *		Bit 5: Position Lock
 *		Bit 4: Autosurvey Mode
 *		Bit 3: Insufficient Visible Satellites
 *		Bit 2-1: Antenna Sense 00 = OK
 *							   01 = OC
 *							   10 = UC
 *							   11 = NV
 *		Bit 0: Code Location 0 = EXTERNAL, 1 = INTERNAL
 */

struct oncore_HaHbEaBa_head {
	word
		id;             /* message identifier */
	byte
		month,          /* 1 .. 12             Date */
		day;            /* 1 .. 31 */
	word
		year;           /* 1980 .. 2079 */
	byte
		hours,          /* 0 .. 23             Time */
		minutes,        /* 0 .. 59 */
		seconds,        /* 0 .. 60 */
		frac_sec   [4], /* 0 .. 999,999,999 - alignment lost, thus [4] */

		latitude   [4], /* signed milliarcseconds (+-90 deg) */
		longitude  [4], /* (+-180 deg) */

		height_ref [4], /* -100,000 .. 1,800,000 cm */
		height_MSL [4]; /* -100,000 .. 1,800,000 cm */
};

struct oncore_vvhh {
	word
		velocity,  /* 0 .. 51400 cm/s (0 to 514.00 m/s) */
		heading;   /* 0 .. 3599 0.1 deg (0.0 to 359.9 o) */
};

/* milliarcseconds to milliminutes */

static void
oncore_coord(long *mas, long *mm)
{
	*mm = EDIVx_signed(*mas, 60, NULL);
}

static void
oncore_gps_update(struct oncore_HaHbEaBa_head *p, struct oncore_vvhh *vh)
{
	struct wallclock wc;
	struct vector v;

	wc.year   = p->year % 100;
	wc.month  = p->month;
	wc.day    = p->day;

	wc.hour   = p->hours;
	wc.minute = p->minutes;
	wc.second = p->seconds;

	oncore_coord((long *) p->latitude,  &v.lat);
	oncore_coord((long *) p->longitude, &v.lon);

	/* cm/s to knots and 0.1 degrees to integer degrees */

	v.knots  = muldiv(vh->velocity, 9, 463);
	v.course = vh->heading / 10;

	gps_fix(&v, &wc);
}

static void
oncore_gps_message(byte *msg, word len)
{
	byte *vvhh = msg + 29;
	bool valid;

	switch (*(word *) msg) {
	default:
		return;

	/* accept Propagate Mode, 2D or 3D */

	case 0x4261:                       /* Ba */
		valid = (msg[62] & 0xB0) != 0;
		break;

	case 0x4561:                       /* Ea */
		valid = (msg[70] & 0xB0) != 0;
		break;

	case 0x4862:                       /* Hb */
		vvhh = msg + 31;
		valid = (msg[39] & 0xE0) >= 0xA0;
		break;

	case 0x4861:                       /* Ha */
		vvhh = msg + 47;
		valid = (msg[127] & 0xE0) >= 0xA0;
		break;
	}
	if (valid)
		oncore_gps_update((void *) msg, (void *) vvhh);
	else
		gps_is_speaking();
}

static void
oncore_gps_input(word c)
{
	static byte sum, len, buf[256];

	sum ^= c;
	buf[len++] = c;   /* wraps to empty if no sync after 256 bytes */

	if (len < 4)
		return;

	if (len == 4) {
		if (*(word *) buf != 0x4040
		 || buf[2] < 'A' || buf[2] > 'Z'
		 || buf[3] < 'a' || buf[3] > 'z') {
			len = 3;
			bcopy(buf + 1, buf, 3);
			return;
		}
		sum = buf[2] ^ buf[3];
		return;
	}
	if (len > 7 && sum == 0x0D ^ 0x0A && *(word *) &buf[len - 2] == 0x0D0A) {

		oncore_gps_message(buf + 2, len - 5);
		len = 0;
	}
}

/* common ------------------------------------------------------------ */

/* cfg.gps == 0 bypasses this */

void
quirky_gps_input(word c)
{
	switch (cfg.gps) {

	case GPS_GENERIC:
	case GPS_GARMIN:
	case GPS_MAGELLAN:
	case GPS_SIRF:
	case GPS_TRIPMATE:
	default:
		nmea_gps_input(c);
		break;

	case GPS_EARTHMATE:
	case GPS_ZODIAC:
		zodiac_gps_input(c);
		break;

	case GPS_AISIN:
		aisin_gps_input(c);
		break;

	case GPS_SEATALK:
		seatalk_gps_input(c);
		break;

	case GPS_ONCORE:
		oncore_gps_input(c);
		break;
	}
}

void
gps_initialize()
{
	/* allow any messages to get thru */

	SCI_drain();

	if (cfg.ttybaud == 0)
		SCI_config(4800, 0);     /* unless overridden in cfg */

	switch (cfg.gps) {

	case GPS_AISIN:              /* proprietary format */
		SCI_config(9600, 2);     /* Aisin must be 9600 8E1 */
		break;

	case GPS_SIRF:               /* command it to emit NMEA */
		SiRF_initialize();
		break;

	case GPS_TRIPMATE:           /* encourage it to emit NMEA */
		SCI_config(4800, 0);
		putstring("\r\nASTRAL\r\n"); /* normal NMEA from now on */
		break;

	case GPS_EARTHMATE:          /* uses rockwell zodiac format */
		SCI_config(9600, 0);
		putstring("\r\nEARTHA\r\n"); /* rockwell zodiac binary from now on */
		break;

	case GPS_ZODIAC:             /* proprietary format XXX some do NMEA ? */
		SCI_config(9600, 0);
		break;

	case GPS_SEATALK:
		SCI_config(4800, 3);     /* 9 bit characters */
		break;

	case GPS_ONCORE:          /* assume these are those not capable of NMEA */
		SCI_config(9600, 0);
		break;
	}
}

