const char ident[]
= "@(#)HAMDR 120 02.01.2006 MDR150";
/*
 *  Mon Jan  2 20:33:04 EET 2006
 *
 *  MDR-150/-450 RVALVE/etc amateur radio conversion.
 *
 *
 *  ----------------------------------------------
 *  Concepts of SmartBeaconing and CornerPegging,
 *  by Tony Arnerich KD7TA and Steve Bragg KA9MVA,
 *  have been used in the tracker. Many thanks.
 *  ----------------------------------------------
 *
*  XXX use 0:0000 to access ram up to 7FFF and SFRs down to F800 if possible
*  XXX shortpath with no RELAY, digipath with RELAY. if not getting thru
*  XXX last_report_seen_repeated == FALSE => dont RELAY ?
*  XXX holdq - leak if personality changes from tracker !
*  XXX do_relay, relay_area - keep 2 packets for my own reports !
*  XXX handle multibyte PIDs.
*  XXX wpt_stash should be hashed.
*  XXX longjumping in remote restart is wrong. XXX at least relse packet
 *
*  XXX short/int/word and char/byte haphazard use
*  XXX time for fixing the long arithmetic. naah.
*  XXX guards for cfg_assign() (remote config happens anytime)
*  XXX get rid of \r in strings, add if (cooked) ... to putchar().
*  XXX nomenclature: addr for shifted "binary", call for ascii.
*  XXX more than 1 arg to function pointer ? thunk manually XXX use jsr 3o,x
*  XXX for tidiness, getpacket() should structurise calls, digis, payload etc
 *
 *! 100 28.06.2005 stop using mker (just boot thru).
 *! 101 05.07.2005 gps seatalk. datetime in zodiac. tx_lockout.
 *! 102 07.07.2005 gps oncore. date/time changes. coordinate revamp. tx_led.
 *! 104 09.07.2005 DSR1 CTS1 DCD1 pins TRUE at boot. TRUE meaning XXX_pin(0).
 *! 105 09.07.2005 respond to ?APRS? query. ascii report with neg altitude.
 *!     12.07.2005 SCI SCSR OR oddity workaround. dunno if it helps.
 *!     18.07.2005 cfg_sanitise less facism.
 *! 106 18.07.2005 app_port auto.
 *! 107 18.07.2005 GPGGA for SiRF.
 *! 110 28.07.2005 cfg sanitizing real bug: cfg_general refers to GLOBAL cfg.
 *! 111 01.08.2005 eepeek
 *! 112 27.11.2005 rn3_function
 *! 113 27.11.2005 vane_to_deg[] gray-decode
 *! 117 14.12.2005 wx.c separation. some wind averaging. APZMDR default dst.
 *!     17.12.2005 wx: humidity from AN7.
 *! 118 17.12.2005 wind direction by pulses, not mph. no loss of information.
 *!     21.12.2005 many EDIV(EMUL()) changed to muldiv(). "sustained wind"
 *! 119 30.12.2005 print_string() into string.s just for fun.
 *!     01.01.2006 all crc with crc_generic.
 *! 120 02.01.2006 tracker_digi=none,relay,wide. txlimit fix.
 *
 ** Achtung: *****************************************************
 *
 *  code and data/stack are separate, but the compiler does not know it.
 *  not everything is possible:
 *  like, hc16-coff-gcc aborts at short = char * int;  arrrrrgh.
 *
 *****************************************************************
 *
 *  XXXN_pin() and read_inputs() to access:
 *
 *  XXX not up-to-date:
 *    DTR_1 input          Config access if TRUE during boot.
 *    RTS_1 input
 *    CTS_1 output
 *    DCD_1 output
 *
 *    DTR_2 input          PTT for some personalities
 *    RTS_2 input          sends 1750 Hz during PTT, in some p:s
 *    CTS_2 output      
 *    DCD_2 output         SERV in some p:s
 *    DSR_2 output         Indicates freq_0 in some p:s XXX not
 *
 *  read_inputs() return bits - 0303h = all inactive, 0 = all active
 *
 *  txproc aka packet_sender executes periodically from PIT, basically
 *     for (;;)
 *       slottime_and_pers()
 *       if (txq)
 *         transmit_some_of_it()
 *
 *  main does everything else.
 */

#include "standard.h"
#include "fix.h"
#include "crc.h"

byte aprs_ztamp[4];      /* day, hour, minute and second */

typedef struct {
	word iz, sp, pk, pc;
} jmp_buf[1];

jmp_buf restart;

STATIC word keycnt, keycnt_previously;
STATIC byte keybuf[16];

byte keydown, fill1;
byte keymap[12] = "9#6380527*41";  /* keypad.s knows about these */

#define INTEREST_IN_CFG 2
#include "hamdr_cfg.h"            /* all the cfg.foobar parameters */

struct cfg cfg;
struct nv  nv;

STATIC bool
	sec_timer(stamp *, word sec),
	select_profile(char *),
	process_manual_command(byte *arg);

STATIC void
	print_freq(dword),
	kickit(void),
	cfg_sanitize(void),
	publish_waypoints(int as_nmea);

word profile_offset;
char profile_tags[4][16];

#define PROFILE_OFFSET(i) ((i) * 8192) /* XXX this must go */

int debug;
int interactive = TRUE;

const byte aprs_destination[8] = "APZMDR";

#define MHz    * 1000L      /* silly */
#define SECOND * 1000L

#define FEND   0300
#define FESC   0333
#define TFEND  0334
#define TFESC  0335

#define KISS_DATA 0
#define KISS_TXD  1
#define KISS_P    2
#define KISS_ST   3
#define KISS_TXT  4
#define KISS_FD   5
#define KISS_HW   6
#define KISS_EXIT 255

#define AXALEN    7   /* [6] and SSID */
#define oSSID     6
#define AXMINLEN 15   /* dst src ctl */

#define DO_ALIAS  1   /* bitmask for is_myaddress() */
#define DO_DOMAIN 2

struct axaddr {
	byte call[6];
	byte cmdrspdig :1,
	     reserved  :2,
	     ssid      :4,
	     last      :1; /* unused, gcc had trouble with bitfields */
};

/* Right side LEDs from top to bottom */

#define LED_1 1
#define LED_2 2
#define LED_3 4

/*
 *  synthesizer 4094. supplies are on when corresponding bit is 0
 *  0x41 is no-connect (or no-document).
 */
#define RFCMD_AUDIO 0x02
#define RFCMD_PLL   0x04
#define RFCMD_VCO   0x08
#define RFCMD_RX    0x10
#define RFCMD_TX    0x20
#define RFCMD_MUTE  0x80
#define RFCMD_CONST 0x41   /* Audio, PLL and VCO on. AF unmuted (hope) */
#define TX_RX_OFF   0x71
#define RX_ON       0x61
#define TX_ON       0x51

#define POWER_LOWEST 3  /* tens of milliwatts (?) */

enum { OFF, TRYING, KEYING, ACTIVE, UNKEYING };
STATIC byte tx_state;
STATIC bool report_queuing_postponed;

/*
 *  tracker reports will not rot in txq,
 *  they are formatted freshly
 *  just before keying tx --- in PIT context !
 *
 *  ... but, tracker_digi=wide does it synchronously
 */

/*
 * XXX mucho copying happening. 150 cps, does not really matter.
 * tbd chop space[] into many many small and a few large packets.
 *
 * afsk.s knows the layout of ?xq (next, len and data[])
 * and uses hardcoded 512 to mask rx index.
 *
 * TXPACLEN must be large enough for any packet sent by tracker,
 * it does not check overflow.
 */

#define TXPACLEN 510      /* simpleminded XXX crc oddity */

struct txq {
	struct txq *next;
	word len;
	byte data[TXPACLEN + 2]; /* PACLEN does not include crc */
};

STATIC bool txq_maxed, fill2; /* XXX just use if (!txq_free || !rxq_free) ... */

struct txq *txq, *txq_free;
STATIC struct txq txq_space_1[10]; /* if 516 bytes each, 20: 10320 bytes */
STATIC struct txq txq_space_2[10]; /* quirky gcc will not allow [20] */

STATIC struct txq *holdq;          /* tracker digipeat q */
STATIC stamp holdq_mark;

/* rx side */

#define RXPACLEN 510      /* simpleminded - see afsk.s if changing */

struct rxq {
	struct rxq *next;
	word len;
	byte data[RXPACLEN + 2];
};

struct rxq *rxq, *rxq_free;
STATIC struct rxq rxq_space[10]; /* if 516 bytes each, 10: 5160 bytes */

static void rxq_relse(struct rxq *);

/* kiss/nmea collector */

STATIC byte nmea_buf[100]; /* NMEA says 82 should be enough */

STATIC struct txq *kiss_buf; /* beware, data[] inserted with -1 offset */

STATIC word
	nmea_len,
	kiss_len;

STATIC bool
	kiss_escaped,
	nmea_flushing,
	any_idea_of_location,
	sendnow,

	gps_has_spoken,
	gps_fix_available,
	target_active,
	target_active_manual,

	last_report_seen_repeated,
	report_inside_digi_area;   /* EOF=dunno, FALSE or TRUE */

STATIC stamp
	long_report_at,
	last_report_at,
	next_report_at,
	gps_fix_at,
	gps_spoke_at;

STATIC enum { VOID, FIX, PINNED } gps_validity;

STATIC struct vector
	gps,            /* RMC */
	target;         /* RMB */

STATIC byte
	target_name[12];    /* not empty if target_active */

STATIC word
	target_generation,
	reported_knots,
	reported_course,
	meters_travelled,
	long_meters_travelled,
	last_report_data_crc,
	gps_satellites;         /* being tracked */

STATIC long
	gps_altitude;

STATIC byte
	user_telemetry_active, fill45,
	user_telemetry[80];    /* external t. */

STATIC word
	telemetry_seqno;       /* internal t. */

STATIC bool
	lcd_gps_good,
	doing_lcd,
	doing_remote,
	fill5;

STATIC word
	manipulator;

STATIC stamp
	manipulator_timer;

/*
 *  Variables to gather responses to remote commands
 */
STATIC byte collected_output[80];
STATIC int  collected_output_len;

static dword atofreq(byte *);

byte *add_cse_spd(byte *, word course, word knots);
extern byte *encode_weather_ascii(byte *);

bool is_valid_symbol(byte tab, byte sym);

#define WX_SYMBOL   '_'

extern dword read_pulsacc(stamp *time);
extern dword ow_accvalue(word rank);

extern word
	read_AN1(), /* 10 MSbits valid, rest zero */
	read_AN3(),
	read_AN5(),
	read_AN7(),
	read_AN7_batt(word scale); /* decivolts, upto 25.5v */

/*
 *  Voltage drops if
 *  - lights turned on. hopefully a small drop.
 *  - engine is being started. implies engine was NOT running before.
 *  - engine killed. that is what we are looking for.
 *  Voltage rises if
 *  - lights turned off. hopefully a small effect on voltage.
 *  - engine started.
 */
STATIC stamp timebatt;
STATIC enum { ARRIVAL = 1, DEPARTURE } dockstat;
STATIC bool  lowbatt, fill3;

static void
battery_monitor()
{
	word batt = read_AN7_batt(cfg.battscale);

	if (batt >= cfg.lobatt) {
		if (lowbatt) {

			if (sec_timer(&timebatt, 10)) {
				lowbatt = FALSE;

				dockstat = DEPARTURE;
				sendnow  = TRUE;    /* car has been started */
			}
		}
	} else {
		if (!lowbatt) {

			if (sec_timer(&timebatt, 10)) {
				lowbatt = TRUE;

				dockstat = ARRIVAL;
				sendnow  = TRUE;    /* car has been stopped */
			}
		}
	}
}

static void
ticking()
{
	if (gps_validity != FIX)
		wallclock_run();

	if (cfg.lobatt)
		battery_monitor();

	if (1)                 /* XXX if mysymbol wx or telemetry_fmt */
		weather_run();
}

/* 0...255 is returned, maybe masked negative shorts */

static byte
telemetry_analog(word channel)
{
	if (channel == 'g')
		return (windgust());

	if (channel == 'w')
		return (windspeed());

	if (channel == 'v')
		return (muldiv(winddir(), 256, 360));

	if (channel == 'h')
		return (relative_humidity());

	if (channel >= 'A')
		return (ow_bytevalue(channel - 'A'));

	switch ((channel - '0') >> 1) {

	case 0: return (read_AN1() >> 8);
	case 1: return (read_AN3() >> 8);
	case 2: return (read_AN5() >> 8);
	case 3: return (read_AN7_batt(cfg.battscale));
	}
	return (0);
}

/* 0 or not 0 (not necessarily 1) word is returned */

static word
telemetry_digital(word channel)
{
	if (channel >= 'A')
		return (ow_bitvalue(channel - 'A'));

	switch (channel - '0') {

	case 0: return (read_inputs() & 0x0001);
	case 1: return (read_inputs() & 0x0002);
	case 2: return (read_inputs() & 0x0100);
	case 3: return (read_inputs() & 0x0200);
	case 4: return (read_PORTQS() & 1);     /* state of pin MISO */
	case 5: return (read_PACTL() & 0x8000); /* state of pin PAI */
	case 6: return (read_PACTL() & 0x0800); /* state of pin PCLK */
	case 7: return (read_PORTQS() & 4);     /* state of pin SCK */
	}
	return (0);
}

static byte *
add_counter_value(byte *p, dword d)
{
	byte *zb = p;

	sprint_010Lu(p, &d);
	if (*p != '0')
		return (p + strlen(p));

	/* find end of unwanted zeroes and tug the rest up */

	while (*p == '0' && p[1])
		++p;

	return (strcpyx(zb, p));
}

static byte *
telemetry_private(word channel, byte *p)
{
	if (channel == 'p')
		return (add_counter_value(p, read_pulsacc(NULL)));

	if (channel >= 'A')
		return (add_counter_value(p, ow_accvalue(channel - 'A')));

	return (p);
}

static byte *
internal_telemetry(byte *p)
{
	byte *channel;

	/* standardized part */

	p[0] = 'T';
	p[1] = '#';
	sprint_03u(p + 2, telemetry_seqno++ % 1000); 
	p[5] = COMMA;
	p += 6;

	channel = cfg.telemetry_format;

	do {
		sprint_03u(p, telemetry_analog(*channel));
		p[3] = COMMA;
		p += 4;
	} while (++channel < &cfg.telemetry_format[5]);

	do {
		*p++ = '0' + !!telemetry_digital(*channel);

	} while (++channel < &cfg.telemetry_format[5 + 8]);

	/* comment field for private stuff */

	while (*channel) {                        /* XXX bit naughty */
		*p++ = ' ';
		p = telemetry_private(*channel++, p);
	}
	return (p);
}

/*
 *  First byte after PID. NULL if no data in packet
 */
static byte *
packet_payload(byte *packet, int len)
{
	do {
		if (len <= AXALEN + 2)  /* address and ctl and pid */
			return (NULL);

		len    -= AXALEN;
		packet += AXALEN;

	} while ((packet[-1] & 1) == 0); /* not last */

	return (packet + 2); /* XXX how about PID extensions ? */
}

static int
valid_packet(struct rxq *p)
{
	word len = p->len - 2;       /* strip off the crc here */
	byte *data = p->data;
	word cksum;

	if (len < AXMINLEN) /* dst src ctl */
		return (FALSE);

	p->len = len;

	cksum = data[len] | (data[len + 1] << 8);

	return (cksum == crc_ax25(data, len));
}

/* NULL if a remote cfg packet was handled in stealth here */

static bool remote_cfg_packet_maybe(byte *, int);

static struct rxq *
getpacket()
{
	struct rxq *p;

	if (!rxq)
		goto out;

	spl7();            /* excessive XXX pit hits at 2 */
	p = rxq;
	if (p)
		rxq = p->next;
	spl0();

	if (!p)
		goto out;

	if (!valid_packet(p))
		goto drop;

	afsk_rx_led(TRUE);  /* turns off automatically after 0.5 seconds */

	if (remote_cfg_packet_maybe(p->data, p->len))
		goto drop;

	return (p);

drop:
	rxq_relse(p);
out:
	return (NULL);
}


/*
 *  gcc rejects txq_space[20], so doing it with smaller pieces _1 and _2.
 */
static void
init_trxq()
{
	int i;
	struct txq *t;

	txq_free = NULL;

	for (i = 0; i < __SIZEOF(txq_space_1); ++i) {
		t = &txq_space_1[i];
		t->next = txq_free;
		txq_free = t;
	}
	for (i = 0; i < __SIZEOF(txq_space_2); ++i) {
		t = &txq_space_2[i];
		t->next = txq_free;
		txq_free = t;
	}

	rxq_free = NULL;

	for (i = 0; i < __SIZEOF(rxq_space); ++i) {
		rxq_space[i].next = rxq_free;
		rxq_free = &rxq_space[i];
	}
}

static struct txq *
txq_alloc(word len)
{
	struct txq *q;

	if (len > TXPACLEN) {
		txq_maxed = TRUE;
		return (NULL);
	}
	spl7();

	q = txq_free;
	if (q) {
		txq_free = q->next;
		q->next = NULL;
		q->len  = TXPACLEN;
	} else
		txq_maxed = TRUE;

	spl0();

	return (q);
}

/* duplicated in OC2, needs high protection */

static void
txq_relse(struct txq *q)
{
	spl7();

	q->next = txq_free;
	txq_free = q;

	txq_maxed = FALSE;

	spl0();
}

static void
rxq_relse(struct rxq *q)
{
	q->len = 0;

	spl7();            /* excessive XXX pit hits at 2 */

	q->next = rxq_free;
	rxq_free = q;

	spl0(); /* only mainline can call */
}

static void
add_crc(struct txq *p)
{
	word cksum = crc_ax25(p->data, p->len);

	p->data[p->len++] = cksum;
	p->data[p->len++] = cksum >> 8;
}

/* unleash sender() too */

static void
txq_enter(struct txq *q)
{
	struct txq **pp;

	if (q->len == 0
	 || q->len > TXPACLEN)
		doh("txq_enter");

	add_crc(q);

	q->next = NULL;
	pp = &txq;

	hold_txproc();

	while (*pp)
		pp = &(*pp)->next; /* simple again */
	*pp = q;

	if (tx_state == OFF)
		tx_state = TRYING;

	okay_txproc();
}

/* kick transmission without queueing anything yet */

static void
kickit()
{
	hold_txproc();

	if (tx_state == OFF)
		tx_state = TRYING;

	okay_txproc();
}

/* queue data and start transmission if idle */

static void
sendit(void *packet, int len)
{
	struct txq *q = txq_alloc(len);

	if (q) {
		q->len = len;
		bcopy(packet, q->data, len);
		txq_enter(q);
	}
}

/* call[10] is ok, "OHnABC-15\0" */

static void
decode_ax25_address(char *call, const byte *a)
{
	int i;

	for (i = 0; i < oSSID; ++i) {
		if (a[i] == (' ' << 1))
			break;
		*call++ = toupper(a[i] >> 1);
	}
	i = (a[oSSID] >> 1) & 0xF;
	if (i) {
		*call++ = '-';
		if (i >= 10) {
			i -= 10;
			*call++ = '1';
		}
		*call++ = i + '0'; 
	}
	*call = 0;
}

/* a is shifted binary, call is string with or without -0 */

static bool
addr_equals(const byte *a, const byte *call)
{
	int i;
	byte c;

	for (i = 0; i < oSSID; ++i) {
		c = (a[i] >> 1);
		if (c == ' ')
			break;

		if (toupper(c) != toupper(*call++))
			return (FALSE);
	}
	if (*call && *call++ != '-')     /* step it past the  - */
		return (FALSE);

	c = (a[oSSID] >> 1) & 0xF;

	if (c == 0) {
		if (*call && *call++ != '0') /* step it past the 0 */
			return (FALSE);
	} else {
		if (c >= 10) {
			c -= 10;
			if (*call++ != '1')      /* step it past the 1 */
				return (FALSE);
		}
		if (c != *call++ - '0')      /* step it past ssid */
			return (FALSE);
	}
	/* call must not contain anything at this point  */

	return (*call == 0);
}

/* prefix variant, call must be in upper case and not empty */

static bool
addr_starts(const byte *a, const byte *call)
{
	do {
		if (toupper(*a >> 1) != *call)
			return (FALSE);
		++a;
	} while (*++call);

	return (TRUE);
}

/*
 *  call looks like "FOO-15\0"
 *  sloppiness for encoding digipath callsigns:
 *  call can be terminated by \0 or , or blank.
 *
 *  XXX prepare the outpacket just once and keep it good
 */
static void
encode_ax25_address(byte *buf, const char *call)
{
	int i;
	byte c;

	for (i = 0; i < oSSID; ++i) {
		c = toupper(*call);
		if (c == '-' || c == 0 || c == COMMA || c == ' ')
			c = ' ';
		else
			call++;              /* not yet at ssid or end */

		*buf++ = (c << 1);
	}
	c = (*call == '-') ? atoi(call + 1) : 0; /* "" or "-15" */

	/*
	 * 0x1E SSID
	 * 0x60 Reserved bits are "1"
	 * 0x80 Command/Response/Digipeated not set here.
	 * 0x01 Last is not set here.
	 */
	*buf = ((c & 0x0F) << 1) | 0x60;
}

void
putchar_pretty(byte c)
{
	if (c < ' ' || c >= 0x7F)
		c = DOT;

	putchar(c);
}

static void
print_shifted_call(byte *a)
{
	int i;
	byte buf[10];

	decode_ax25_address(buf, a);

	for (i = 0; buf[i]; ++i)
		putchar_pretty(buf[i]);
}

static void
prettyprint_packet(byte *data, int len)
{
	byte *dst = data;
	byte *end = data + len;
	bool rep = FALSE;

	/* source>destination */

	print_shifted_call(data + AXALEN);
	putchar('>');
	print_shifted_call(dst);

	data += 2 * AXALEN;

	/* any digis ? */

	while (data + AXALEN <= end && (data[-1] & 1) == 0) {

		if (data[oSSID] & 0x80)
			rep = TRUE;
		else if (rep) {
			putchar('*');
			rep = FALSE;
		}
		putchar(',');
		print_shifted_call(data);

		data += AXALEN;
	}
	if (rep)
		putchar('*');
#if 0
	if (data < end) {
		print_byte(*data++);        /* ctl */
		putchar(' ');

		if (data < end) {
			print_byte(*data++);    /* pid */
			putchar(' ');

			while (data < end)
				putchar_pretty(*data++);
		}
	}
#else
	if (data + 2 < end && data[1] == 0xF0) {
		data += 2;        /* ctl pid */
		putchar(':');

		/* drop newline accumulation from end */

		while (end > data && (end[-1] == '\r' || end[-1] == '\n'))
			--end;

		while (data < end) {
			byte c = *data++;

			/* pass scands and some benign control characters (mic_e) */

			if (c < 28 || c == 0x7F)
				c = DOT;

			putchar(c);
		}
	}
#endif
	eol();
}

static void
repeater_beep()
{
	tone_generator(4 * 1750); /* four times the Hertz */

	delay_ms(500);

	afsk_rx_on();                 /* and back */
}

static void
ctcss(word decihz)
{
	if (decihz) {
		tone_generator(muldiv(decihz, 4, 10));
	} else {
		afsk_rx_on();              /* restore */
	}
}

void synth_R(word);
void synth_NA(dword);

#define REF_6_250 (12800L * 1000 / 6250)
#define REF_5_000 (12800L * 1000 / 5000)

/*
 *  x in integer kHz
 *
 *  6.25 kHz step is tried first, if it is
 *  one of .000, .006, .012, .018 or .025
 *  5 kHz is taken otherwise.
 */
static void
vco_update(dword freq)
{
	word r;
	dword na;

	/*
	 *  see if 6.25 kHz step is workable,
	 *  else take 5 kHz
	 */

	na = EDIV(freq, 25, &r);

	if (r == 0
	 || r == 6
	 || r == 12 || r == 13
	 || r == 18 || r == 19) {

		ASLx(&na, 2);
		na += r / 6;
		r = REF_6_250;

	} else {
		na = EDIV(freq, 10, &r);

		ASLx(&na, 1);
		na += r / 5;
		r = REF_5_000;
	}
	synth_R(r);
	synth_NA(na);
}

static void
vco_rx(dword freq)
{
	vco_update(freq - cfg.rx_if_offset);
}

static void
vco_tx(dword freq)
{
	if (freq > 144 MHz
	 && freq < 146 MHz)
		goto go;

	if (freq > 432 MHz
	 && freq < 438 MHz)
		goto go;

	doh("illegal");
go:

	vco_update(freq);
}

/* outside packet_sender_task */

static void
update_rx_freq()
{
	hold_txproc();

	if (tx_state <= TRYING)
		vco_rx(cfg.freq);

	/* else unkeying will change rx */

	okay_txproc();
}

static bool
tx_is_locked_out()
{
	return (cfg.tx_lockout && RTS1_pin() != 0);
}

/*
 *  TRUE if selected PTT is active, and has
 *  changed since being selected as the ptt origin
 *  Parameter tx_lockout can block ptt.
 *  RTS1 pin can override parameter tx_lockout.
 */
static bool
ptt_pressed()
{
	static byte states, fill;
	bool is = FALSE;

	if (tx_is_locked_out())
		return (FALSE);

	switch (cfg.ptt_origin) {

		/* DTR2 modem control input on serial port 2 */
	case 0:
		is = (DTR2_pin() == 0);
		if (is)
			states |= 1;
		else
			states |= 2;

		if ((states &= 0x3) != 0x3)
			is = FALSE;
		break;

		/* MISO pin on D9 connector. pullup mod required */
	case 1:
		is = ((read_PORTQS() & 1) == 0);
		if (is)
			states |= 4;
		else
			states |= 8;

		if ((states &= 0xC) != 0xC)
			is = FALSE;
		break;

	/* case 2: void - no ptt */
	}
	return (is);
}

#if 0
static void
nullop()
{
}
#endif

/* TRUE if RF carrier present */

static bool
carrier(int for_LEDs)
{
	extern word afsk_dcd_l, afsk_dcd_h;

	if (cfg.use_softdcd)
		return ((for_LEDs ? afsk_dcd_h : afsk_dcd_l) == ~0);

	if (cfg.use_busy)
		return (busy_channel());

	return (read_rssi() > cfg.squelch);
}

/* TRUE if should keep deferring. Poor random, yes. */

static byte
not_at_all_random()
{
	static word rnd;

	if (rnd == 0)
		rnd = read_RSSI() * 0x3F;

	rnd *= 3475;
	rnd += 62303;

	return (rnd >> 8);
}

static bool
persist()           /* always FALSE if persistence=255 (never defer) */
{
	return (not_at_all_random() > cfg.persistence);
}

void
putchar(byte c)
{
	if (doing_remote) {
		if (collected_output_len < sizeof(collected_output))
			collected_output[collected_output_len++] = c;
		return;
	}
	if (doing_lcd)
		lcd_chr(c);
	else
		SCI_putc(c);
}

byte
getchar()
{
	word c;

	if (doing_remote)
		return (0);

	while ((c = SCI_getc()) == EOF)  /* wait until one is received */
		;
	return (c);
}

bool
ischar()
{
	return (doing_remote || SCI_canget());
}
 
static void
kiss_output(byte *data, int len, int port)
{
	byte c;

	SCI_putc(FEND);      /* purge garbage */
	SCI_putc(KISS_DATA | (port << 4));

	do {
		switch (c = *data++) {
		case FEND:
			SCI_putc(FESC);
			c = TFEND;
			break;
		case FESC:
			SCI_putc(FESC);
			c = TFESC;
			break;
		}
		SCI_putc(c);
	} while (--len);

	SCI_putc(FEND);
}

static void
kiss_flush_until_fend()
{
	if (kiss_buf) {
		txq_relse(kiss_buf);
		kiss_buf = NULL;
	}
}

/*
 * another data byte, not a FEND.
 * forget it if there has been an error after the last FEND.
 */
static void
kiss_chr(byte c)
{
	if (!kiss_buf)
		return;

	if (kiss_escaped) {
		kiss_escaped = FALSE;

		switch (c) {
		default:
			kiss_flush_until_fend();
			return;

		case TFEND:
			c = FEND;
			break;
		case TFESC:
			c = FESC;
			break;
		}
	} else if (c == FESC) {
		kiss_escaped = TRUE;
		return;
	}
	/*
	 *  XXX oddities to align data[] and command byte
	 *  kiss_buf->len is temporarily trampled by KISS_DATA chr.
	 */
	if (kiss_len >= TXPACLEN + 1) {
		kiss_flush_until_fend();
		return;
	}
	(kiss_buf->data - 1)[kiss_len++] = c;
}

/* one of those register allocator failures */

static word
kiss_to_msec(byte arg)
{
	word x = arg;

	return (10 * x);
}

/*
 *  FEND. Transmit the collected packet
 *  or process parameter.
 *  Unless there has been an error, of course.
 */
static void
kiss_fend()
{
	byte *dat, *end, cmd;

	if (kiss_buf && kiss_len && !kiss_escaped) {

		dat = kiss_buf->data - 1;
		end = dat + kiss_len;

		if (*dat == KISS_EXIT && kiss_len == 1) /* just C0 FF C0 will do */
			longjmp(restart, 2);

		/* other commands all have at least a byte of data appended */

		while (dat + 2 <= end) {
			cmd = *dat++;

			/* these settings are volatile */

			switch (cmd) {

			/* the other 2 delays cannot be set here */
			case KISS_TXD:
				cfg.data_delay = kiss_to_msec(*dat++);
				continue;

			case KISS_ST:
				cfg.slottime = kiss_to_msec(*dat++);
				continue;

			case KISS_P:
				cfg.persistence = *dat++;
				continue;
			}

			/* all the rest and always out */

			if ((cmd & 0x0F) == KISS_DATA) {
				cmd >>= 4;

				if (cmd != cfg.kiss_trick) {
					/* not ours. pass it downstream asynch line */

					if (cfg.kiss_trick)
						kiss_output(dat, end - dat, cmd);

				} else if ( ! tx_is_locked_out()) {

					/* data[] _is_ already aligned, normally */

					if (dat != kiss_buf->data)
						bcopy(dat, kiss_buf->data, end - dat);

					kiss_buf->len = end - dat; /* fix len, finally */
					txq_enter(kiss_buf);
					kiss_buf = NULL;
				}
			}
			break;
		}
	}
	kiss_escaped = FALSE;
	kiss_len = 0;
	if (kiss_buf == NULL)
		kiss_buf = txq_alloc(TXPACLEN);
}

/* harrrgh XXX */

static stamp
sec_to_msec(word sec)
{
	return (EMUL(sec, 1000));
}

static bool
sec_timer(stamp *t, word sec)
{
	if (*t) {                     /* running ? */
		if (time_past(*t)) {
			*t = 0;
			return (TRUE);        /* time is up */
		}
	} else
		future_sec(t, sec);       /* start it now */

	return (FALSE);
}

/*
 *  Personality 2
 */
static void
kiss_tnc()
{
	word leds, c;
	struct rxq *pk;

	/*
	 *  flush out garbage.
	 *  input garbage is rejected until a FEND allocates kiss_buf.
	 */
	SCI_putc(FEND);

	for (;;) {

		if (pk = getpacket()) {
			kiss_output(pk->data, pk->len, cfg.kiss_trick);
			rxq_relse(pk);
		}
		if ((c = SCI_getc()) != EOF) {
			if (c == FEND)
				kiss_fend();
			else
				kiss_chr(c);
		}
		leds = 0;
		if (tx_state <= TRYING && carrier(1))
			leds |= LED_3;
		if (txq)
			leds |= LED_2;
		if (txq_maxed)
			leds |= LED_1;

		green_leds(leds);
	}
}

/* sum is xor of characters after $ and before *XX */

static bool
nmea_checksum(byte *s)
{
	byte sum = 0;

	do {
		if (*s == 0)
			return (FALSE);

		sum ^= *s;
	} while (*++s != '*');

	return (sum == ascii_hex(s + 1, 2));
}

static void
nmea_print(byte *s)
{
	byte sum = 0;

	putchar('$');

	do {
		putchar(*s);
		sum ^= *s;
	} while (*++s);

	putchar('*');
	print_byte(sum);
	putchar('\r');
	putchar('\n');
}

static void
nmea_regen()
{
	static char s[56]
	 = "GPRMC,hhmmss,A,lati.tud,n,longi.tud,e,spd,cse,ddmmyy,,";

	/*  01234567890123456789012345678901234567890123456789012
	 *            1         2         3         4         5 */

	if (cfg.nmea_regen) {

		sprint_02u(s +  6, wallclock.hour);
		sprint_02u(s +  8, wallclock.minute);
		sprint_02u(s + 10, wallclock.second);

		sprint_02u(s + 46, wallclock.day);
		sprint_02u(s + 48, wallclock.month);
		sprint_02u(s + 50, wallclock.year);

		sprint_lat(s + 15, &gps.lat);
		sprint_lon(s + 26, &gps.lon);

		sprint_03u(s + 38, gps.knots);
		sprint_03u(s + 42, gps.course);

		nmea_print(s);
	}
}

static void
nmea_coord(long *pmm, byte *s, byte hem)
{
	word dg, mm, decim = 3;

	dg = EDIV(atoi(s), 100, &mm);

	while (*s && *s++ != DOT)  /* land over \0 or one past . */
		;

	do {
		mm *= 10;
		if (isdigit(*s))
			mm += *s++ - '0';
	} while (--decim);

	*pmm = EMUL(dg, _DEG_AS_mm) + mm;

	if (hem == 'S' || hem == 'W')
		*pmm = lnegate(*pmm);
}

static byte *
pick_nmea_field(byte **pp)
{
	byte *f = *pp;
	byte *cp = f;

	while (*cp != COMMA)
		if (*cp++ == 0)
			return (NULL);

	*cp++ = 0;
	*pp = cp;

	return (f);
}

/* XXX fails if all fields are wanted (no comma at end) */

static bool
split_nmea(byte *sentence, byte **v, char n)
{
	byte *fld;

	do {
		fld = pick_nmea_field(&sentence);
		if (!fld)
			return (FALSE);
		*v++ = fld;
	} while (--n);

	return (TRUE);
}

/*
 *  kmh = knots * 1.852 and 50 / 27 is about 1.85185
 *  accurate enough even without rounding properly.
 *  overflows short at 65536 / 50 = 1310 knots.
 */
static short
knots_to_kmh(word knots)
{
	return (muldiv(knots, 50, 27));
}

/*
 *  Same for knots to meter/second conversion.
 *  overflows around 500 knots.
 */
static short
knots_to_ms(word knots)
{
	return (muldiv(knots, 125, 243));
}

static void
new_gps_data()
{
	any_idea_of_location = TRUE;
	gps_validity = FIX;

	now(&gps_fix_at);
	gps_fix_available = TRUE; /* set flag to process it */
}

void
gps_is_speaking()
{
	gps_has_spoken = TRUE; /* flag that there has been some input from gps */
}

/* XXX non-NMEA path */

void
gps_fix(struct vector *v, struct wallclock *wc)
{
	/* be consistent with 0...359 */

	if (v->course > 359)
		v->course = 0;

	wallclock_reset();
	wallclock = *wc;

	hold_txproc();

	aprs_ztamp[0] = wallclock.day;
	aprs_ztamp[1] = wallclock.hour;
	aprs_ztamp[2] = wallclock.minute;
	aprs_ztamp[3] = wallclock.second;

	gps = *v;

	okay_txproc();

	new_gps_data();

	nmea_regen();
}

static void
jorgio(byte **date, byte **utc)
{
	static byte giorgio[6 + 6];

	byte *p;
	word i;

	for (i = 0; i < sizeof(giorgio); ++i)
		giorgio[i] = '0';

	p = *date + strlen(*date);
	i = 6;
	while (i && p > *date)
		giorgio[--i] = *--p;

	for (p = *utc; *p; ++p)
		if (*p == DOT)
			break;
	i = 6;
	while (i && p > *utc)
		giorgio[6 + --i] = *--p;

	*date = giorgio;
	*utc  = giorgio + 6;
}

static void
nmea_new_datetime(byte *date, byte *utc) /* ddmmyy and HHMMSS[.sssssssss] */
{
	byte *p;

	for (p = utc; *p; ++p)
		if (*p == DOT)
			break;

	if (p - utc != 6 || strlen(date) != 6)
		jorgio(&date, &utc);

	wallclock_reset();

	wallclock.year  = sscan_02u(date + 4);
	wallclock.month = sscan_02u(date + 2);
	wallclock.day   = sscan_02u(date + 0);

	wallclock.hour   = sscan_02u(utc + 0);
	wallclock.minute = sscan_02u(utc + 2);
	wallclock.second = sscan_02u(utc + 4);
}

/*
 * GPRMC,212909.00,A,4915.607,N,12310.537,W,000.0,360.0,111198,020.3,E*68
 *       0         1 2        3 4         5 6     7     8
 *       utc         lat        lon         knot  crs   date
 */
static void
do_gprmc(byte *sentence)
{
	byte *field[9];

	if (!split_nmea(sentence + 5 + 1, field, 9))
		return;

	/* assume good timekeeping from rtc, even if unfixed */

	nmea_new_datetime(field[8], field[0]);

	gps_is_speaking();

	if (*field[1] != 'A') {     /* A = Valid */
		gps_validity = VOID;
		return;
	}

	hold_txproc();

	aprs_ztamp[0] = wallclock.day;
	aprs_ztamp[1] = wallclock.hour;
	aprs_ztamp[2] = wallclock.minute;
	aprs_ztamp[3] = wallclock.second;

	nmea_coord(&gps.lat, field[2], *field[3]);
	nmea_coord(&gps.lon, field[4], *field[5]);

	gps.knots  = atoi(field[6]); /* discard the decimals */
	gps.course = atoi(field[7]);

	okay_txproc();

	new_gps_data();

	nmea_regen();
}

/* Wrap it to get 1...9999,0... steady sequence. */

static void
bump_target_generation()
{
	if (++target_generation > 9999)
		target_generation = 0;
}

/*
 * $GPRMB,A,4.08,L,EGLL,EGLM,5130.02,N,00046.34,W,004.6,213.9,122.9,A*3D
 *        1   2  3   4    5    6     7   8      9   10    11    12  13
 *
 *
 *     1    A         validity
 *     2    4.08      off track
 *     3    L         Steer Left (L/R)
 *     4    EGLL      last waypoint
 *     5    EGLM      next waypoint
 *     6    5130.02   Latitude of Next waypoint
 *     7    N         North/South
 *     8    00046.34  Longitude of next waypoint
 *     9    W         East/West
 *     10   004.6     Range
 *     11   213.9     bearing to waypt.
 *     12   122.9     closing velocity
 *     13   A         validity
 *     14   *3D       checksum
 *
 *  This is more complicated than it should be
 *  to handle separate waypoints all named "MOB".
 */
static void
do_gprmb(byte *sentence)
{
	struct vector old;
	byte *field[9];

	if (!split_nmea(sentence + 5 + 1, field, 9))
		return;

	if (*field[4] == 0) {
		target_active = FALSE;   /* XXX ok to clear now ? */
		return;                  /* no target waypoint in sentence */
	}

	/* the validity does not matter, or does it ? XXX */

	if (select_profile(field[4]))
		;                        /* current profile. continue anyway XXX */


	/* XXX */

	if (!target_active && !strcmp(field[4], "GIMME"))
		publish_waypoints(TRUE);

	/* whenever target changes, send a report */

	if (!target_active || strcmp(field[4], target_name))
		sendnow = TRUE;

	hold_txproc();

	target_active = TRUE;        /* XXX when to clear ? */

	/* remember what it used to be */

	old = target;

	/* copy new data */

	strncpyx(target_name, field[4], sizeof(target_name));

	nmea_coord(&target.lat, field[5], *field[6]);
	nmea_coord(&target.lon, field[7], *field[8]);

	/*
	 * What the waypoint name is or was, does not matter.
	 * Whenever coordinates do change, increment generation.
	 * Whoa, so silly in many ways.
	 */
	if (target.lat != old.lat
	 || target.lon != old.lon)
		bump_target_generation();

	okay_txproc();
}

/*
 * $GPGGA,123519,4807.038,N,01131.000,E,1,08,0.9,545.4,M,46.9,M,,*47
 *         0         1    2      3    4 5  6  7     8  9  10  11
 *
 *   1  Fix quality:
 *     0 = invalid
 *     1 = GPS fix (SPS)
 *     2 = DGPS fix
 *     3 = PPS fix
 *     4 = Real Time Kinematic
 *     5 = Float RTK
 *     6 = estimated (dead reckoning) (2.3 feature)
 *     7 = Manual input mode
 *     8 = Simulation mode
 *
 *   545.4,M Altitude, Meters, above mean sea level
 *   46.9,M  Height of geoid (mean sea level) above WGS84 ellipsoid
 */
static void
do_gpgga(byte *sentence)
{
	static dword ftscale = 1309106032; /* *= 0.3048 */

	byte *field[12];
	long altitude;

	if (!split_nmea(sentence + 5 + 1, field, 12))
		return;

	gps_satellites = atoi(field[6]);

	if (atoi(field[5]) == 0)   /* Fix quality */
		return;

	atol_signed(&altitude, field[8]);

	switch (*field[9]) {       /* Altitude unit */
	default:
		return;

	case 'M': /* meters */
	case 'm':
		break;

	case 'F': /* feet to meters */
	case 'f':
		altitude = SCALE32_signed(&altitude, &ftscale);
		break;
	}
	hold_txproc(); /* STED could be used XXX */

	gps_altitude = altitude;

	okay_txproc();
}


static void
nmea_sentence(byte *sentence)
{
	if (*sentence != 'G')
		return;

	if (!bcmp(sentence, "GPRMC,", 6)) {
		if (nmea_checksum(sentence))
			do_gprmc(sentence);
		return;
	}
	if (!bcmp(sentence, "GPRMB,", 6)) {
		if (nmea_checksum(sentence))
			do_gprmb(sentence);
		return;
	}
	if (!bcmp(sentence, "GPGGA,", 6)) {
		if (nmea_checksum(sentence))
			do_gpgga(sentence);
		return;
	}
}

void
nmea_gps_input(byte c)
{
	/* CR or LF terminates line */

	if (c == '\r' || c == '\n') {
		int ok = !nmea_flushing && nmea_len;

		nmea_flushing = FALSE;
		nmea_len = 0;

		if (ok) {
			if (nmea_buf[0] == '$')
				nmea_sentence(nmea_buf + 1);
			else
				process_manual_command(nmea_buf);
		}
		return;
	}
	/* other control characters or buffer overflow flush the line */

	if (c < ' ' || c >= 0x80
	 || nmea_len >= sizeof(nmea_buf) - 1) {    /* -1 for \0 at end */
		nmea_flushing = TRUE;
		return;
	}
	nmea_buf[nmea_len++] = c;
	nmea_buf[nmea_len] = 0;
}

void
sprint_03d_signed(byte *p, short nnn)
{
	if (nnn < 0) {
		if (nnn < -99)
			nnn = -99;
		p[0] = '-';
		sprint_02u(&p[1], -nnn);
	} else
		sprint_03u(p, nnn);
}

void
sprint_06Lu(byte *p, dword *nnn_nnn)
{
	word hi, lo;

	hi = EDIV(*nnn_nnn, 1000, &lo);

	sprint_03u(p + 0, hi);
	sprint_03u(p + 3, lo);
}

static bool
gps_data_is_void()
{
	return (gps_validity <= VOID);
}

static bool
gps_data_is_not_fit_for_report()
{
	return (!any_idea_of_location || gps_validity <= VOID); /* XXX for now */
}

/* in 10 seconds of silence, go from FIX to VOID */

static void
gps_inactive_timeout()
{
	if (gps_validity == FIX && time_past(gps_fix_at + 10 * 1000))
		gps_validity = VOID;
}

static bool
is_myaddress(byte *a, int which)
{
	if (cfg.mycall[0] && addr_equals(a, cfg.mycall))
		return (TRUE);

	if (which & DO_ALIAS)
		if (cfg.myalias[0] && addr_equals(a, cfg.myalias))
			return (TRUE);

	if (which & DO_DOMAIN)
		if (cfg.mydomain[0] && addr_equals(a, cfg.mydomain))
			return (TRUE);

	return (FALSE);
}

static bool
is_other_net(byte *a)
{
	byte call[10];

	decode_ax25_address(call, a);

	return (wildmat(cfg.altnet, call) == FALSE);
}

static void
show_aprs_symbols()
{
	int lines = 0;
	byte buf[16];
	word off = 0;

	while (off = next_aprs_symbol(off, buf)) {

		if (!more(&lines))
			break;

		putchar(buf[0]);
		putchar(buf[1]);
		putchar(' ');
		print_string(&buf[2]);
		eol();
	}
}

static void
show_aprs_symbol(word which)
{
	byte buf[16];
	word off = 0;

	putchar(which >> 8);
	putchar(which);

	while (off = next_aprs_symbol(off, buf)) {
		if (*(word *) buf == which) {
			putchar(' ');
			print_string(&buf[2]);
			break;
		}
	}
}

static word
lookup_aprs_symbol(byte *name)
{
	byte buf[16];
	word off = 0;

	while (off = next_aprs_symbol(off, buf))
		if (!strcmp(name, &buf[2]))
			return (*(word *) buf);

	return (0);
}

/* collection of data for report formatters */

struct repdata {
	word generation;
	struct vector vec;
	long altitude;
	word symbol;
	byte *digipath;
	byte *comments;
	byte type, name[10];
};

byte *
add_cse_spd(byte *p, word course, word knots)
{
	sprint_03u(&p[0], course ? course : 360); /* oddity */
	sprint_03u(&p[4], knots);
	p[3] = '/';

	return (p + 7);
}

static byte *
altimaatumiasma(long alt, byte *p)
{
	static dword ftscale = 215013;

	bool neg = FALSE;

	if (alt < 0) {
		alt = lnegate(alt);
		alt = SCALE16(&alt, &ftscale);
		if (alt > 99999)
			alt = 99999;
		neg = TRUE;
	} else {
		alt = SCALE16(&alt, &ftscale);
		if (alt > 999999)
			alt = 999999;
	}
	p = strcpyx(p, "/A=");

	sprint_06Lu(p, &alt);
	if (neg)
		*p = '-';

	return (p + 6);
}

static byte *
altimaatumice(long alt, byte *p)
{
	word rem, rest;

	alt += 10000; /* meters above marianes deep */

	if (alt < 0)
		alt = 0;
	else
	if (alt > 91 * 91 * 91L - 1)
		alt = 91 * 91 * 91L - 1; /* must be jean-luc, not jacques */

	rest = EDIV(alt, 91, &rem);

	p[0] = 33 + rest / 91;
	p[1] = 33 + rest % 91;
	p[2] = 33 + rem;
	p[3] = '}';

	return (p + 4);
}

static void
add_report_comments(struct txq *q, struct repdata *rd)
{
	byte *p = q->data + q->len;

	if (cfg.target_comment && target_active && target_name[0]) {
		p = strcpyx(p, " ->");
		p = strcpyx(p, target_name);
	}
	if (dockstat) {
		p = strcpyx(p, (dockstat == ARRIVAL)
			? " QTP"   /* I am going to enter dock or port. */
			: " QTO"); /* I have left dock or port. */
		dockstat = FALSE;
	}

	/* fixed comment ? */

	if (rd->comments && rd->comments[0]) {
		*p++ = ' ';
		p = strcpyx(p, rd->comments);
	}

	/* anything else ? in a shopkeeper's assistant kind of way */

	q->len = p - q->data;
}

/* turn coordinate digit ... to encoded '0', or 'A' or 'P' ... */

static byte
mic_e_dstbyte(byte num, byte base, int bit)
{
	if (bit)
		base = '0';
	num += base;

	return (num << 1);
}

static void
mic_e_destination(byte *p, struct repdata *rd, bool lon_offset)
{
	byte msg, custom;
	word D, d, M, m, H, h;

	d = EDIV(labs(rd->vec.lat), _DEG_AS_mm, &h);
	m = EDIV(h, 1000, &h);
	h /= 10;

	D = EDIV(d, 10, &d);
	M = EDIV(m, 10, &m);
	H = EDIV(h, 10, &h);

	custom = 'P';          /* assume not custom */
	msg = cfg.mic_e_msg;
	if (msg & ~7)          /* relies on the order of tab_mic_e_msg[] */
		custom = 'A';

	p[0] = mic_e_dstbyte(D, custom, msg & 4);
	p[1] = mic_e_dstbyte(d, custom, msg & 2);
	p[2] = mic_e_dstbyte(M, custom, msg & 1);

	p[3] = mic_e_dstbyte(m, 'P', negative_hemisphere(rd->vec.lat));
	p[4] = mic_e_dstbyte(H, 'P', ! lon_offset);
	p[5] = mic_e_dstbyte(h, 'P', ! negative_hemisphere(rd->vec.lon));

	p[oSSID] = ((cfg.mic_e_ssid & 15) << 1) | 0x60;
}

/* pbuf has space for TXPACLEN bytes, currently ample room */

int
encode_aprs_packet_mic_e(byte *pbuf, struct repdata *rd)
{
	byte *p = pbuf;
	word d, m, h, spd, cse, spH, csH;

	d = EDIV(labs(rd->vec.lon), _DEG_AS_mm, &h); /* degrees and milliminutes */
	h /= 10;                                     /* centiminutes */
	m = EDIV(h, 100, &h);                        /* minutes and centiminutes */

	if (d < 10)
		d += 190;
	else if (d >= 100 && d < 110)
		d += 80;

	if (m < 10)
		m += 60;

	spd = rd->vec.knots;
	cse = rd->vec.course;

	if (spd < 100)
		spd += 800;
	if (spd % 10 == 0)
		cse += 400;

	/* destination address ----------------------- */

	mic_e_destination(p, rd, d >= 100);
	p += AXALEN;

	/* source address ---------------------------- */

	encode_ax25_address(p, cfg.mycall);
	p += AXALEN;

	/* control and pid --------------------------- */

	p[-1] |= 1;        /* last address field in packet, poke a mark */

	*p++ = 0x03;
	*p++ = 0xF0;

	/* information field: id, longitude, speed and course. 9 bytes */

	d %= 100;

	spH = EDIV(spd,  10, &spd);
	csH = EDIV(cse, 100, &cse);

	spd *= 10;
	csH += spd;             /* to simplify formulas below */

	p[0] = '`';           /* APRS data type: MIC-E, current GPS data */

	p[1] = 28 + d;
	p[2] = 28 + m;
	p[3] = 28 + h;

	p[4] = 28 + spH;
	p[5] = 28 + csH;
	p[6] = 28 + cse;

	p[7] = rd->symbol;      /* code */
	p[8] = rd->symbol >> 8; /* table id */

	p += 9;

	if (cfg.send_altitude && rd->altitude)
		p = altimaatumice(rd->altitude, p);

	return (p - pbuf);
}

static byte *
add_coords_standard(byte *p, struct repdata *rd)
{
	/*
	 *  6103.259,N
	 *  6103.25N/
	 *  02803.123,E
	 *  02803.12Ec
	 *  01234567890
	 */
	sprint_lat(p, &rd->vec.lat); /* must fix the last characters */
	p[7] = p[9];
	p[8] = rd->symbol >> 8;
	p += 9;

	sprint_lon(p, &rd->vec.lon);
	p[8] = p[10];
	p[9] = rd->symbol;
	p += 10;

	if ((byte) rd->symbol == WX_SYMBOL) {
		p = encode_weather_ascii(p);
	} else {
		if (rd->vec.knots)
			p = add_cse_spd(p, rd->vec.course, rd->vec.knots);

		if (cfg.send_altitude && rd->altitude)
			p = altimaatumiasma(rd->altitude, p);
	}
	return (p);
}

static void
add_base91(byte *p, dword n)
{
	byte Hi, Lo;
	word hi, lo;

	hi = EDIV(n, 91 * 91, &lo);

	Hi = EDIV(hi, 91, &hi);
	Lo = EDIV(lo, 91, &lo);

	p[0] = '!' + Hi;
	p[1] = '!' + hi;
	p[2] = '!' + Lo;
	p[3] = '!' + lo;
}

static byte *
add_coords_compress(byte *p, struct repdata *rd)
{
	/* lat degree is 380926 ticks, lon 190463. use .24 */

	static long latscale = 106514629;
	static long lonscale =  53257314;

	dword x;

	/* Sym Table ID - harumpfh, dancing at the table
	 * Compressed Lat  YYYY
	 * Compressed Long XXXX
	 * Symbol Code */

	x = _90_DEG_AS_mm;
	x -= rd->vec.lat;
	x = SCALE24(&x, &latscale);
	add_base91(p + 1, x);

	x = _180_DEG_AS_mm;
	x += rd->vec.lon;
	x = SCALE24(&x, &lonscale);
	add_base91(p + 5, x);

	p[0] = rd->symbol >> 8;
	p[9] = rd->symbol;

	if (isdigit(p[0]))
		p[0] += 'a' - '0';  /* 0... overlays are a... with compression */

	p += 10;

	if (rd->vec.knots) {
		byte c = (rd->vec.course + 2) / 4;
		byte s = compressed_knots(rd->vec.knots);

		p[0] = '!' + (c < 90 ? c : 0); /* wrap if rounded 359 */
		p[1] = '!' + s;
		p[2] = '!' + 0x3E;    /* compression type byte - RMC oddity */

	} else if (rd->altitude) {
		word cs = compressed_altitude(&rd->altitude);

		p[0] = '!' + cs / 91; /* MSByte */
		p[1] = '!' + cs % 91; /* LSByte */
		p[2] = '!' + 0x36;    /* compression type byte - GGA oddity */

	} else {
		byte s = compressed_range(cfg.radio_range);

		p[0] = '{';        /* } align. pre-calculated radio range */
		p[1] = '!' + s;
		p[2] = '!' + 0x2E; /* compression type byte - GLL oddity */
	}
	return (p + 3);
}

/*
 *  Craziness for unique target names.
 *
 *  Any '*' characters in names are replaced by
 *  least significant decimal digits of target_generation.
 */
static void
target_name_generation_game(byte *buf, byte *end, word digits)
{
	while (--end >= buf)
		if (*end == '*') {
			*end = '0' + digits % 10;
			digits /= 10;
		}
}

static byte *
add_object_item_name(byte *p, byte *name, int minlen)
{
	byte *padto = p + minlen;
	byte *cutat = p + 9;

	while (p < cutat && *name)
		*p++ = *name++;
	while (p < padto)
		*p++ = ' ';

	return (p);         /* return "one past" */
}

static byte *
add_digipath(byte *p, byte *path)
{
	for (;;) {
		while (*path == COMMA || *path == ' ')
			++path;
		if (*path == 0)
			break;

		encode_ax25_address(p, path);
		p += AXALEN;

		while (*path && *path != COMMA && *path != ' ')
			++path;
	}
	p[-1] |= 1;        /* last address field in packet, poke a mark */

	return (p);
}

/* tbd calculate, do not integrate, the long area */
 
static void
select_digipath_and_comments(struct repdata *rd)
{
	if (cfg.longpath[0]) {

		if (cfg.long_area && long_meters_travelled >= cfg.long_area)
			goto longer;

		if (cfg.long_time
		 && time_past(long_report_at + sec_to_msec(cfg.long_time)))
			goto longer;
	}
	rd->digipath = cfg.digipath;
	rd->comments = cfg.short_comment;

	return;

longer:

	now(&long_report_at);
	long_meters_travelled = 0;

	rd->digipath = cfg.longpath;
	rd->comments = cfg.long_comment;
}

/* destination, source and possible digipeaters */

static byte *
aprs_packet_header(byte *p, struct repdata *rd)
{
	encode_ax25_address(p, cfg.aprs_dst[0] ? cfg.aprs_dst : aprs_destination);
	p += AXALEN;

	encode_ax25_address(p, cfg.mycall);
	p += AXALEN;

	p = add_digipath(p, rd->digipath);

	/* control = UI and pid = No_L3 */

	*p++ = 0x03;
	*p++ = 0xF0;

	return (p);
}

/* pbuf has space for TXPACLEN bytes, currently ample room */

static int
encode_aprs_packet_ascii(byte *pbuf, struct repdata *rd)
{
	byte *p = pbuf;
	byte *d;

	p = aprs_packet_header(p, rd);

	/*
	 *  payload
	 *                    !6103.52N/02806.18E>  normal station
	 *   ;KOIVU    *312359z6103.52N/02806.18E>  object (
	 *              )KOIVU!6103.52N/02806.18E>  item
	 */
	if (*rd->name == 0)
		goto noname;

	switch (rd->type) {

	default:                  /* normal station */
	noname:
		*p++ = '!';
		break;

	case 1:                   /* object */
		*p++ = ';';
		d = p;
		p = add_object_item_name(p, rd->name, 9);
		target_name_generation_game(d,  p, rd->generation);

		p[0] = '*';           /* live one */ 
		p[7] = 'z';           /* zulu time always */
		sprint_02u(p + 1, aprs_ztamp[0]);
		sprint_02u(p + 3, aprs_ztamp[1]);
		sprint_02u(p + 5, aprs_ztamp[2]);
		p += 8;
		break;

	case 2:                   /* item ( */
		*p++ = ')';
		d = p;
		p = add_object_item_name(p, rd->name, 3);
		target_name_generation_game(d,  p, rd->generation);

		*p++ = '!';           /* it's alive, it's alaaaaaiiiiv */
		break;
	}

	if (cfg.use_compression)
		p = add_coords_compress(p, rd);
	else
		p = add_coords_standard(p, rd);

	return (p - pbuf);
}

static void
remember_mine(byte *packet, word len)
{
	byte *payload = packet_payload(packet, len);

	if (payload) {
		len -= payload - packet;
		last_report_data_crc = crc_ax25(payload, len);
	} else
		;  /* huh ? */
}

/* fish around and grab interesting stuff in this little red ledger */

static void
my_report_data(struct repdata *rd)
{
	rd->generation = 0;             /* my position has never a generation */
	rd->type       = 0;
	rd->symbol     = cfg.mysymbol;
	rd->vec        = gps;
	rd->altitude   = gps_altitude;

	/* need a name to be an object or item */

	if (cfg.myname[0]) {
		bcopy(cfg.myname, rd->name, sizeof(rd->name));
		rd->type = cfg.mytype;
	}

	/* clear speed and altitude if no recent fix */

	if (gps_data_is_void()) {
		rd->altitude  = 0;
		rd->vec.knots = 0;
	}
	/* core out meaningless speeds */

	if (gps.knots <= 1)
		rd->vec.knots = 0;

	select_digipath_and_comments(rd);
}

static void
queue_aprs_report()
{
	struct txq *q;
	byte *p;
	struct repdata rd;

	my_report_data(&rd); /* fish around for data to be sent */

	/* XXX */

	if (rd.vec.knots) {
		reported_knots  = rd.vec.knots;
		reported_course = rd.vec.course;
	} else
		reported_knots  = 0;

	sendnow = FALSE;
	last_report_seen_repeated = FALSE;
	now(&last_report_at);

	/* (1) regular report -------------------------------- */

	if (any_idea_of_location) {

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		if (cfg.use_mic_e) {
			q->len = encode_aprs_packet_mic_e(q->data, &rd);
		} else {
			q->len = encode_aprs_packet_ascii(q->data, &rd);
		}
		add_report_comments(q, &rd);

		remember_mine(q->data, q->len);
		txq_enter(q);
	}

	/* (2) piggyback the internal telemetry as another packet ----- */

	if (cfg.telemetry_format[0]) {

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		/*
		 *  Headers up to and including PID, are duplicate of the first packet.
		 */
		p = aprs_packet_header(q->data, &rd);
		p = internal_telemetry(p);

		q->len = p - q->data;

		txq_enter(q); /* this one is not remembered */
	}

	/* (3) piggyback the external telemetry as well --------------- */

	if (user_telemetry_active) {

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		/*
		 *  Headers up to and including PID, are duplicate of the first packet.
		 *  Then payload, any user supplied text.
		 *  Protect against changes.
		 */
		p = aprs_packet_header(q->data, &rd);
		p = strcpyx(p, user_telemetry);

		q->len = p - q->data;

		txq_enter(q); /* this one is not remembered */
	}

	/* (4) piggyback the target report ------------------- */

	if (target_active && cfg.target_type) {

		/* only if target is active and was configured */

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		/* rd.stamp and digipath still ok */

		/* confusing XXX */

		bcopy(cfg.target_name[0] ? cfg.target_name : target_name,
			rd.name, sizeof(rd.name));

		rd.generation = target_generation;
		rd.type       = cfg.target_type;
		rd.symbol     = cfg.target_symbol;
		rd.vec        = target;

		/* tbd should the target report include OUR altitude ? XXX */

		target_active = target_active_manual; /* XXX for now */

		/* object or item can not be MIC-E encoded */

		q->len = encode_aprs_packet_ascii(q->data, &rd);

		txq_enter(q); /* this one is not remembered */
	}
}

/*
 *  1 milliminute is 1.852 meters, 0.001852 kilometers.
 *  1/540 is about 0.001852
 *  result, max 40'000 kilometers, fits in word nicely
 */
static word
lon_mm_to_km(dword mm)
{
	word corr;

	/*
	 *  abs distance from equator in full degrees, 0...90
	 *  cosine of that, 0...255
	 *  prescale milliminutes
	 *  finally convert to kilometers.
	 */
	corr = EDIV(labs(gps.lat), _DEG_AS_mm, NULL);
	corr = eight_bit_sin_from_abs_degrees(90 - corr);

	mm = EMULx(mm, corr);
	LSRx(&mm, 8);

	return (EDIV(mm, 540, NULL));
}

static word
lat_mm_to_km(dword mm)
{
	return (EDIV(mm, 540, NULL));
}

STATIC long mm_to_m_scale = 1852 * 65536L / 1000;

static dword
lon_mm_to_m(dword mm)
{
	word corr;

	corr = EDIV(labs(gps.lat), _DEG_AS_mm, NULL);
	corr = eight_bit_sin_from_abs_degrees(90 - corr);

	mm = EMULx(mm, corr);
	LSRx(&mm, 8);

	return (SCALE16(&mm, &mm_to_m_scale));
}

static dword
lat_mm_to_m(dword mm)
{
	return (SCALE16(&mm, &mm_to_m_scale));
}


/*
 *  Must return 0 or 1, not 0 or nz, for slimy reasons.
 *  temporary loss of fix does not hurt here. XXX ?
 */
static bool
inside_area(long y, long x, word kilometers)
{
	word dy, dx;

	if (!any_idea_of_location   /* cannot determine distances */
	 || kilometers == 0)        /* accepting the world */
		return (TRUE);

	/*
	 * XXX this fails around poles in many ways
	 *
	 * avoid unnecessary sqrt, compare dx and dy separately
	 * before calculating distance.
	 * first, latitude
	 * difference in milliminutes to kilometers of distance
	 */
	y -= gps.lat;
	if (y < 0)
		y = lnegate(y);

	dy = lat_mm_to_km(y);
	if (dy > kilometers)
		return (FALSE);

	/* longitudinal kilometers need correction */

	x -= gps.lon;
	if (x < 0)
		x = lnegate(x);

	dx = lon_mm_to_km(x);
	if (dx > kilometers)
		return (FALSE);

	if (ihypot(dy, dx) > kilometers)
		return (FALSE);

	return (TRUE);        /* close, maybe even a cigar */
}

/*
 *  storage in name[] can be changed at will,
 *  but not extended much. adding couple of bytes is ok.
 */
static bool
wash_waypoint_name(byte *name)
{
	byte *pat = cfg.waypoint_tidy;
	byte *s;
	byte caseconv = 'u';
	int best = 0;

	for (;;) {
		/* find start of next prefix, if any */

		while (*pat == ' ' || *pat == COMMA)
			++pat;
		if (*pat == 0)
			break;

		/* special -pat ? */

		if (*pat == '-') {
			switch (*++pat) {

			case 'a': /* -a means as is */
			case 'l': /* -l means lowercase */
			case 'u': /* -u means uppercase */
				caseconv = *pat;
				break;

			case 'F': /* -F means compress ssid to single hex digit */
			default:  /* - is the same, deprecated */

				s = name + strlen(name) - 3; /* position at "-15" */
				if (s > name
				 && s[0] == '-'
				 && s[1] == '1' && s[2] >= '0' && s[2] <= '5') {
					s[1] = 'A' + s[2] - '0';
					s[2] = 0;
				}
				break;
			}
			/* step over the code letter and start again */

			if (*pat)
				++pat;
			continue;
		}
		/*
		 *  else must be just another prefix string.
		 *  find length of match against this prefix.
		 */
		for (s = name; toupper(*s) == *pat; ++s, ++pat)
			if (*pat == 0 || *pat == ' ' || *pat == COMMA)
				break;

		if (*s == 0)      /* full match ? no waypoint at all  */
			return (FALSE);

		if (best < s - name)
			best = s - name;

		/* skip over possible mismatch in prefix */

		while (*pat && *pat != ' ' && *pat != COMMA)
			++pat;
		if (*pat == 0)
			break;
	}
	if (best)
		strcpyx(name, name + best);


	switch (caseconv) {

	case 'u':
		for (s = name; *s; ++s)
			*s = toupper(*s);
		break;

	case 'l':
		for (s = name; *s; ++s)
			*s = tolower(*s);
		break;

	case 'a':
		/* as is */
		break;
	}

	/* $, COMMA and * are nonos too wrt GPS. squeeze them out */

	for (s = name; *s; )
		switch (*s) {

		case COMMA:
		case '*':
		case '$':
			strcpyx(s, s + 1);
			break;

		default:
			++s;
		}

	/* return TRUE only if not washed totally invisible */

	return (*name ? TRUE : FALSE);
}

/*
 *  Keep position data for 100 latest.
 */
struct waypoint {
	byte tag[10];   /* callsigns, object/item names all fit in 9+1, unwashed */

	dword lat, lon; /* milliminutes */

	struct wallclock date; /* pillage some memory */

	struct waypoint *next; /* 1 byte indexes instead would save memory */
};

struct waypoint *wpt_stash, wpt_stash_space[100];
struct waypoint *wpt_tracked;

static void
init_wpt_stash()
{
	int i = 0;

	do {
		wpt_stash_space[i].next = wpt_stash;
		wpt_stash = &wpt_stash_space[i];
	} while (++i < __SIZEOF(wpt_stash_space));
}

static void
print_date(struct wallclock *wc)
{
	print_02u(wc->day);    putchar('.');
	print_02u(wc->month);  putchar('.');
	print_02u(wc->year);   putchar(' ');
	print_02u(wc->hour);   putchar(':');
	print_02u(wc->minute); putchar(':');
	print_02u(wc->second);
}

void
sprint_lat(byte *buf, long *lat)
{
	word D, M, mm;
	dword x;

	/*
	 *  6103.259,N     latitude
	 *  0123456789
	 */
	if (negative_hemisphere(*lat)) {
		buf[9] = 'S';
		x = lnegate(*lat);
	} else {
		buf[9] = 'N';
		x = *lat;
	}
	x = EDIV(x, 1000, &mm);
	D = EDIV(x, 60, &M);

	sprint_02u(buf + 0, D);
	sprint_02u(buf + 2, M);
	sprint_03u(buf + 5, mm);

	buf[4] = DOT;
	buf[8] = COMMA;       /* lat punctuation */
}

void
sprint_lon(byte *buf, long *lon)
{
	word D, M, mm;
	dword x;

	/*
	 *  06103.259,E
	 *  01234567890
	 */
	if (negative_hemisphere(*lon)) {
		buf[10] = 'W';
		x = lnegate(*lon);
	} else {
		buf[10] = 'E';
		x = *lon;
	}
	x = EDIV(x, 1000, &mm);
	D = EDIV(x, 60, &M);

	sprint_03u(buf + 0, D);
	sprint_02u(buf + 3, M);
	sprint_03u(buf + 6, mm);

	buf[5] = DOT;
	buf[9] = COMMA;        /* lon punctuation */
}

static byte *
sprint_wpt_coords(byte *buf, struct waypoint *wpt)
{

	/*
	 *  6103.259,N,     latitude
	 *  012345678901
	 */
	sprint_lat(buf, &wpt->lat);
	buf[10] = COMMA;
	buf += 11;

	/*
	 *  06103.259,E     longitude - NOTE no comma at end
	 *  012345678901
	 */
	sprint_lon(buf, &wpt->lon);
	buf[11] = 0;            /* put \0 for convenience */
	buf += 11;

	return (buf);
}

/*
 * $PMGNWPL,6103.52,N,02806.18,E,altitude,F/M,call,,icon,*XX <CRLF>
 * $PMGNWPL,6103.520,N,02806.180,E,0,M,OHnABC-15,HH:MM DD.MM.,a*XX <CRLF>
 * $PMGNWPL,6103.520,N,02806.180,E,,,OH5ABC-15*XX <CRLF>
 * $GPWPL,6103.52,N,02806.18,E,OH5ABC-15*XX <CRLF>
 */
static byte *
waypoint_comment_datetime(byte *cp, struct wallclock *date)
{
	/*
	 *  HH:MMz DD.MM.
	 *  0123456789012
	 */
	sprint_02u(cp +  0, date->hour);
	sprint_02u(cp +  3, date->minute);
	sprint_02u(cp +  7, date->day);
	sprint_02u(cp + 10, date->month);
	cp[2] = ':';
	cp[5] = 'z';
	cp[6] = ' ';
	cp[9] = cp[12] = DOT;

	return (cp + 13);
}

static void
print_waypoint(struct waypoint *wpt)
{
	static byte buf[82]; /* big enough, BUT NOT MUCH MORE */

	byte *cp = buf;

	switch (cfg.gps) {
	case GPS_MAGELLAN:
		cp = strcpyx(cp, "PMGNWPL");
		break;
	case GPS_GARMIN:
	default:
		cp = strcpyx(cp, "GPWPL");
		break;
	}
	*cp++ = COMMA;

	cp = sprint_wpt_coords(cp, wpt); /* 6103.259,N,06103.259,E */
	*cp++ = COMMA;

	if (cfg.gps == GPS_MAGELLAN) {
		*cp++ = '0';
		*cp++ = COMMA;
		*cp++ = 'M';
		*cp++ = COMMA;
	}
	/* waypoint name ends the sentence. max 9 and \0 */

	strcpyx(cp, wpt->tag);

	/* rejection here should not be so common */

	if (!wash_waypoint_name(cp))
		return;

	if (cfg.gps == GPS_MAGELLAN) {
		cp += strlen(cp);           /* f/f over the washed tag */

		*cp++ = COMMA;
		cp = waypoint_comment_datetime(cp, &wpt->date);
		*cp++ = COMMA;
		*cp++ = 'a';
		*cp = 0;
	}
	nmea_print(buf);
}

static void
print_distance(word num, word k)
{
#if 1
	if (k == 0) {
		if (num < 1000) {        /* 0 m 999 m */
			print_short(num);
			print_string(" m");
			return;
		}
		if (num < 10000) {         /* 1.0 km 9.9 km */
			num /= 100;
			print_short(num / 10);
			putchar(DOT);
			print_short(num % 10);
			print_string(" km");
			return;
		}
		num /= 1000;
	}
	print_short(num);
	print_string(" km");
#else
	if (k == 0) {
		if (num < 1000) {        /* __0 m_  999 m_ */
			if (num < 100)
				putchar(' ');
			if (num < 10)
				putchar(' ');
			print_short(num);
			print_string(" m ");
			return;
		}
		if (num < 10000) {         /* 1.0 km  9.9 km */
			num /= 100;
			print_short(num / 10);
			putchar(DOT);
			print_short(num % 10);
			print_string(" km");
			return;
		}
		num /= 1000;
	}
	/* __1 km __9 km */
	/* _10 km 999 km */

	/* 1000 km 9999 km   but these are handled as just "far" */
	/* 10000 km 20000 km */

	if (num < 100)
		putchar(' ');
	if (num < 10)
		putchar(' ');
	print_short(num);
	print_string(" km");
#endif
}

static void
print_distance_and_bearing(struct waypoint *wpt)
{
	long y, x;
	word dy, dx;
	word distance = 0;
	byte k = 0;
	byte ns = 'N';
	byte ew = 'E';

	y = wpt->lat - gps.lat;
	if (y < 0) {
		y = lnegate(y);
		ns = 'S';           /* johnny */
	}
	y = lat_mm_to_m(y);

	x = wpt->lon - gps.lon;
	if (x < 0) {
		x = lnegate(x);
		ew = 'W';           /* story */
	}
	x = lon_mm_to_m(x);

	if (y || x) {
		if (y >= 30000 || x >= 30000) {
			dy = EDIV(y + 500, 1000, NULL);
			dx = EDIV(x + 500, 1000, NULL);
			k = 1;
		} else {
			dy = y;
			dx = x;
		}
		if (dx / 3 > dy)
			ns = 0;
		if (dy / 3 > dx)
			ew = 0;

		distance = ihypot(dy, dx);
	}
	if (distance >= 1000 && k)
		print_string("far");

	else if (distance < 1 && !k)
		print_string("close");

	else {
		print_distance(distance, k);
		putchar(' ');
		if (ns)
			putchar(ns);
		if (ew)
			putchar(ew);
	}
}

static void
dump_waypoint(struct waypoint *wpt)
{
	byte buf[28]; /* 5130.019,N,00046.339,W */

	print_date(&wpt->date);
	putchar(' ');

	print_string(wpt->tag);
	print_string("            " + strlen(wpt->tag));

	sprint_wpt_coords(buf, wpt);
	print_string(buf);

	if (any_idea_of_location) {
		putchar(' ');
		print_distance_and_bearing(wpt);
	}
	eol();
}

static void
publish_waypoints(int as_nmea)
{
	struct waypoint *p;

	/* any number of unfilled entries at the tail. name then "" */

	for (p = wpt_stash; p && p->tag[0]; p = p->next)
		if (as_nmea)
			print_waypoint(p);
		else
			dump_waypoint(p);
}

/*
 *  this one will _always_ become the stash head
 */
static void
shuffle_waypoint_stash(byte *tag)
{
	struct waypoint *p, **pp;

	for (pp = &wpt_stash; ; pp = &p->next) {
		p = *pp;

		if (p->tag[0] == 0)
			goto new;        /* still space to grow into */

		if (!strcmp(p->tag, tag))
			goto old;        /* old entry, just refresh coords */

		if (p->next == NULL)
			break;           /* do not slip past end */
	}
	/* stash is full, oldest reused for new entry */
new:
	strncpyx(p->tag, tag, sizeof(p->tag));
old:
	*pp = p->next;           /* remove */

	p->next = wpt_stash;     /* prepend */
	wpt_stash = p;
}

static void
lcd_report(struct waypoint *wpt)
{
	if (manipulator && *tab_lcd[cfg.lcd] < '3')
		return;

	lcd_row(2);      /* LCD2: heard reports */
	doing_lcd++;

	lcd_str(wpt->tag);

	if (any_idea_of_location) {
		lcd_chr(' ');
		print_distance_and_bearing(wpt);
	}
	if (wpt == wpt_tracked)
		lcd_chr(4);

	doing_lcd--;
	lcd_finis();
}

/*
 *  addr is always binary callsign of packet source
 *  name is be non-empty, if this is an object or item
 *  lat and lon are signed milliminutes, NE positive.
 *
 *  tbd find and delete waypoints which are _now_ too distant.
 */
static void
process_report(byte *name, byte *addr, long lat, long lon)
{
	if (inside_area(lat, lon, cfg.waypoint_area)) {

		byte call[10];

		if (!name || !*name) {
			name = call;
			decode_ax25_address(call, addr);
		}
		/* becomes the first one, always */

		shuffle_waypoint_stash(name);

		wpt_stash->lat  = lat;
		wpt_stash->lon  = lon;
		wpt_stash->date = wallclock;  /* XXX some reports had timestamps */

		if (cfg.lcd)
			lcd_report(wpt_stash);

		if (cfg.waypoints)
			print_waypoint(wpt_stash);
	}
	if (cfg.tracker_digi)
		report_inside_digi_area = inside_area(lat, lon, cfg.digi_area);
}

#if 0 /* in tables.s */

static bool
overlayable(byte sym)
{
	byte *laid = "#&0>AW^_acinsuvz"; /* switch() _hogged_ space */

	do {
		if (*laid == sym)
			return (TRUE);
	} while (*++laid);

	return (FALSE);
}

static bool
is_valid_symbol(byte tab, byte sym)
{
	if (tab == '/' || tab == '\\')
		return (sym >= '!' && sym <= '~'); /* primary or alternate */

	if ((tab >= 'A' && tab <= 'Z')
	 || (tab >= '0' && tab <= '9'))
		return (overlayable(sym));

	return (FALSE);
}

static bool
is_valid_symbol_with_compression(byte tab, byte sym)
{
	if (tab == '/' || tab == '\\')
		return (sym >= '!' && sym <= '~'); /* primary or alternate */

	if ((tab >= 'A' && tab <= 'Z')
	 || (tab >= 'a' && tab <= 'j'))
		return (overlayable(sym));

	return (FALSE);
}

#endif /* in tables.s */

static int
is_digits(byte *p, int cc)
{
	byte *end = p + cc;

	do {
		if (!isdigit(*p))
			return (FALSE);
	} while (++p < end);

	return (TRUE);
}

static int
dec_base91(byte *b, dword *bin)
{
	byte c;
	int i = 0;
	dword x = 0;

	do {
		c = *b++ - '!';
		if (c >= 91)
			return (FALSE);
		x = EMULx(x, 91);
		x += c;
	} while (++i != 4);

	*bin = x;
	return (TRUE);
}

/*
 *   convert compressed APRS coordinates into milliminutes.
 *   return FALSE if not possible.
 */
static int
decode_compressed_aprs_coordinates(byte *compressed, long *lat, long *lon)
{
	static dword latscale =  676504197; /* 32.32 fxpt to  5400000 mm */
	static dword lonscale = 1353008394; /* 32.32 fxpt to 10800000 mm */

	dword y, x;

	if (!dec_base91(compressed + 0, &y)
	 || !dec_base91(compressed + 4, &x))
		return (FALSE);

	y = SCALE32(&y, &latscale);
	*lat = _90_DEG_AS_mm;
	*lat -= y;

	x = SCALE32(&x, &lonscale);
	*lon = x;
	*lon -= _180_DEG_AS_mm;

	return (TRUE);       /* good */
}

static bool
copy_object_name(byte *to, byte *from, int len)
{
	int i = len;

	bcopy(from, to, len);

	do {
		to[i] = 0;
	} while (--i >= 0 && to[i] == ' ');

	return (i >= 0); /* reject empties */
}

static int
copy_item_name(byte *to, byte *from)
{
	byte c;
	int i;

	/* name is 3-9 characters long, end in ! or _ */

	for (i = 0; ; ) {
		c = from[i];
		if (c < ' ')   /* XXX accept blanks and scands in item names */
			return (0);
		if (c == '!' || c == '_')
			break;
		to[i] = c;

		if (++i > 9)
			return (0);  /* too long */
	}
	if (i < 3)
		return (0);  /* too short */

	to[i] = 0;
	return (i); /* length, not including the separator */
}

static bool
process_compressed_report(byte *p, byte *name, byte *src)
{
	dword lat, lon;

	/*
	 * !/5L!"<*e9>
	 * 012345678901
	 *
	 * c-byte has only certain values XXX more checking for sT
	 */
	if (p[11] < ' ' || p[11] > '{')            /* align } */
		return (FALSE);

	if (!decode_compressed_aprs_coordinates(p + 2, &lat, &lon))
		return (FALSE);

	process_report(name, src, lat, lon);
	return (TRUE);
}

/* aprs coordinate in ascii format */

static dword
ascii_latitude(byte *p)
{
	dword lat;

	/* degrees and integer minutes to milliminutes, then add centiminutes */

	lat = EMUL(1000, sscan_02u(p + 0) * 60 + sscan_02u(p + 2));
	lat += sscan_02u(p + 5) * 10;

	if (p[7] == 'S')
		lat = lnegate(lat);

	return (lat);
}

static dword
ascii_longitude(byte *p)
{
	dword lon;

	/* degrees here 3 characters */

	lon = EMUL(1000, sscan_03u(p + 0) * 60 + sscan_02u(p + 3));
	lon += sscan_02u(p + 6) * 10;

	if (p[8] == 'W')
		lon = lnegate(lon);

	return (lon);
}

static bool
process_ascii_report(byte *p, byte *name, byte *src)
{
	/* lat and lon punctuation */

	if (p[5] != DOT || (p[8] != 'N' && p[8] != 'S'))  /* dot and N/S */
		return (FALSE);

	if (p[15] != DOT || (p[18] != 'E' && p[18] != 'W'))  /* dot and E/W */
		return (FALSE);


	if (!is_valid_symbol(p[9], p[19]))
		return (FALSE);


	/* lat and lon digits */

	if (!is_digits(p +  1, 4) || !is_digits(p +  6, 2)
	 || !is_digits(p + 10, 5) || !is_digits(p + 16, 2))
		return (FALSE);

	/* parrot badge for the inventor of 'port' in internet protocols */

	/*
	 *  store coordinates in milliminutes
	 *  6103.25N/02806.18E
	 *  0 2  56789
	 *  02806.18E
	 *  0  3  678
	 *  XXX KJASDKJSDFDD this format shuffling in nauseatic
	 */
	p += 1;                               /* point at lat start */

	process_report(name, src, ascii_latitude(p), ascii_longitude(p + 9));

	return (TRUE);
}

/*
 *  only 03F0 packets with some data get to maybe_xxx_waypoint()
 *
 *  p for payload
 */
static bool
maybe_ascii_waypoint(byte *packet, byte *p, int len)
{
	byte *endpacket = &packet[len];
	int i;
	byte *timestamp;
	byte name[12];          /* if object or item */

	/*
	 *  !6103.52N/02806.18E>
	 *  01234567890123456789
	 *            1
	 *
	 *  /312359z6103.52N/02806.18E>  same with timestamp
	 *
	 *
	 *  !/YYYYXXXX>csT    another encoding, the 'compressed' one.
	 *  01234567890123    this can have a timestamp in front as well
	 *            1
	 *
     *  csT could be just a " " with no s or T. sT are plain ignored.
	 *
	 *
	 *  ;OBJECT   *            live object, 7 byte timestamp follows
	 *   123456789             align (
	 *
	 *  )ITEM!                 live item, no timestamp
	 */

	if (p + 11 > endpacket)    /* too short - will test length again */
		return (FALSE);



	/* timestamp ? object ? */


	timestamp = NULL;
	name[0] = 0;            /* assume no name, use src callsign */


	switch (p[0]) {         /* first character first */
	default:
	case 0: case 255:       /* seduce gcc not to hog space with a jump table */
		return (FALSE);


	case '!':               /* the 'normal' ones */
	case '=':
		break;


	case ')':               /* item. no timestamp */
		p++;

		i = copy_item_name(name, p);
		if (i == 0)
			return (FALSE);
		if (p[i] != '!')      /* only live item will do */
			return (FALSE);

		p += i;               /* adjust p to coordinates */

		if (p + 11 > endpacket)  /* recheck length - not last check */
			return (FALSE);

		break;


	case ';':               /* object. always has a timestamp */
		p++;

		if (p[9] != '*')      /* only live object will do */
			return (FALSE);

		if (!copy_object_name(name, p, 9))
			return (FALSE);

		p += 9;               /* adjust p for timestamp case  */

		/* and fall thru to check/strip it */

	case '/':               /* timestamp in front */
	case '@':
		p++;
		timestamp = p;
		p += 6;             /* adjust p to match 'normal' indexing */

		if (p + 11 > endpacket)  /* recheck length - not last check */
			return (FALSE);

		switch (timestamp[6]) {
		default:
			return (FALSE);

		case 'z': /* DDHHMM zulu time */
		case '/': /* DDHHMM local time */
		case 'h': /* HHMMSS no day of month, but seconds */
			;
		}
		if (!is_digits(timestamp, 6))
			return (FALSE);

		/* good place to preset wallclock XXX but what year/month ? */
	}


	/* any timestamp in front has been handled, p[0] is undefined */


    /* compression used ? */

	if (is_valid_symbol_with_compression(p[1], p[10]))
		return (process_compressed_report(p, name, packet + AXALEN));


	/* rest is about the common !6103.52N/02806.18E> format */


	if (p + 19 > endpacket)    /* too short - final check for length */
		return (FALSE);

	return (process_ascii_report(p, name, packet + AXALEN));
}

/* c is known to be valid */

static int
mic_e_latchr(int c)
{
	if (c >= 'P')
		return (c - 'P');

	if (c >= 'A')
		return (c - 'A');

	return (c - '0');
}

/* p for payload */

static bool
maybe_mic_e_waypoint(byte *packet, byte *p, int len)
{
	int i;
	unsigned D, M, cm;
	dword lat, lon;         /* signed milliminutes */
	byte dst[oSSID];        /* "VQ2Y83" (just 6 letters) */

	if (p + 9 > &packet[len])     /* too short */
		return (FALSE);

	switch (p[0]) {         /* first character */
	default:
		return (FALSE);

	case '`':
	case '\'':               /* XXX rev.0, whatever it is, is ignored */
	}
	if (!is_valid_symbol(p[8], p[7]))
		return (FALSE);

	for (i = 1; i < 7; ++i)
		if (p[i] < 28 || p[i] >= 128)    /* not watertight */
			return (FALSE);

	/* ascify destination, checking for valid characters */

	for (i = 0; i < oSSID; ++i) {
		byte c = packet[i] >> 1;

		if (c == 'M' || c == 'N' || c == 'O'
		 || c > 'Z' || c < '0' || (c > '9' && c < 'A'))
			return (FALSE);

		dst[i] = c;
	}

	/* longitude. last minute checks */

	D = p[1] - 28;
	if (dst[4] >= 'A')
		D += 100;

	if (D >= 200)
		return (FALSE);
	if (D >= 190)
		D -= 190;
	else if (D >= 180)
		D -= 80;

	M = p[2] - 28;
	if (M >= 70)
		return (FALSE);
	if (M >= 60)
		M -= 60;

	cm = p[3] - 28; /* all combinations are valid */

	lon = EMULx(D, _DEG_AS_mm) + 1000 * M + 10 * cm;

	/* XXX check for range [0,180] degrees */

	if (dst[5] >= 'A')               /* west hemisphere */
		lon = lnegate(lon);

	/* latitude from prechecked destination address */

	D  = 10 * mic_e_latchr(dst[0]) + mic_e_latchr(dst[1]);
	M  = 10 * mic_e_latchr(dst[2]) + mic_e_latchr(dst[3]);
	cm = 10 * mic_e_latchr(dst[4]) + mic_e_latchr(dst[5]);

	lat = EMULx(D, _DEG_AS_mm) + 1000 * M + 10 * cm;

	/* XXX check for range [0,90] degrees */

	if (dst[3] < 'A')                /* south hemisphere */
		lat = lnegate(lat);

	process_report(NULL, packet + AXALEN, lat, lon);
	return (TRUE);
}

static bool
compare_ax25_address(byte *a, byte *b)
{
	return (bcmp(a, b, oSSID) || ((a[oSSID] ^ b[oSSID]) & 0x1E));
}

static word
length_without_crlf(byte *data, byte *end)
{
	while (end > data) {
		switch (end[-1]) {
		case '\r':
		case '\n':
		case ' ':
			--end;
			continue;
		}
		break;
	}
	return (end - data);
}

static void
sendit_not(byte *packet, word len)
{
	struct txq *p, **pp;
	byte *data, *hdata;
	word datalen;

	if (!holdq)
		return;

	data = packet_payload(packet, len);
	if (!data)
		return;

	/*
	 *  Do not include any CR, LF or blank
	 *  at end of packet in the comparisons
	 */
	datalen = length_without_crlf(data, &packet[len]);

	/*
	 *  Compare just source addresses and dataparts.
	 *  holdq access needs no protection.
	 */
	for (pp = &holdq; p = *pp; pp = &p->next) {

		if (compare_ax25_address(p->data + AXALEN, packet + AXALEN))
			continue;

		hdata = packet_payload(p->data, p->len);
		if (!hdata)
			doh("holdq");

		if (datalen != length_without_crlf(hdata, &p->data[p->len])
		 || bcmp(data, hdata, datalen))
			continue;

		/* found it. remove and release */

		*pp = p->next;
		txq_relse(p);

		break;
	}
}

static word
two_bits_of_random()
{
	return (read_RSSI() & 3); /* not the averaged one */
}

static void
sendit_later(byte *packet, word len)
{
	struct txq *p, **pp;
	byte *data, *pdata;
	word datalen;

	/*
	 *  Allow only one packet from single source address. XXX ?
	 *  It could be that this and the held packet are the same,
	 *  but release/allocate is always done. Should not happen often.
	 *
	 *  holdq access needs no protection.
	 */
	for (pp = &holdq; p = *pp; pp = &p->next) {

		if (compare_ax25_address(p->data + AXALEN, packet + AXALEN))
			continue;

		*pp = p->next; /* remove from q and reuse */
		txq_relse(p);
		break;
	}
	p = txq_alloc(len);
	if (!p)
		return; /* oopsidaisies */

	p->len = len;
	bcopy(packet, p->data, len);

	while (*pp)
		pp = &(*pp)->next;   /* in case for() broke early */

	if (!holdq)
		future_sec(&holdq_mark, cfg.relay_viscosity + two_bits_of_random());

	*pp = p;
}

static void
send_held_packets()
{
	struct txq *p;

	/* holdq access needs no protection */

	while (p = holdq) {
		holdq = p->next;

		txq_enter(p);

		if (cfg.waypoints == FALSE) {
			print_string("relay: ");
			prettyprint_packet(p->data, p->len);
		}
	}
}

/* already zero not allowed */

static void
ssid_decrement(byte *a)
{
	byte ssid;

	a += oSSID;

	ssid = ((*a & 0x1E) >> 1) - 1;
	*a &= ~0x1E;
	*a |= ssid << 1;
}

/*
 *  Pared down special case of digipeat() for tracker use.
 *  Packet is known to be an APRS report, worthy of repeat.
 */
static void
do_relay(byte *packet, int len)
{
	byte *digi, last;

	if (packet[AXALEN + oSSID] & 1)
		return;                        /* no digis, src is last */

	digi = &packet[2 * AXALEN];        /* first digi */

	if (&digi[AXALEN] > &packet[len])  /* packet is invalid ? */
		return;

	if (digi[oSSID] & 0x80) {          /* already repeated by 1st digi */

		sendit_not(packet, len);
	} else {
		if (!is_myaddress(digi, DO_ALIAS)) /* XXX hardwire "RELAY" ? */
			return;

		/* committed. update digipeater callsign */
		/* XXX reserved bits in digi-ssid get overwritten. good or bad. */

		last = digi[oSSID] & 1;            /* remember if Last */

		encode_ax25_address(digi, cfg.mycall);
		digi[oSSID] |= last | 0x80;        /* Digipeated, maybe Last */

		/* XXX swap packets between rx and txq_delayed */

		sendit_later(packet, len);
	}
}

/*
 *  More extensive digipeating, APRS style.
 *  Packet is known to be non-empty "UI Text".
 *  Touching packet data is ok at this point.
 */
static word
dupe_digest(byte *packet, int len)
{
	byte *data = packet_payload(packet, len);
	word crc;

	/*
	 * source, with ssid
	 * destination, no ssid
	 * payload, no trailing whitespace
	 */
	crc = crc_generic(packet + AXALEN, -1, 0x8408, AXALEN);
	crc = crc_generic(packet, crc, 0x8408, oSSID);
	if (data)
		crc = crc_generic(data, crc, 0x8408,
			length_without_crlf(data, &packet[len]));

	return (crc);
}

struct dupe {
	struct dupe *next;
	word  digest;
	stamp mark;        /* XXX very sloppy ram usage */
};

static struct dupe dupe_space[128], *dupeq, *dupeq_free;

static void
dupeq_refresh()
{
	struct dupe *d;

	/* need init, if both ptrs are still null */

	if (!dupeq_free && !dupeq) {
		for (d = dupe_space; d < &dupe_space[__SIZEOF(dupe_space)]; ++d) {
			d->next = dupeq_free;
			dupeq_free = d;
		}
	}

	/* release all stale ones from the top */

	while ((d = dupeq) && time_past(d->mark)) {
		dupeq = d->next;
		d->next = dupeq_free;
		dupeq_free = d;
	}
}

static bool
is_dupe(byte *packet, int len)
{
	struct dupe *d;
	word digest = dupe_digest(packet, len);

	for (d = dupeq; d; d = d->next)
		if (d->digest == digest)
			return (TRUE);

	return (FALSE);
}

static void
sendit_digi(byte *packet, int len)
{
	struct dupe *d, **pp;

	if ((d = dupeq_free) == NULL)
		return;                    /* clogged, discard */

	dupeq_free = d->next;
	d->next = NULL;

	/* promptly out */

	sendit(packet, len);

	future_sec(&d->mark, cfg.dupe_age);
	d->digest = dupe_digest(packet, len);

	/* voyerism */

	if (cfg.waypoints == FALSE) {
		print_string("digi: ");
		prettyprint_packet(packet, len);
	}

	/* append */

	pp = &dupeq;
	while (*pp)
		pp = &(*pp)->next;
	*pp = d;
}

static void
aprs_digipeat_if_needed(byte *packet, int len)
{
	byte *digi;
	byte ssidx;

	/* initialize and purge stale entries */

	dupeq_refresh();

	/* ignore those just recently digipeated */

	if (is_dupe(packet, len))
		return;

	/* ignore packets for and from me, including aliases */

	if (is_myaddress(packet, DO_ALIAS)
	 || is_myaddress(packet + AXALEN, DO_ALIAS))
		return;

	digi = &packet[2 * AXALEN];
	if (digi[-1] & 1)
		digi = NULL;                    /* no digis */

	if (packet[oSSID] & 0x1E) {
		/*
		 *  destination-SSID based digipeating,
		 *  only possible with empty or virgin digipath
		 */
		if (digi == NULL || (digi[oSSID] & 0x80) == 0) {
			ssid_decrement(packet);

			sendit_digi(packet, len);  /* simply this */
			return;
		}
	}

	if (!digi)
		return;                          /* nothing to do */

	/* short-circuit RELAY --- XXX */

	if (addr_equals(digi, "RELAY")) {
		digi[oSSID] |= 0x80;             /* nasty */
		digi += AXALEN;

		if (digi[-1] & 1)                /* huh ? just RELAY */
			return;
	}
	while (digi + AXALEN <= &packet[len - 1]) {

		ssidx = digi[oSSID];             /* remember Last-bit */

		if ((ssidx & 0x80) == 0) {       /* not repeated */

			if (addr_starts(digi, "WIDE")) {
				if (ssidx & 0x1E) {
					ssid_decrement(digi);

					/* fresh WIDEn-N or no available slot before this ? */

					if (((ssidx ^ digi[4]) & 0x1E) == 0
					 || digi == &packet[2 * AXALEN])
						goto squeeze_me_in;

					/* else overwrite marker to prev slot */

					digi -= AXALEN;
					goto marker_slot;
				}
				goto claim_slot;
			}
			if (addr_starts(digi, "TRACE")) {
				if (ssidx & 0x1E) {
					ssid_decrement(digi);

		squeeze_me_in:
					/* XXX knowledge about struct rxq */
					if (len + AXALEN > TXPACLEN)
						return;

					bcopy(digi, digi + AXALEN, &packet[len] - digi);
					len += AXALEN;

					goto marker_slot;
				}
				goto claim_slot;
			}
			if (addr_starts(digi, "RELAY"))
				return;

			if (addr_equals(digi, cfg.myalias))
				goto claim_slot;

			if (addr_equals(digi, cfg.mycall))
				goto slot_claimed;

			/* hmph. ignore the packet */

			break;

		marker_slot:
			ssidx = 0;            /* inserted slot is not last */

		claim_slot:
			encode_ax25_address(digi, cfg.mycall);

		slot_claimed:
			digi[oSSID] |= (ssidx & 1) | 0x80; /* Last? and repeated */

			sendit_digi(packet, len);
			break;
		}
		if (ssidx & 1)
			break;                        /* (repeated and) Last */

		if (is_myaddress(digi, 0))
			break;                       /* _I_ have already repeated it */

		digi += AXALEN;
	}
}

STATIC stamp news_timer;
STATIC byte news, news_finger, news_end, news_space[256], fill6;

static void
clear_news()
{
	news = FALSE;
	news_finger = news_end = 0;
	bzero(news_space, sizeof(news_space));
}

#define BULLHUH 0xA5
#define BULLSEP 0xA1
#define BULLEOT 0xDB

static byte
news_chr(byte c)
{
	return ((c == BULLSEP || c < ' ') ? BULLHUH : c);
}

static void
news_addc(byte c)
{
	news_space[news_end++] = c;
}

static bool
news_exists(byte *text, word len)
{
	byte *t;
	word n, x = 0;

	do {
		if (news_space[x++ & 0xFF] == BULLSEP) {
			t = text;
			n = len;
			while (news_space[x++ & 0xFF] == news_chr(*t++)) {
				if (--n == 0) {
					if (news_space[x & 0xFF] == 0
					 || news_space[x & 0xFF] == BULLEOT)
						return (TRUE);
					break;
				}
			}
		}
	} while (x < 256);

	return (FALSE);
}

static word
news_clean(byte **text, word len)
{
	byte *end = *text + len;

	while (*text < end && **text <= ' ')
		++*text;
	while (*text < end && end[-1] <= ' ')
		--end;

	return (end - *text);
}

static void
store_news(byte *a, byte *spec, byte *text, word len)
{
	byte *p;
	byte buf[10];

	len = news_clean(&text, len);
	if (len == 0)                   /* empty ? */
		return;

	if (news_exists(text, len))     /* already have it ? */
		return;

	news_addc(BULLEOT);

	news_finger = news_end;

	decode_ax25_address(buf, a);
	p = buf;
	while (*p && *p != '-')      /* ssid not printed */
		news_addc(*p++);

	if (spec) {
		news_addc('/');
		p = spec + 5;
		do {
			if (*spec > ' ')
				news_addc(*spec);
		} while (++spec < p);
	}
	news_addc(BULLSEP);

	do {
		news_addc(news_chr(*text++));
	} while (--len);

	news = TRUE;
	future_msec(&news_timer, 2000);
}

static void
billboard()
{
	byte x;

	if (!news || manipulator)
		return;

	if (!time_past(news_timer))
		return;

	lcd_row(4);      /* LCD4: all sorts of junk */

	x = news_finger;
	do {
		lcd_chr(news_space[x]);

		if (++x == news_end) {
			lcd_finis();
			future_msec(&news_timer, 2000);
			news = TRUE;
			news_finger = news_space[news_end] ? news_end : 1; /* 1st EOT */
			return;
		}
	} while (lcd_row_space());

	lcd_finis();
	future_msec(&news_timer, (news == 2) ? 250 : 2000);
	news = 2; /* still true */
	news_finger++;
}

static void
lcd_traffic(byte *packet, int len)
{
	word nleft;
	byte *data;

	if (manipulator || news)
		return;

	lcd_row(4);      /* LCD4: all sorts of junk */
	doing_lcd++;

	print_shifted_call(packet + AXALEN);
	lcd_chr('>');
	print_shifted_call(packet);

	nleft = lcd_row_space();
	if (nleft > 1
	 && (data = packet_payload(packet, len))
	 && (data[-1] == 0xF0)) {

		lcd_chr(' ');
		nleft--;

		if (nleft > &packet[len] - data)
			nleft = &packet[len] - data;

		lcd_buf_pretty(data, nleft);
	}
	doing_lcd--;
	lcd_finis();
}

static void
process_keys()
{
	word x;

	keycnt_previously = keycnt;

	if (cfg.lcd) {
		lcd_row(4);      /* LCD4: all sorts of junk */
		lcd_buf(keybuf, keycnt);
		lcd_finis();
	}
	if (keycnt == 0)
		return;

	if (keycnt == 4 && keybuf[3] == '#' && is_digits(keybuf, 3)) {
		cfg.freq    = EMULx(EDIV(cfg.freq, 1000, NULL), 1000) + atoi(keybuf);
		cfg.tx_freq = 0;
		update_rx_freq();
		goto accept;
	}

	if (keycnt == 7 && keybuf[6] == '#' && is_digits(keybuf, 6)) {
		keybuf[6] = 0;
		cfg.freq    = atofreq(keybuf);
		cfg.tx_freq = 0;
		update_rx_freq();
		goto accept;
	}

	if (keycnt == 3 && keybuf[2] == '#' && is_digits(keybuf, 2)) {
		x = atoi(keybuf);

		if (x < __SIZEOF(cfg.channel)) {
			cfg.freq    = cfg.channel[x].freq;
			cfg.tx_freq = cfg.channel[x].tx_freq;
			update_rx_freq();
		}
		goto accept;
	}
	return;

accept:
	keycnt = 0;
}

/* XXX ptt and manipulator ought to be merged */

static bool
manipulator_state()
{
	static byte states, fill;
	bool is = FALSE;

	switch (cfg.manipulator) {

#if 0
		/* SCK pin on D9 connector. pullup mod required */
	case 1:
		is = ((read_PORTQS() & 4) == 0);
		if (is)
			states |= 4;
		else
			states |= 8;

		if ((states &= 0xC) != 0xC)
			is = FALSE;
		break;
#endif

		/* Keypad * */
	case 1:
		is = (keydown == '*');
		break;

		/* DTR2 modem control input on serial port 2 */
	case 2:
		is = (DTR2_pin() == 0);
		if (is)
			states |= 1;
		else
			states |= 2;

		if ((states &= 0x3) != 0x3)
			is = FALSE;
		break;
	}
	return (is);
}

static char *
manipulator_exec(int x)
{
	switch (x) {

	case 1:
		return ("Send now");

	case 10:
		sendnow = TRUE;
		return ("Pending");

	case 2:
		return ("Clear");

	case 20:
		clear_news();
		lcd_gps_good = 2;       /* XXX force redraw */
		wpt_tracked = NULL;
		target_active = target_active_manual = FALSE;
		return ("Cleared");

	case 3:
		return ("Track");

	case 30:
		wpt_tracked = wpt_stash;
		if (wpt_tracked->tag[0] == 0)
			wpt_tracked = NULL;

		return (wpt_tracked ? wpt_tracked->tag : (byte*) "None");

	case 31:
		manipulator--; /* XXX lump 32, 33 etc into one */

		/* absurd */
		if (wpt_tracked)
			wpt_tracked = wpt_tracked->next;
		if (wpt_tracked == NULL || wpt_tracked->tag[0] == 0)
			wpt_tracked = wpt_stash;
		if (wpt_tracked->tag[0] == 0)
			wpt_tracked = NULL;

		return (wpt_tracked ? wpt_tracked->tag : (byte*) "None");

	case 4:
		return ("Mark");

	case 40:
		bump_target_generation();
		target = gps;
		target_active = target_active_manual = TRUE;

		sendnow = TRUE;
		return ("Mark set");
	}
	return ("No action");
}

static void
lcd_message(word row, char *what)
{
	if (cfg.lcd) {
		lcd_row(row);
		if (what)
			lcd_str(what);
		lcd_finis();
	}
}

static void
manipulator_call(int x)
{
	lcd_message(4, manipulator_exec(x));

	future_msec(&manipulator_timer, 1000);
}

static void
do_manipulator()
{
	/* idle state, wait a keypress */

	if (manipulator == 0) {
		if (manipulator_state())
			manipulator_call(manipulator = 1);
		return;
	}
	/* during keypress, step thru selections */

	if (manipulator_state()) {
		if (time_past(manipulator_timer))
			manipulator_call(++manipulator);
		return;
	}
	/* key released, execute it really */

	if (manipulator < 10) {
		manipulator_call(manipulator *= 10);
		return;
	}
	/* clear the pending state, back to idle */

	if (time_past(manipulator_timer)) {
		lcd_message(4, NULL);
		manipulator = 0;
	}
}

static void
fix_ssid_counter(byte *call)
{
	byte *p;
	byte count, left;

	/* some malicious cases not tested against */

	for (p = call; *p; ++p)
		if (*p == '-') {
			count = p[-1] - '0';
			left  = p[1]  - '0';
			if (count < 10 && left < 10 && p[2] == 0 && count > left) {
				count -= left;
				p[1] = count + '0';
			}
			break;
		}
}

static byte *
ack_path_reversal(byte *p, byte *odigis)
{
	byte *od = odigis;
	byte call[10];

	/* over used digis */

	while ((od[-1] & 1) == 0 && (od[oSSID] & 0x80))
		od += AXALEN;

	/* flip those */

	while (od > odigis) {
		od -= AXALEN;

		decode_ax25_address(call, od);

		if (!strcmp(call, "RELAY"))  /* XXX should not happen ideally */
			continue;

		fix_ssid_counter(call);
		encode_ax25_address(p, call);
		p += AXALEN;
	}
	return (p);
}

static byte *
encode_ack_string(byte *p, byte *arecip, byte *id, word idlen)
{
	*p++ = ':';

	decode_ax25_address(p, arecip);
	strcat(p, "         ");

	p = strcpyx(p + 9, ":ack");

	while (idlen-- && isalnum(*id))
		*p++ = *id++;

	return (p);
}

static void
message_ack(byte *opacket, byte *id, word idlen)
{
	struct txq *q = txq_alloc(TXPACLEN);
	byte *p;

	if (!q)
		return;

	p = q->data;

	encode_ax25_address(p, cfg.aprs_dst[0] ? cfg.aprs_dst : aprs_destination);
	p += AXALEN;

	encode_ax25_address(p, cfg.mycall);
	p += AXALEN;

	p = ack_path_reversal(p, opacket + 2 * AXALEN);
	p[-1] |= 1;                                      /* Last */

	*p++ = 0x03;
	*p++ = 0xF0;

	p = encode_ack_string(p, opacket + AXALEN, id, idlen);

	q->len = p - q->data;
	txq_enter(q);
}

static byte *
bull_dressing(byte *neat, byte *s, int n)
{
	/* copy so many bytes */

	bcopy(s, neat, n);

	/* check last and strip it until not blank */

	do {
		neat[n] = 0;
	} while (--n >= 0 && neat[n] == ' ');

	/* remove -0 at end */

	if (n > 0 && neat[n] == '0' && neat[--n] == '-')
		neat[n] = 0;

	/* squeeze blanks from the beginning too */

	while (*neat == ' ')
		++neat;

	strupper(neat);

	return (neat);
}

static bool
bull_match(byte *bird, byte *pats)
{
	while (*bird == toupper(*pats) && *bird)
		++bird, ++pats;

	return (*bird == 0 && (*pats == 0 || *pats == ' ' || *pats == COMMA));
}

static bool
match_bullgruppa(byte *g)
{
	byte *list = cfg.bullgroups;

	do {
		while (*list == ' ' || *list == COMMA)
			++list;

		if (bull_match(g, list))
			return (TRUE);

		while (*list && *list != ' ' && *list != COMMA)
			++list;
	} while (*list);

	return (FALSE);
}

static bool
match_message_recipient(byte *recip, byte *pat)
{
	while (*recip == toupper(*pat) && *recip)
		++recip, ++pat;

	if (*recip == 0)
		return (*pat == 0 || !strcmp(pat, "-0"));

	return (FALSE);
}

/*
 *  packet/len is the total packet
 *  p is the payload
 *  XXX this _does_ return TRUE for _any_ message/bulletin
 *
 *  :QST      :This is a broadcast message. can contain id
 *  :BLN9     :This is a general bulletin
 *  :BLNZ     :This is an announcement
 *  :BLN9NAME :This is a bulletin for group NAME
 *  012345678901
 */
static bool
maybe_messaging(byte *packet, byte *p, word len, bool not_altnet)
{
	byte *end = &packet[len];
	byte *recip, *id;
	byte buf[9 + 1];

	if (p + 11 > end || p[0] != ':' || p[10] != ':')
		return (FALSE);

	recip = bull_dressing(buf, p + 1, 9);
	p += 11;

	if (match_message_recipient(recip, cfg.mycall)
	 || match_message_recipient(recip, cfg.myalias)
	 || match_message_recipient(recip, cfg.mydomain)) {
		id = end;
		while (--id >= p)
			if (*id == '{') {            /* balance '}' */
				message_ack(packet, id + 1, end - id - 1);
				end = id;
				break;
			}
		store_news(packet + AXALEN, NULL, p, end - p);

	} else if (not_altnet) {

		if (!strcmp(recip, "QST")
		 || !strcmp(recip, "ALL")
		 || !strcmp(recip, "CQ")) {
			id = end;
			while (--id >= p)
				if (*id == '{') {        /* balance '}' */
					end = id;
					break;
				}
		} else if (!strncmp(recip, "BLN", 3) && isalnum(recip[3])) {

			recip += 4;                 /* XXX ignore sequence */

			if (cfg.bullgroups[0] && *recip && !match_bullgruppa(recip))
				recip = NULL;
		} else
			recip = NULL;

		if (recip)
			store_news(packet + AXALEN, recip, p, end - p);
	}
	return (TRUE);
}
 
static void
deg_dot_rees_to_mm(byte *s, long *pmm)
{
	bool neg = FALSE;
	word deg, weight, mm = 0;

	if (*s == ' ' || *s == '+' || *s == '-')
		neg = (*s++ == '-');

	deg = atoi(s);

	while (isdigit(*s))
		++s;

	if (*s == DOT) {
		weight = 6000;         /* 0.1 degrees */

		while (isdigit(*++s) && weight) {
			mm += (*s - '0') * weight;
			weight /= 10;
		}
	}
	*pmm = EMUL(deg, _DEG_AS_mm) + mm;
	if (neg)
		*pmm = lnegate(*pmm);
}

static word
miles_to_km(word miles)
{
	return (muldiv(miles, 103, 64));
}

/*
 *  ?APRS?
 *  ?APRS? 34.02,-117.15,0200
 *   radius 200 miles, fixed 4 digits
 *   34.02 degrees north, 117.15 degrees west
 *   ' ' = north/east, '-' = south/west
 *   floating point degrees, not minutes
 *
 *  respond by sending a report (and a "status". wazzt ?)
 */

static bool
maybe_aprs_query(byte *packet, byte *p, word len)
{
	byte *plat, *plon, *end = &packet[len];
	long lat, lon;
	word radius;

	if (p + 6 > end || bcmp(p, "?APRS?", 6) != 0)
		return (FALSE);

	p += 6;

	/* query to a limited area ? look for two commas */

	plat = p;
	while (p < end)
		if (*p++ == COMMA)
			break;
	plon = p;
	while (p < end)
		if (*p++ == COMMA)
			break;

	if (p + 4 <= end && is_digits(p, 4)) {

		/* ... looks like so */

		radius = ascii_num(p, 4);   /* miles, always 4 digits */
		if (radius == 0)
			radius = 1;             /* cannot be that precise */

		deg_dot_rees_to_mm(plat, &lat);
		deg_dot_rees_to_mm(plon, &lon);

		if ( ! inside_area(lat, lon, miles_to_km(radius)))
			goto not_for_us;
	}
	sendnow = TRUE;

not_for_us:
	return (TRUE);
}

/*
 *  1: tracker might digipeat
 *  2: coordinates might be sent to GPS
 * XXX
 * XXX packet[] trampled on.
 */
static void
tracker_input_packet(byte *packet, int len)
{
	byte *data = packet_payload(packet, len); /* if any */

	/* dump it for voyerism */

	if (cfg.waypoints == FALSE)
		prettyprint_packet(packet, len);

	if (cfg.lcd)
		lcd_traffic(packet, len);

	/* we are only interested in non-empty "UI Text" packets */

	if (data == NULL
	 || data[-2] != 0x03
	 || data[-1] != 0xF0)
		return;

	/*
	 *  If the last report has not yet been heard digipeated,
	 *  check if this one is a digipeat of that.
	 */
	if (last_report_seen_repeated == FALSE
	 && is_myaddress(packet + AXALEN, 0)
	 && crc_ax25(data, &packet[len] - data) == last_report_data_crc)
		last_report_seen_repeated = TRUE;


	report_inside_digi_area = EOF; /* "dunno". XXX disgusting in many ways */

	if (cfg.altnet[0] && is_other_net(packet)) {

		/* ignore other networks, if we have ours */

		maybe_messaging(packet, data, len, FALSE); /* my only */
	} else {
		/*
		 *  Could it be a position report ?
		 *  XXX maybe_xxx should return TRUE if any decision was made,
		 *  XXX _not_ if report was rejected for some sundry reason.
		 */
		maybe_ascii_waypoint(packet, data, len)  ||
		maybe_mic_e_waypoint(packet, data, len)  ||
		maybe_messaging(packet, data, len, TRUE) ||
		maybe_aprs_query(packet, data, len);
	}

	if (tx_is_locked_out())
		return;              /* rest requires permission to transmit */

	/* XXX safe to touch packet data now, prettyprinted and all, above */

	switch (cfg.tracker_digi) {

	case 1:
		if (report_inside_digi_area == TRUE) /* if not dunno nor FALSE */
			do_relay(packet, len);
		break;

	case 2:
		if (report_inside_digi_area != FALSE) /* if dunno or TRUE */
			aprs_digipeat_if_needed(packet, len);
		break;
	}
}

static short
course_change(word new, word old)
{
	short turn = new - old;

	if (turn > +180)
		turn -= 360;
	if (turn < -180)
		turn += 360;

	return (turn);
}

/*
 *  "unfixed GPS is stationary", or,
 *  this is done only if new fix data is available.
 *  Normally, once per second, or onece per every 2 seconds.
 *  If no fix, just the track_interval runs.
 */
static void
integrate_distance_and_catch_corners()
{
	static stamp prev_fix_at;
	static word  prev_course, prev_knots;

	word msec = gps_fix_at - prev_fix_at;
	word meters_per_sec = 0;

	if (gps.knots)
		meters_per_sec = knots_to_ms(gps.knots);

	/* more miles if moving */

	if (meters_per_sec) {
		word meters = muldiv(meters_per_sec, msec, 1000);

		meters_travelled      += meters;
		long_meters_travelled += meters;
	}

	/* CornerPegging if enabled and moving now and then */

	if (cfg.track_turnslope && gps.knots && reported_knots) {

		short turn = course_change(gps.course, reported_course);
		word  threshold;
		word  kmh  = knots_to_kmh(gps.knots);

		if (turn && kmh) {
			threshold = cfg.track_turnslope / kmh;

			if (abs(turn) > threshold + 1)
				sendnow = TRUE;
		}
	}

	/* GornerPegging if enabled */

	if (cfg.track_highg) {
		/*
		 *  constant length cumulator would be this:
			oacc = acc;
			qi++;
			qi %= 8;
			acc -= q[qi];
			acc += q[qi] = g;

		 *  only a coarse _approximation_ of sentripetal force.
		 *  over 90 degrees sharp turns are not handled correctly.
		 *  decim/s^2 scale is more practical to use.
		 *  if previous data is more than couple of seconds old,
		 *  assume g as 0.
		 *  Also, g is forced to 0 if previous or current speed is 0.
		 */
		static bool turning, fill;
		static short sum, qi, q[16];

		short turn, g;

		g = turn = 0;

		if (prev_knots && meters_per_sec && msec < 2500) {
			turn = course_change(gps.course, prev_course);

			if (turn) {
				g = meters_per_sec;
				g *= eight_bit_sin_from_abs_degrees(abs(turn));
				g *= 4;         /* *= 1000 for msec, /= 256 for sine scale */
				g /= msec / 10; /* *= 10 for decimeter scale */
				if (turn < 0)
					g = -g;     /* left or right sign */
			}
		}
		if (debug) {
			print_signed_short(turn);
			print_string(" ");
			print_short(meters_per_sec);
			print_string(" m/s, g");
		}

		/* if gps does not update every second, fake it */

		for (;;) {
			qi++;
			qi %= __SIZEOF(q);
			q[qi] = g;
			if (msec < 1500)
				break;
			msec -= 1000;
		}
		/*
		 *  length of sum(g) is proportional to speed;
		 *  300m turns at highway, 10m turns at downtown.
		 *
		 *  108 km/h  30 m/s  300m   10 second
		 *   32 km/h   9 m/s   10m    3 second
		 *    0 km/h   0 m/s    0m    0 second, but 1 second still used.
		 *
		 *  first in sum is always current g.
		 */
		if (debug) {
			print_string(" ");
			print_signed_short(g);
		}

		sum = g;

		if (meters_per_sec >= 3) {
			word n = meters_per_sec / 3;
			word i = qi;

			if (n > __SIZEOF(q))
				n = __SIZEOF(q);
			do {
				i--;
				i %= __SIZEOF(q);
				sum += q[i];

				if (debug) {
					print_string(" ");
					print_signed_short(q[i]);
				}
			} while (--n);
		}
		if (debug) {
			print_string(", sum=");
			print_signed_short(sum);
			eol();
		}
		/* catch the end of turn */

		if (abs(sum) > cfg.track_highg)
			turning = TRUE;

		if (abs(sum) <= cfg.track_lowg) {
			if (turning)
				sendnow = TRUE;
			turning = FALSE;
		}
	}

	/* remember what happened second ago */

	prev_fix_at = gps_fix_at;
	prev_course = gps.course;
	prev_knots  = gps.knots;
}

static bool
report_is_due()
{
	if (tx_is_locked_out())
		return (FALSE);

	if (gps_data_is_not_fit_for_report())
		return (FALSE);

	/*
	 *  too little time since last report ? NO
	 *  round a corner or something else like that ? YES
	 *  full interval elapsed ? YES
	 *  travelled so many kilometers ? YES
	 *  else NO, not yet.
	 */
	if (cfg.track_limiter) {

		if (!time_past(last_report_at + sec_to_msec(cfg.track_limiter)))
			return (FALSE);
	}
	if (sendnow)
		return (TRUE);

	if (cfg.track_interval) {

		if (time_past(next_report_at))
			return (TRUE);
	}
	if (cfg.track_distance) {

		if (meters_travelled > cfg.track_distance)
			return (TRUE);
	}

	/*
	 *  XXX
	 *  arrival or departure
	 *  if:
	 *   reported_knots >> "walking speed"
	 *   and gps.knots == 0
	 *   or the other way around.
	 *      TRUE
	 *
	 *  change of highway
	 *  if:
	 *   reported vector mul current vector > some limit
	 *      TRUE
	 */
	return (FALSE);
}

static void
check_tracker_input()
{
	struct rxq *pk;

	if (pk = getpacket()) {
		tracker_input_packet(pk->data, pk->len); /* pk->data trampled on */
		rxq_relse(pk);
	}
}

/* XXX dammiy */

static void
tracker_preset_times()
{
	stamp t;

	now(&t);

	gps_fix_at     = t; /* XXX */
	next_report_at = t; /* immediately */
	last_report_at = t - sec_to_msec(cfg.track_limiter);
	long_report_at = t - sec_to_msec(cfg.long_time);
}

static void
kicked()
{
	sendnow = FALSE;
	meters_travelled = 0;
	future_sec(&next_report_at, cfg.track_interval);
}

/*
 *  LED_2 is
 *  - lit permanently, when coordinates are fake
 *  - lit, but pulses after each Active GPRMC.
 *  - short blinks after each Void GPRMC
 *  - dim if no GPRMC received at all
 */
static bool
gps_validity_led()
{
	bool blink;

	if (gps_validity == PINNED)
		return (TRUE);

	if (gps_validity == FIX)
		return (time_past(gps_fix_at + 50u));

	blink = gps_has_spoken;
	gps_has_spoken = FALSE;
	return (blink);
}

static const byte *
symbolic_course(word deg)
{
	static const byte *x[] = {
		"N", "NE", "E", "SE", "S", "SW", "W", "NW", "N" };

	return (x[(deg + 22) / 45]);
}

static void
locator(byte *p, dword *LL)
{
	word N, n, c;

	N = EDIV(LL[0] + _90_DEG_AS_mm, 2500, &n);

	c = EDIV(N, 240, &N);
	p[1] = 'A' + c;
	c = EDIV(N, 24, &N);
	p[3] = '0' + c;
	p[5] = 'A' + N;

	c = EDIV(n, 250, &n);
	p[7] = '0' + c;
	p[9] = 'A' + n / 11;

	N = EDIV(LL[1] + _180_DEG_AS_mm, 5000, &n);

	c = EDIV(N, 240, &N);
	p[0] = 'A' + c;
	c = EDIV(N, 24, &N);
	p[2] = '0' + c;
	p[4] = 'A' + N;

	c = EDIV(n, 500, &n);
	p[6] = '0' + c;
	p[8] = 'A' + n / 21;
}

static void
my_locator(byte *p)
{
	dword LL[2];

	LL[0] = gps.lat;
	LL[1] = gps.lon;

	locator(p, LL);
}

#define LCD_CG_MARK  5
#define LCD_CG_RSSI  6

static void
lcd_gps_void()
{
	lcd_gps_good = FALSE;

	lcd_row(1);               /* LCD1: gps data */

	if (target_active_manual)
		lcd_chr(LCD_CG_MARK);

	lcd_finis();
}

static void
lcd_gps_active()
{
	byte loc[10];

	lcd_gps_good = TRUE;

	my_locator(loc);

	lcd_row(1);      /* LCD1: gps data */
	doing_lcd++;

	if (gps.knots) {
		lcd_buf(loc, 6);
		lcd_chr(' ');
		print_short(knots_to_kmh(gps.knots));
		lcd_str(" kmh ");
		lcd_str(symbolic_course(gps.course));
	} else {
		lcd_buf(loc, 10);
		lcd_chr(' ');
		print_signed_long(&gps_altitude);
		lcd_str(" m");
	}
	if (target_active_manual)
		lcd_chr(LCD_CG_MARK);

	doing_lcd--;
	lcd_finis();

	/* XXX quite expensive to do once per second... */

	if (wpt_tracked)
		lcd_report(wpt_tracked);
	else
	if (1 && wpt_stash->tag[0])
		lcd_report(wpt_stash);
}

static void
process_gps_input()
{
	word c;

	while ((c = SCI_getc()) != EOF)
		if (cfg.gps)
			quirky_gps_input(c);
		else
			nmea_gps_input(c);
}

/*
 *  Personality 1
 */
static void
aprs_tracker()
{
	word leds;

	if (gps_validity == PINNED)
		lcd_gps_active();

	if (!any_idea_of_location)
		lcd_message(1, "Location unknown");

	gps_initialize();

	tracker_preset_times();

	for (;;) {
		/*
		 *  "smart" variables are reset here, however a
		 *  report is sent some time later, depending on
		 *  channel usage. No biggie ?
		 */
		if (gps_fix_available) {
			gps_fix_available = FALSE;
			integrate_distance_and_catch_corners();

			if (cfg.lcd)
				lcd_gps_active();
		}
		if (tx_state == OFF && (ptt_pressed() || report_is_due())) {

			if (cfg.tracker_digi != 2) {
				report_queuing_postponed = TRUE; /* until keydown */
				kickit();
			} else
				queue_aprs_report();   /* synchronously */

			kicked();
		}
		check_tracker_input();

		process_gps_input();
		gps_inactive_timeout();

		if (cfg.lcd && lcd_gps_good && gps_data_is_void())
			lcd_gps_void();

		if (cfg.manipulator)
			do_manipulator();

		if (keycnt != keycnt_previously)
			process_keys();

#if 0
	chatboard_debug();
#endif

		/* XXX should do this also when holdq grows large ? */

		if (holdq && time_past(holdq_mark))
			send_held_packets();

		leds = 0;
		if (tx_state <= TRYING && carrier(1))
			leds |= LED_3;
		if (gps_validity_led())
			leds |= LED_2;
		if (last_report_seen_repeated)
			leds |= LED_1;

		green_leds(leds);

		if (news && cfg.lcd)
			billboard();

		ticking();
	}
}

/* kind of synchronously. tx must be on already XXX */

static void
send_ptt_packet()
{
	struct repdata rd;

	struct txq *p = txq_alloc(TXPACLEN); /* XXX */

	if (!p)
		return;

	switch (cfg.ptt_packet) {
	default:
		txq_relse(p);
		return;

	case 1: /* shortest id */

		encode_ax25_address(p->data +      0, "ID");
		encode_ax25_address(p->data + AXALEN, cfg.mycall);

		p->data[AXMINLEN - 2] |= 1;    /* last address field in packet */
		p->data[AXMINLEN - 1] = 0x03;  /* control is UI */

		p->len = AXMINLEN;
		break;

	case 2: /* MIC-E */

		my_report_data(&rd); /* only the fish which are used by MIC-E */

		p->len = encode_aprs_packet_mic_e(p->data, &rd);

		add_report_comments(p, &rd);
		break;
	}
	afsk_tx_idle();
	delay_ms(cfg.data_delay);

	afsk_tx_go(p);

	while (afsk_tx_going())
		;

	afsk_rx_on();
}

static void
voice_transmission(int beep)
{
	if (tx_is_locked_out())
		doh("tx_lockout2");    /* should not get this far in that case */

	tx_led(TRUE);

	radio_control(RX_ON | RFCMD_MUTE);

	vco_tx(cfg.tx_freq ? cfg.tx_freq : cfg.freq);
	delay_ms(cfg.pll_delay);

	radio_control(TX_ON | RFCMD_MUTE);  /* RX off */

	if (cfg.faith_in_pll || synth_locked())
		tx_power(cfg.tx_power);
	else
		green_leds(0);

	delay_ms(cfg.pwr_delay);

	if (beep)
		repeater_beep();

	if (cfg.ctcss)
		ctcss(cfg.ctcss);

	while (ptt_pressed())     /* busywait the ptt */
		;

	if (cfg.ctcss)
		ctcss(0);

	if (cfg.ptt_packet && cfg.mycall[0])
		send_ptt_packet();

	tx_power(POWER_LOWEST);
	delay_ms(2);
	radio_control(RX_ON | RFCMD_MUTE);  /* TX off */

	vco_rx(cfg.freq);

	tx_led(FALSE);

	/* at this point, must be rx and squelch closed */
}

static bool
need_autobeep()
{
	return (cfg.tx_freq != 0 && cfg.tx_freq != cfg.freq && cfg.autobeep);
}

static void
power_leds()
{
	switch (cfg.tx_power) {
	default:
		green_leds(0);
		break;
	case 2:
		green_leds(LED_3);
		break;
	case 1:
		green_leds(LED_3 | LED_2);
		break;
	case 0:
		green_leds(LED_3 | LED_2 | LED_1);
		break;
	}
}

static void
rssi_leds()
{
	word rssi = read_rssi();

	if (rssi < cfg.rssi_1led)
		green_leds(0);
	else
	if (rssi < cfg.rssi_2led)
		green_leds(LED_3);
	else
	if (rssi < cfg.rssi_3led)
		green_leds(LED_3 | LED_2);
	else
		green_leds(LED_3 | LED_2 | LED_1);

	if (cfg.lcd)
		lcd_rssi(rssi);
}

STATIC struct {
	int   muted;
	stamp timer;
	stamp at_bilevel;
} squelch;

static void
close_squelch()
{
	DCD2_pin(1);

	radio_control(RX_ON | RFCMD_MUTE);

	squelch.muted = TRUE;
	squelch.timer = 0;

	squelch.at_bilevel = 0;
}

static void
open_squelch()
{
	DCD2_pin(0);

	radio_control(RX_ON);            /* UN_MUTE */

	squelch.muted = FALSE;
	squelch.timer = 0;
}

static int
squelch_threshold()
{
	return (cfg.use_busy ? busy_channel() : (read_rssi() > cfg.squelch));
}

static void
squelch_update()
{
	if (squelch.muted) {
		squelch.at_bilevel = 0;

		if (squelch_threshold()) {
			if (squelch.timer == 0)
				future_msec(&squelch.timer, cfg.squelch_delay);

			if (time_past(squelch.timer))
				open_squelch();
		} else
			squelch.timer = 0;
	} else {
		if (read_rssi() > cfg.squelch_bilevel)
			future_msec(&squelch.at_bilevel, 500);

		if (squelch_threshold())
			squelch.timer = 0;
		else {
			if (squelch.timer == 0)
				future_msec(&squelch.timer, cfg.squelch_tail);

			if ((squelch.at_bilevel && !time_past(squelch.at_bilevel))
			 || time_past(squelch.timer))
				close_squelch();
		}
	}
}

static void
postpone_idle(stamp *when)
{
	future_sec(when, cfg.idle_time);
}

static void
prettyprint_packets()
{
	struct rxq *pk;

	if (pk = getpacket()) {
		prettyprint_packet(pk->data, pk->len);
		rxq_relse(pk);
	}
}

static void
lcd_freq_etc(dword *fq)
{
	lcd_row(3);      /* LCD3: freq and rssi etc */
	doing_lcd++;

	print_freq(*fq);
	lcd_chr(' ');

	lcd_chr(LCD_CG_RSSI);
	lcd_chr(' ');

	lcd_str(tab_tx_power[cfg.tx_power]);

	doing_lcd--;
	lcd_finis();
}

static void
lcd_scanning()
{
	lcd_row(3);      /* LCD3: freq and rssi etc */

	lcd_str("Scan");

	lcd_finis();
}

/*
 *  Personality 0
 *
 *  If this returns, scanner() is started.
 */
static void
simple_voice_radio()
{
	stamp idle;

	if (cfg.lcd)
		lcd_freq_etc(&cfg.freq);

	postpone_idle(&idle);

	for (;;) {
		prettyprint_packets();      /* for now */

		if (ptt_pressed()) {

			close_squelch();
			CTS2_pin(0);

			power_leds();
			voice_transmission(squelch.muted && need_autobeep()); /* beep ? */
			CTS2_pin(1);

			postpone_idle(&idle);

		}
		rssi_leds();
		squelch_update();

		if (!squelch.muted)
			postpone_idle(&idle);

		if (cfg.autoscan && time_past(idle)) /* nothing happens - scan */
			break;

		process_gps_input();

		if (keycnt != keycnt_previously)
			process_keys();
	}
}

/* XXX dumb digipeater is next to useless */

static void
dumb_digipeat_if_needed(byte *packet, int len)
{
	byte *digi;

	if (tx_is_locked_out())
		return;

	if (len < AXMINLEN + AXALEN)
		return;               /* cannot be digis in that */

	if (packet[AXALEN + oSSID] & 1)
		return;               /* no digis, source address was last */

	if (is_myaddress(packet, DO_ALIAS))
		return;               /* for us. would be silly to peat it */


	/* check digi addresses, find the first not repeated slot */

	digi = &packet[2 * AXALEN];

	while (digi + AXALEN <= &packet[len - 1]) {

		if ((digi[oSSID] & 0x80) == 0) { /* not repeated */

			if (is_myaddress(digi, DO_ALIAS)) {

				digi[oSSID] |= 0x80;     /* we are asked to repeat it */
				/*
				 * XXX replace alias(es) with mycall ?
				 * XXX downcount certain SSIDs ?
				 */

				/*
				 * XXX now scan list of recently heard packets,
				 * or digests of them,
				 * if 'packet' is there, drop it. already repeated
				 * by someone ?
				 */
				sendit(packet, len);
			}
			break;                 /* rest can be skipped */
		}
		if (digi[oSSID] & 1)       /* had passed all digis already */
			break;

		if (is_myaddress(digi, DO_ALIAS))
			break;                 /* _I_ have already repeated it */

		digi += AXALEN;
	}
}

/*
 *  Personality 3
 */
static void
digipeater()
{
	word leds;
	struct rxq *pk;

	for (;;) {

		if (pk = getpacket()) { 
			dumb_digipeat_if_needed(pk->data, pk->len);
			prettyprint_packet(pk->data, pk->len);
			rxq_relse(pk);
		}
		leds = 0;
		if (tx_state <= TRYING && carrier(1))
			leds |= LED_3;
		if (txq)
			leds |= LED_2;
		if (txq_maxed)
			leds |= LED_1;

		green_leds(leds);
	}
}

static void
next_channel(int *i)
{
	static int blink;
	int n = __SIZEOF(cfg.channel);

	do {
		if (++*i >= __SIZEOF(cfg.channel))
			*i = 0;
	} while (cfg.channel[*i].freq == 0 && --n);

	blink ^= LED_2;
	green_leds(blink);

	close_squelch();
	vco_rx(cfg.channel[*i].freq);
}

/* confusing mess */

static int
scanner_ptt()
{
	prettyprint_packets(); /* for now */

	process_gps_input();

	if (keycnt != keycnt_previously)
		process_keys();

	return (ptt_pressed());
}

/*
 *  Personality 4
 *
 *  If this returns, voice is selected.
 */
static void
scanner()
{
	stamp mark;
	int channel = -1; /* preincrements to 0 */
	struct channel last;

	bzero(&last, sizeof(last));
	last.freq    = cfg.freq;
	last.tx_freq = cfg.tx_freq;

	for (;;) {

		/* scanning phase */
next:
		if (cfg.lcd)
			lcd_scanning();

		do {
			next_channel(&channel);
			delay_ms(50);

			if (scanner_ptt())
				goto participate;

			squelch_update();
		} while (!squelch_threshold());

		/* possible signal, give time for squelch to open */

		future_msec(&mark, cfg.squelch_delay);
		do {
			if (time_past(mark))      /* nothing after all */
				goto next;

			if (scanner_ptt())
				goto participate;

			squelch_update();
		} while (squelch.muted);

		/* listening/lingering phase */

		last = cfg.channel[channel];

		if (cfg.lcd)
			lcd_freq_etc(&last.freq);

		for (;;) {
			if (scanner_ptt())
				goto participate;

			rssi_leds();
			squelch_update();

			if (squelch.muted) {
				if (time_past(mark))
					goto next;             /* looks like qso finished */
			} else
				future_sec(&mark, cfg.scan_tail); /* not */
		}
	}

participate:

	cfg.freq    = last.freq;         /* no perm */
	cfg.tx_freq = last.tx_freq;
}

/*
 *  AFSK output stuff
 */
static struct txq *
pick_some_packets_to_send()
{
	struct txq *list, **pp = &list; /* frames to send this time */
	word cc = 1;                    /* bytes to send this time. starting flag */

	while (txq) {
		cc += txq->len + 1;     /* plus one flag */
		*pp = txq;
		pp = &txq->next;
		txq = txq->next;        /* bit silly shuffle */

		if (cfg.txlimit && cc >= cfg.txlimit) /* max bytes per keydown */
			break;
	}
	*pp = NULL;

	return (list);
}

static word
	transmission_idle(),
	transmission_1(),
	transmission_2(),
	transmission_3(),
	transmission_4(),
	transmission_5();

word (*txproc)() = transmission_idle;

STATIC stamp slot, dcd_glitch;

		/*
		 *  Wait the channel to clear.
		 *  Synchronize time slots to end of any transmission,
		 *  including ours. Ignore short glitches.
		 *
		 *  Do not retake channel without at least one slot.
		 *  Retry channel takeover at end of each slottime.
		 *  Then,
		 *  Send something, usually everything. If
		 *  something remains, do it all again immediately
		 *  after giving some chance for others.
		 */
static void
opportunity()
{
	now(&slot);

	if (tx_state == OFF || persist())
		return;

	/* last call for (normal or relaying) tracker */

	if (cfg.personality == PERSONALITY_TRACKER
	 && report_queuing_postponed) {
		report_queuing_postponed = FALSE;
		queue_aprs_report();
	}

	if (txq)
		txproc = transmission_1;
	else
		tx_state = OFF;
}

static word
transmission_wait()
{
	if (carrier(0))      /* keep waiting */
		return (0);

	txproc = transmission_idle;

	if (time_since(&dcd_glitch) > 100)
		opportunity();

	return (0);
}

static word
transmission_idle()
{
	if (carrier(0)) {
		now(&dcd_glitch);
		txproc = transmission_wait;
		return (0);
	}
	if (time_since(&slot) > cfg.slottime)
		opportunity();

	return (0);
}

static word
transmission_1()
{
	tx_state = KEYING;               /* Keying now */

	afsk_tx_led(TRUE);

	vco_tx(cfg.tx_freq ? cfg.tx_freq : cfg.freq);

	txproc = transmission_2;
	return (cfg.pll_delay);          /* PLL locking time */
}

static word
transmission_2()
{
	if (tx_is_locked_out())
		doh("tx_lockout");       /* should not get this far in that case */

	radio_control(TX_ON);            /* RX off */

	if (cfg.faith_in_pll || synth_locked())
		tx_power(cfg.tx_power);
	else
		green_leds(0);               /* XXX no place to gripe to ? */

	afsk_tx_idle();

	txproc = transmission_3;
	return (cfg.pwr_delay + cfg.data_delay);  /* Wait TX power to rise */
}

static word
transmission_3()
{
	tx_state = ACTIVE;

	if (txq)
		afsk_tx_go(pick_some_packets_to_send());

	txproc = transmission_4;
	return (0);
}

static word
transmission_4()
{
	if (afsk_tx_going())
		return (0);

	tx_state = UNKEYING;

	tx_power(POWER_LOWEST);

	txproc = transmission_5;
	return (2);                       /* Wait RF go down */
}

static word
transmission_5()
{
	radio_control(RX_ON);                 /* TX off */

	vco_rx(cfg.freq);

	afsk_rx_on();
	afsk_tx_led(FALSE);

	tx_state = txq ? TRYING : OFF;

	now(&slot);
	txproc = transmission_idle;
	return (2);
}

/*
 * and now something completely different
 */

static void
print_freq(dword f)
{
	byte buf[12]; /* must hold 4G and \0, 10+1 bytes */
	int i;

	sprint_010Lu(buf, &f);

	for (i = 0; buf[i] == '0' && i < 6; ++i) /* zeroblank to 1 kHz */
		;

	/*
	 * add dots to resemble 1.295.500
	 *                  0 123 456 789
	 */

	do {
		putchar(buf[i]);
		if (i == 0 || i == 3 || i == 6)
			putchar(DOT);
	} while (++i < 10);
}

/* XXX handle 145 = 145.000 and 145.0 = 145.000 */

static dword
atofreq(byte *p)
{
	dword k = 0;

	for ( ; *p; ++p) {
		if (*p == DOT)
			continue;
		if (*p < '0' || *p > '9')
			return (0);

		k = EMULx(k, 10) + (*p - '0');
	}
	return (k);
}

/*
 * pick first blank-separated word from *pp
 * advance *pp to start of second b.s. word
 */
static byte *
blanksplit(byte **pp)
{
	byte *p = *pp;
	byte *p0;

	while (*p == ' ')
		++p;

	for (p0 = p; *p; ++p) {
		if (*p == ' ') {
			*p++ = 0;
			while (*p == ' ')
				++p;
			break;
		}
	}
	*pp = p;

	return (p0);
}

static void
show_possible_options(byte **opts)
{
	int i;
	char *pretty = ": ";

	print_string("possible options");

	for (i = 0; opts[i]; ++i) {
		if (i && opts[i + 1] == NULL)
			pretty = " and ";
		print_string(pretty);
		pretty = ", ";

		print_string(opts[i]);
	}
	eol();
}

static void
channel_printout(int i, struct channel *p)
{
	print_short_N(i, 2);
	print_string(": ");

	if (p->freq == 0) {
		print_string("empty channel");
	} else {
		print_freq(p->freq);
		print_string(" ");

		if (p->tx_freq == 0)
			print_string("simplex");
		else {
			print_freq(p->tx_freq);
			print_string(" ");

			if (p->tx_freq < p->freq) {
				print_freq(p->freq - p->tx_freq);
				print_string("-");
			} else {
				print_freq(p->tx_freq - p->freq);
				print_string("+");
			}
		}
	}
	print_string("\n\r");
}

#if 0

static dword
getset_foobar(short is_set, struct cfgparm *p)
{
	if (is_set)
		;
	return (42);
}

#endif

static void
cfg_printout(struct cfgparm *p)
{
	union {
		void *v;
		byte *b;
		word *w;
		long *l;
	} xx;
	int i;

	xx.v = p->address;

	print_string(p->name);
	i = strlen(p->name);
	do {
		print_string(" ");
	} while (++i < 16);

#if 0
	if (p->getset)
		call_near_pointer(p->getset, FALSE, p); /* XXX */
#endif
	switch (p->type) {
	default:
		print_string("\"");
		print_buf(xx.b, p->size);
		print_string("\", type unknown");
		break;

	case CFG_STR:
		print_string(xx.b);
		break;

	case CFG_BOOL:
		print_string(*xx.b ? "on" : "off");
		if (p->size != 1)
			doh("bool size");
		break;

	case CFG_BYTE:
		print_short(*xx.b);
		break;

	case CFG_WORD:
		print_short(*xx.w);
		break;

	case CFG_FREQ:
	case CFG_LONG:            /* XXX */
		print_freq(*xx.l);
		break;

	case CFG_TAB:
		for (i = 0; p->optlist[i]; ++i) /* catch short list */
			if (i == *xx.b)
				break;

		if (p->optlist[i] == NULL) {
			print_short(*xx.b);
			print_string(", name unknown");
		} else
			print_string(p->optlist[i]);

		if (p->size != 1)
			doh("tab size");
		break;

	case CFG_APRSYM:
		if (is_valid_symbol(*xx.w >> 8, *xx.w))
			show_aprs_symbol(*xx.w);
		else
			print_string("INVALID");
		break;
	}
	eol();
}

static bool
cfg_assign(struct cfgparm *p, byte *arg)
{
	union {
		void *v;
		byte *b;
		word *w;
		long *l;
	} xx;
	int i;

	xx.v = p->address;

	switch (p->type) {
	default:
		if (arg) {
			print_string(p->name);
			print_string(" has unknown type, unassignable\n\r");
		}
		return (FALSE);

	case CFG_STR:
		if (arg)
			strncpyx(xx.b, arg, p->size);
		else
			bzero(xx.b, p->size);
		break;

	case CFG_BOOL:
		if (arg) {
			if (!strcmp(arg, "0") || !strcasecmp(arg, "off"))
				i = 0;
			else
			if (!strcmp(arg, "1") || !strcasecmp(arg, "on"))
				i = 1;
			else {
				print_string("possible options: off/on\n\r");
				return (FALSE);
			}
		} else
			i = p->defval;
		*xx.b = i;
		break;

	case CFG_BYTE:
		*xx.b = arg ? atoi(arg) : p->defval;
		break;

	case CFG_WORD:
		*xx.w = arg ? atoi(arg) : p->defval;
		break;

	case CFG_FREQ:
	case CFG_LONG:
		*xx.l = arg ? atofreq(arg) : p->defval;
		break;

	case CFG_TAB:
		if (arg) {
			for (i = 0; p->optlist[i]; ++i)
				if (!strcasecmp(p->optlist[i], arg))
					break;

			if (p->optlist[i] == NULL) {
				show_possible_options(p->optlist);
				return (FALSE);
			}
		} else
			i = p->defval;
		*xx.b = i;
		break;

	case CFG_APRSYM:
		if (arg) {
			word sym;

			if (is_valid_symbol(arg[0], arg[1]))
				sym = (arg[0] << 8) | arg[1];
			else
				sym = lookup_aprs_symbol(arg);

			if (!sym) {
				print_string("not a valid symbol, try help symbols");
				eol();
				return (FALSE);
			}
			*xx.w = sym;
		} else
			*xx.w = p->defval;
		break;
	}
#if 0
	if (p->getset)
		call_near_pointer(p->getset, TRUE, p);
#endif
	return (TRUE);
}

static void
stat_line()
{
	byte c;

	print_string("Busy ");    print_string(busy_channel() ? "1" : "0");
	print_string(" RSSI ");   print_short_N(read_rssi(), 5);
	print_string(" DTR2 ");   print_short( DTR2_pin() == 0);
	print_string(" MISO ");   print_short(read_PORTQS() & 1 ^ 1);
	print_string(" SCK ");    print_short((read_PORTQS() & 4) == 0);
	print_string(" Inputs "); print_word(read_inputs());
	print_string(" Batt ");   print_short_N(read_AN7_batt(cfg.battscale), 3);

	print_string(synth_locked() ? " Locked  " : " Unlocked");
}

extern byte _eeprom[];

static void
eeprom_erase()
{
	byte *e = _eeprom + profile_offset;

	flash_erac(e);
}

static void
eeprom_write(void *ram, word cc)
{
	byte *e = _eeprom + profile_offset;
	byte *p = ram;

	do {
		flash_writc(e++, *p++);
	} while (--cc);
}

static bool
permanent()
{
	static struct cfg cfg_sibling_profile; /* 2 profiles per flash erase */

	cfg.config_size = sizeof(cfg);
	cfg.cfg_magic_1 = CFG_MAGIC_1;
	cfg.cfg_magic_2 = CFG_MAGIC_2;

	/*
	 *  load the sibling profile into ram.
	 *  erase will zap 16 kB at once.
	 */
	profile_offset ^= PROFILE_OFFSET(1); /* XXX tsk */

	eeprom_read(&cfg_sibling_profile, sizeof(cfg));

	profile_offset ^= PROFILE_OFFSET(1); /* XXX tsk */


	print_string("erase ");
	eeprom_erase();

	/* save the sibling profile back, if it is valid */

	if (cfg_sibling_profile.cfg_magic_1 == CFG_MAGIC_1) {

		profile_offset ^= PROFILE_OFFSET(1); /* XXX tsk */

		eeprom_write(&cfg_sibling_profile, sizeof(cfg));

		if (eeprom_compare(&cfg_sibling_profile, sizeof(cfg)) != 0)
			print_string("- sibling? - ");

		profile_offset ^= PROFILE_OFFSET(1); /* XXX tsk */
	}

	/* back with this one */

	print_string("write ");
	eeprom_write(&cfg, sizeof(cfg));  /* en force */

	print_string("compare ");
	if (eeprom_compare(&cfg, sizeof(cfg)) != 0)
		goto errout;

	print_string("OK\n\r");

	return (TRUE);

errout:
	print_string(" failed\n\r");
	return (FALSE);
}

static struct cfgparm *
parm_lookup(char *name)
{
	int i;

	for (i = 0; i < __SIZEOF(cfg_general); ++i)

		if (!strcasecmp(cfg_general[i].name, name))
			return (&cfg_general[i]);

	return (NULL);
}

bool
more(int *lines)
{
	byte ans;

	if (interactive
	 && cfg.ttylines
	 && ++*lines >= cfg.ttylines - 1) {

		print_string("-More-");
		ans = getchar();
		print_string("\r       \r");

		switch (ans) {
		case 0x00:
		case 0x03:
		case 'q':
			return (FALSE);

		default:
			*lines = 0;

		case '\r':
		case '\n':
			break;
		}
	}
	return (TRUE);
}

static void
process_remote_command(byte *cmd)
{
	doing_remote++;
	collected_output_len = 0;

	process_manual_command(cmd); /* hmph */

	doing_remote--;
}

struct remote_cfg_packet {
	byte dst[AXALEN];
	byte src[AXALEN];
	byte ctl;
	byte pid;
	byte zero;
	byte seq;
	byte data[80];
};

static void
send_remote_response(byte *addr, byte seq)
{
	static struct remote_cfg_packet pk;
	unsigned cc;

	bcopy(addr, pk.dst, AXALEN);
	pk.dst[oSSID] &= ~1;

	encode_ax25_address(pk.src, cfg.mycall);
	pk.src[oSSID] |= 1; /* last address field in packet */

	pk.ctl  = 0x03;
	pk.pid  = 0xEC;
	pk.zero = 1;
	pk.seq  = seq;

	cc = collected_output_len;
	if (cc > sizeof(pk.data))
		cc = sizeof(pk.data);

	bcopy(collected_output, pk.data, cc);

	sendit(&pk, &pk.data[cc] - (byte *) &pk);
}

static bool
remote_cfg_packet_maybe(byte *packet, int len)
{
	struct remote_cfg_packet *pk = (void *) packet;

	if (!doing_remote                /* can it happen ? */
	 && cfg.mycall[0]                /* allowed locally */
	 && cfg.rfpasswd[0]              /* ... these two must be set */
	 && packet + len > pk->data      /* at least header exists */
	 && pk->ctl  == 0x03             /* proper header */
	 && pk->pid  == 0xEC
	 && pk->zero == 0
	 && packet[len - 1] == 0         /* ends in \0 - lazy ... */
	 && !strcmp(pk->data, cfg.rfpasswd)
	 && is_myaddress(pk->dst, DO_ALIAS | DO_DOMAIN)) {

		byte *cmd = pk->data + strlen(pk->data) + 1; /* over password */

		if (cmd < packet + len) {
			process_remote_command(cmd);
			send_remote_response(pk->src, pk->seq);
		}
		return (TRUE);
	}
	return (FALSE);
}

static bool
remote_response(byte *remote_addr, byte seq)
{
	struct remote_cfg_packet *cp;
	int len;
	struct rxq *pk;
	int result = FALSE;

	if (pk = getpacket()) {
		cp = (void *) pk->data;
		len = pk->len;

		if (pk->data + len > (byte *) &cp->seq
		 && cp->ctl  == 0x03
		 && cp->pid  == 0xEC
		 && cp->zero == 1
		 && cp->seq  == seq
		 && is_myaddress(cp->dst, 0)) /* just mycall will do */
			result = TRUE;

		if (result) {
			int i, n = &pk->data[pk->len] - cp->data;

			/* if the sender is unexpected (mass config ?)
			 * print his callsign */

			if (!addr_equals(cp->src, remote_addr)) {
				print_shifted_call(cp->src);
				print_string(":");
				eol();
			}
			/* and print the reply, having faith
			 * in printable characters. */

			for (i = 0; i < n; ++i)
				putchar(cp->data[i]);
			eol();
		}
		rxq_relse(pk);
	}
	return (result);
}

static void
remote_parameditor(byte *remote_addr)
{
	static struct remote_cfg_packet pk;

	byte *line, *cmd, *end, *p;
	int try;

	char addr[12], ps[16];

	if (strlen(remote_addr) >= sizeof("OHnABC-15")) {
		print_string("too long address, sorry\n\r");
		return;
	}
	strcpyx(addr, remote_addr);
	strcpyx(strcpyx(ps, addr), " $ ");

	encode_ax25_address(pk.dst, addr);
	encode_ax25_address(pk.src, cfg.mycall);

	pk.src[oSSID] |= 1; /* last address field in packet */

	pk.ctl = 0x03;              /* UI */
	pk.pid = 0xEC;              /* pid */

	pk.zero = 0;

	do {
		line = prompt("password: ");
		if (!line || !*line)
			return;
	} while (strlen(line) > 15);

	cmd = strcpyx(pk.data, line);
	cmd++;

	end = &pk.data[sizeof(pk.data)];

	/*
	 *  payload will be \0 token pass \0 cmd \0
	 *
	 *  cmd after the \0 ending pass string
	 *  end after last available byte
	 */
	print_string("enter commands, exit with ^D\n\r");

	while (line = prompt(ps)) {
		if (*line == 0)
			continue;

		if (strlen(line) + 1 > end - cmd) {
			print_string("too long command, sorry\n\r");
			continue;
		}
		p = strcpyx(cmd, line);
		p++;

		++pk.seq;
		try = 3;

		do {
			stamp timo;

			future_msec(&timo, 3 * 1000);

			sendit(&pk, p - (byte *) &pk); /* XXX plaintext */
			do {
				if (remote_response(addr, pk.seq) || SCI_canget())
					goto got;
			} while (!time_past(timo));
		} while (--try);

		print_string("no response\n\r");
	got:
		;
	}
}

static byte
disp_group_mask(byte c)
{
	if (c == 't')
		return (1 << PERSONALITY_TRACKER);
	if (c == 'd')
		return (1 << PERSONALITY_PEATER);
	if (c == 'k')
		return (1 << PERSONALITY_KISS);
	if (c == 'v')
		return (1 << PERSONALITY_VOICE);
	if (c == 's')
		return (1 << PERSONALITY_SCANNER);

	return (0); /* dunno what it is */
}

static void
show_stack(word *f)
{
	word n;

	print_word((word) f);
	print_string(" ");
	n = 0;
	while (*f++ == 0)
		n += 2;
	print_word((word) f);
	print_string(" ");
	print_short(n);
	eol();
}

/* XXX getting too long - funny cc does not complain yet */

static bool
process_manual_command(byte *arg)
{
	byte *cmd;
	int i;
	struct cfgparm *p;
	int lines = 0;

	cmd = blanksplit(&arg); /* first word and rest */

	if (!strcasecmp(cmd, "reboot"))
		reboot(0);

	if (!strcasecmp(cmd, "restart"))
		longjmp(restart, 1);

	if (!strcasecmp(cmd, "nop"))
		return (TRUE);

	if (!strcasecmp(cmd, "version")) {

		print_string(ident + 4);
		eol();
		return (TRUE);
	}
	if (!strcasecmp(cmd, "wpt_stash")) {

		publish_waypoints(FALSE);
		return (TRUE);
	}
	if (!strcmp(cmd, "debug")) {
		word a = atoi(blanksplit(&arg));

		debug = a;
		return (TRUE);
	}
#if 0
	if (!strcmp(cmd, "loadtest1")) {
		stamp start;
		word a = atoi(blanksplit(&arg));

		now(&start);

		while (a--)
			for (i = 0; ++i; )
				nullop();

		print_short(time_since(&start));
		print_string(" msec");
		eol();
		return (TRUE);
	}
#endif
#if 0
	if (!strcmp(cmd, "loadtest2")) {
		stamp start;
		word a = atoi(blanksplit(&arg));

		now(&start);

		for (i = 0; i < a; ++i)
			strcmp(ident + 2, ident);

		print_short(time_since(&start));
		print_string(" msec");
		eol();
		return (TRUE);
	}
#endif
	if (!strcasecmp(cmd, "batt")) {

		print_short(read_AN7_batt(cfg.battscale));
		eol();
		return (TRUE);
	}
	if (!strcmp(cmd, "peek")) {
		char *as = blanksplit(&arg);
		int n = atoi(blanksplit(&arg));
		static int a;

		if (*as)
			a = ascii_hex(as, strlen(as));
		if (n == 0)
			n = 1;

		do {
			if (!more(&lines))
				break;
			print_string("0");
			print_word(a);
			print_string(": ");
			print_word(read_ram(a));
			eol();
			a += 2;
		} while (--n);
		return (TRUE);
	}
	if (!strcmp(cmd, "eepeek")) {
		char *as = blanksplit(&arg);
		int n = atoi(blanksplit(&arg));
		static int a;

		if (*as)
			a = ascii_hex(as, strlen(as));
		if (n == 0)
			n = 1;

		do {
			if (!more(&lines))
				break;
			print_string("2");
			print_word(a);
			print_string(": ");
			print_word(read_eepage(a));
			eol();
			a += 2;
		} while (--n);
		return (TRUE);
	}
	if (!strcmp(cmd, "register")) {
		char *as = blanksplit(&arg);
		int n = atoi(blanksplit(&arg));
		static int a;

		if (*as)
			a = ascii_hex(as, strlen(as));
		if (n == 0)
			n = 1;

		do {
			if (!more(&lines))
				break;
			print_string("F");
			print_word(a);
			print_string(": ");
			print_word(read_register(a));
			eol();
			a += 2;
		} while (--n);

		return (TRUE);
	}
	if (!strcmp(cmd, "stacks")) {
		extern word _lostack[];

		show_stack(_lostack);
		return (TRUE);
	}
	if (!strcmp(cmd, "wallclock")) {

		print_date(&wallclock);
		eol();
		return (TRUE);
	}
	if (!strcasecmp(cmd, "profile")) {
		if (*arg) {
			i = atoi(blanksplit(&arg));

			if ((unsigned) i >= __SIZEOF(profile_tags)) {
				print_string("no such profile number\n\r");
				return (TRUE);
			}
			profile_offset = PROFILE_OFFSET(i);

			eeprom_read(&cfg, sizeof(cfg));
			cfg_sanitize();
		} else
			i = profile_offset / PROFILE_OFFSET(1); /* XXX tsk */

		print_short(i);

		if (cfg.profile_tag[0]) {
			print_string(" \"");
			print_string(cfg.profile_tag);
			print_string("\"");
		} else
			print_string(", with no name,");

		print_string(" is current profile");
		eol();
		return (TRUE);
	}
	if (!strcasecmp(cmd, "perm")) {

		permanent();
		return (TRUE);
	}
	if (!strcasecmp(cmd, "disp")) {
		byte which = ~0;

		if (*arg) {
			which = 0;
			do {
				which |= disp_group_mask(*arg);
			} while (*++arg);
		}
		for (i = 0; i < __SIZEOF(cfg_general); ++i)
			if (which & cfg_general[i].group) {

				if (!more(&lines))
					break;

				cfg_printout(&cfg_general[i]);
			}
		return (TRUE);
	}
	if (!strcasecmp(cmd, "factorydefaults")) {

		for (i = 0; i < __SIZEOF(cfg_general); ++i)
			cfg_assign(&cfg_general[i], NULL);

		print_string("you probably want to do PERM now\n\r");
		return (TRUE);
	}
	if (!strcasecmp(cmd, "stat")) {
		do {
			print_string("\r");
			stat_line();
			print_string(" ");
			delay_ms(333);
		} while (!ischar());

		getchar();
		eol();

		return (TRUE);
	}

	if (!strcasecmp(cmd, "channel")) {
		int n;

		/* channel <ret> */

		if (*arg == 0) {
			n = 0;
			for (i = 0; i < __SIZEOF(cfg.channel); ++i) {

				if (cfg.channel[i].freq) {

					if (!more(&lines))
						break;

					channel_printout(i, &cfg.channel[i]);
					++n;
				}
			}
			if (i == __SIZEOF(cfg.channel) && n == 0)
				print_string("channel table is empty\n\r");
			return (TRUE);
		}
		/* channel n <ret> */

		i = atoi(blanksplit(&arg));

		if (i >= __SIZEOF(cfg.channel)) {
			print_string("not so many channels\n\r");
			return (TRUE);
		}
		if (*arg == 0) {
			channel_printout(i, &cfg.channel[i]);
			return (TRUE);
		}
		/* channel n r [t] <ret> */

		cfg.channel[i].freq    = atofreq(blanksplit(&arg));
		cfg.channel[i].tx_freq = atofreq(blanksplit(&arg));

		channel_printout(i, &cfg.channel[i]);

		return (TRUE);
	}
	if (!strcasecmp(cmd, "remote")) {
		if (*arg == 0) {
			print_string("need the callsign of the remote set\n\r");
			return (TRUE);
		}
		remote_parameditor(arg);
		return (TRUE);
	}
	if (!strcasecmp(cmd, "telemetry")) {    /* arg = "" or not */

		if (*arg) {
			hold_txproc();

			strncpyx(user_telemetry, arg, sizeof(user_telemetry));
			user_telemetry_active = TRUE;

			okay_txproc();
		} else
			user_telemetry_active = FALSE;

		return (TRUE);
	}
	if (!strcasecmp(cmd, "sendnow")) {
		sendnow = TRUE;
		return (TRUE);
	}
	if (!strcasecmp(cmd, "help")) {

		if (!strcasecmp(arg, "symbols")) {
			show_aprs_symbols();

		} else if (*arg) {
			p = parm_lookup(arg);
			if (p) {
				cfg_printout(p);

				switch (p->type) {
				case CFG_TAB:
					show_possible_options(p->optlist);
					break;
				case CFG_BOOL:
					print_string("possible options: off/on\n\r");
					break;
				case CFG_BYTE:
					print_string("range: 0 ... 255\n\r");
					break;
				case CFG_WORD:
					print_string("range: 0 ... 65535\n\r");
					break;
				case CFG_STR:
					print_string("max ");
					print_short(p->size - 1);
					print_string(" characters\n\r");
					break;
				}
				if (p->info) {
					print_string(p->info);
					eol();
				}
			} else {
				print_string("no help for \"");
				print_string(arg);
				print_string("\", sorry\n\r");
			}
		} else {
			int col = 0;      /* expand tabs from helptext[] */
			byte c;

			for (i = 0; c = helptext(i); ++i) {

				if (col == 0 && !more(&lines))
					break;

				if (c == '\n') {
					eol();
					col = 0;
					continue;
				}
				if (c == '\t') {
					do {
						putchar(' ');
					} while (++col < 18);
					continue;
				}
				putchar(c);
				++col;
			}
		}
		return (TRUE);
	}

	/* confused GPS ? pat it */

	if (!strcmp(cmd, "ASTRAL")) {

		print_string("\r\nASTRAL\r\n");
		return (TRUE);
	}

	/* not a command, maybe a parameter ? "set parm ..." or just "parm ..." */

	if (!strcasecmp(cmd, "set")) {

		cmd = blanksplit(&arg);
		p = parm_lookup(cmd);
		if (p) {
			cfg_assign(p, *arg ? arg : NULL);
			cfg_printout(p);
		} else  {
			print_string("no such parameter \"");
			print_string(cmd);
			print_string("\", sorry\n\r");
		}
		return (TRUE);
	}

	p = parm_lookup(cmd);
	if (p) {
		if (*arg)
			cfg_assign(p, arg);

		cfg_printout(p);
		return (TRUE);
	}

	return (FALSE);
}

static void
cfg_sanitize()
{
	byte *cfg_caddr, *end_old, *end_new;
	int i;

	/* if there is nonsense in cfg, set defaults to all */

	if (cfg.cfg_magic_1 != CFG_MAGIC_1
	 || cfg.config_size < 256
	 || cfg.config_size > 8 * 1024)
		cfg.config_size = 0;

	cfg_caddr = (void *) &cfg;

	end_old = cfg_caddr + cfg.config_size;
	end_new = cfg_caddr + sizeof(cfg);

	/*
	 * brute force first ...
	 */
	if (end_new > end_old)
		bzero(end_old, end_new - end_old);

	/*
	 *  ... then defaults if any
	 */
	for (i = 0; i < __SIZEOF(cfg_general); ++i) {
		/*
		 *  If the parameter lives in the end of the cfg structure,
		 *  after any previous config_size,
		 *  it is uninitialized.
		 */
		struct cfgparm *p = &cfg_general[i];

		if ((byte *) p->address >= end_old)
			cfg_assign(p, NULL);
	}
}

static byte
diddle_getchar()
{
	static const word worm[] = { LED_2, LED_1, LED_2, LED_3, };
	static word which, timo = 2000;
	stamp mark;
	word c;

	future_msec(&mark, timo);

	while ((c = SCI_getc()) == EOF) {

		if (time_past(mark)) {
			if (timo > 100)
				timo -= timo / 32;
			future_msec(&mark, timo);

			green_leds(worm[which++ % 4]);
		}
	}
	timo = 2000;

	return (c);
}

byte *
prompt(byte *PS1)
{
	byte c;
	int i;
	static char line[70];

redo:
	for (;;) {
		if (PS1)
			print_string(PS1);
		i = 0;
		for (;;) {
			c = diddle_getchar();
			switch (c) {

			/*
			 *  NMEA ? Tracker.
			 *  KISS ? ...well, KISS.
			 */
			case '$':
				if (i != 0)
					goto defalut;

				cfg.personality = PERSONALITY_TRACKER; /* no perm though */
				return (NULL);

			case FEND:
				cfg.personality = PERSONALITY_KISS; /* no perm though */
				return (NULL);

			case 0x03:
				print_string("^C\n\r");
				goto redo;

			case 0x00: /* does BRK come thru ? */
			case 0x04:
				print_string("eof\n\r");
				return (NULL);

			case '\r':
			case '\n':
				eol();
				line[i] = 0;
				return (line);

			case '\b':
			case 0x7F:
				if (i > 0) {
					line[--i] = 0;
					print_string("\b \b");
				}
				break;

			default:
		defalut:
				if (c >= ' ' && c < 0x7F && i < sizeof(line) - 1) {
					line[i++] = c;
					line[i] = 0;
					putchar(c);
				}
			}
		}
	}
}

/* meditate the parameters */

static void
parameditor()
{
	byte *cmd;

	eol();

	if (cfg.config_size != sizeof(cfg)
	 || cfg.cfg_magic_1 != CFG_MAGIC_1) {

		print_string(
			"\n\r"
			"\n\r"
			"Some parameters have defaulted. Try help.\n\r"
			"Verify mysymbol and target_symbol.\n\r"
			"\n\r");
	}

	while (cmd = prompt("hamdr $ ")) {
		if (*cmd == 0)
			continue;

		if (!process_manual_command(cmd)) {
			print_string("\"");
			print_string(cmd);
			print_string("\" ?\n\r");
		}
	}
}

static void
default_coordinates()
{
	if (cfg.gps_lat[0] && cfg.gps_lon[0]) {

		nmea_coord(&gps.lat, cfg.gps_lat, cfg.gps_lat[strlen(cfg.gps_lat) - 1]);
		nmea_coord(&gps.lon, cfg.gps_lon, cfg.gps_lon[strlen(cfg.gps_lon) - 1]);

		gps_validity = PINNED; /* until first GPRMC says otherwise */

		any_idea_of_location = TRUE;
	}
}

static bool
select_profile(char *tag)
{
	int i;

	for (i = 0; i < __SIZEOF(profile_tags); ++i)
		if (!strcasecmp(tag, profile_tags[i])) {
			if (profile_offset != PROFILE_OFFSET(i)) {

				while (tx_state != OFF)
					;

				profile_offset = PROFILE_OFFSET(i);
				eeprom_read(&cfg, sizeof(cfg));
				cfg_sanitize();

				longjmp(restart, 3);
			}
			return (TRUE); /* already current profile */
		}
	return (FALSE); /* not a profile selector */
}

static void
prepare_profiles()
{
	int i;

	i = __SIZEOF(profile_tags); /* for fun, read from 3 to 0 */

	while (--i >= 0) {
		profile_offset = PROFILE_OFFSET(i);

		eeprom_read(&cfg, sizeof(cfg));
		cfg_sanitize();

		strncpyx(profile_tags[i], cfg.profile_tag,
			sizeof(profile_tags[i]));
	}
	/* and leaving with profile 0 loaded */
}

/* --------------------------------------------------- */
/* interrupt context, IPL 6. be quick */

void
keypad(byte key)
{
	if (key != '*' || cfg.manipulator != 1) {
		if (keycnt < sizeof(keybuf))
			keybuf[keycnt++] = key;
		if (keycnt == sizeof(keybuf))
			keycnt = 0;
	}
}

/* --------------------------------------------------- */

static void
select_app_port(word port)
{
	SCI_drain();                       /* make sure any messages get thru */

	if (port < 2) {                    /* that's serial 1 or 2 */
		SCI_ssel(port);
		return;
	}
	/*
	 * app_port auto
	 *
	 * few rounds, 1 second for each port.
	 * speed might be wrong at this point XXX use RAF and NF flags from SCI
	 */

	for (port = 0; port < 30; ++port) {
		stamp mark;

		SCI_ssel(port & 1);                /* serial 1 or 2 */

		while (SCI_getc() != EOF)          /* possible switching glitch */
			;

		now(&mark);
		while (time_since(&mark) < 1000) {
			if (SCI_getc() != EOF)
				return;                    /* got it */
		}
	}
}

/*
 *  =============================================================
 */
int
main(void)
{
	default_exception_vectors();

	green_leds(1);

	init_windvane();                /* PORTE generic input */
	init_trxq();
	init_wpt_stash();

	init_SCI(19200, 0);
	init_PIT();

	tx_power(POWER_LOWEST);
	radio_control(RX_ON);            /* TX off, no mute */

	afsk_rx_on();

	spl0();                          /* allow interrupts */

	green_leds(2);

	prepare_profiles();

	spl7();

	switch (cfg.rn3_function) {
	case 1:                         /* weather */
		init_pulsacc();
		break;
	case 2:                         /* keypad */
		init_keypad();
		break;
	}

	spl0();

	/* XXX */

	if (!is_valid_symbol(cfg.mysymbol >> 8, cfg.mysymbol))
		cfg.mysymbol = ('/' << 8) | '>';

	green_leds(4);

	vco_rx(cfg.freq);                /* always sets R and NA */

	if (setjmp(restart))
		;

	green_leds(5);

	/* restarting position. txproc can be active or not */

	ow_reinit();         /* SPI configuration must be done, ow or not */

	update_rx_freq();

	kiss_flush_until_fend();         /* release any current kiss_buf */

	green_leds(6);                   /* async initialization */

	if (cfg.ttybaud != 0)
		SCI_config(cfg.ttybaud, 0);    /* if 0, use 19200 as loader did */

	CTS1_pin(0);
	DCD1_pin(0);
	DSR1_pin(0);

	CTS2_pin(1); /* not sending */
	DCD2_pin(1); /* no open squelch */

	while (SCI_getc() != EOF)        /* purge any poweron async garbage */
		;

	green_leds(7);

	if (cfg.lcd)
		lcd_init(tab_lcd[cfg.lcd]);

#if 0
	if (cfg.chatboard)
		chatboard_init(cfg.chatboard);
#endif

	/* wired for configuration mode ? meditate ! */

	if (DTR1_pin() == 0) {
stick_at_prompt:
		interactive = TRUE;
		parameditor();
	}

	if (cfg.cfg_magic_1 != CFG_MAGIC_1)
		goto stick_at_prompt;          /* insist on good configuration */

	interactive = FALSE;

	update_rx_freq();

	default_coordinates();

	tx_led(TRUE);                      /* act as power led (for a while) */

	unleash_txproc();

	/* allow any messages to get thru to previous port */

	green_leds(~0);                    /* initialization almost ... */

	select_app_port(cfg.app_port);

	green_leds(0);                     /* initialization complete */

	switch (cfg.personality) {

	default:                           /* insist on good configuration */
		goto stick_at_prompt;

	case PERSONALITY_TRACKER:
		aprs_tracker();
		break;

	case PERSONALITY_KISS:
		kiss_tnc();
		break;

	case PERSONALITY_PEATER:
		digipeater();
		break;

		for (;;) {
	case PERSONALITY_VOICE:
			tx_led(FALSE);             /* not a power led */
			simple_voice_radio();   /* these bounce back and forth */

	case PERSONALITY_SCANNER:
			tx_led(FALSE);             /* not a power led */
			scanner();
		}
		break;
	}
}

