/*
 *  MDR150 loader
 *
 *  Sun Sep 26 18:23:30 EEST 2004
 */

static char *version = "@(#)MDR150 loader 26.10.2004";

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include <setjmp.h>
#include <stdio.h>

#define BAUDRATE B19200

static char *serport  = "/dev/cuaa1";
static char *hexfile = "./hamdr.out";
static int debug;
static int mdr, file;
static int keybrd, screen;
static jmp_buf outtahere;
static int hexbuf_cc, hexbuf_idx;
static u_char hexbuf[64];
static struct termios savetty;

static void
doh()
{
	*(int *) 0 = 0;
}

static void
timeout(int sig)
{
	longjmp(outtahere, 1);
}

static int
open_serial(char *dev)
{
	int fd;
	struct termios t;

	fd = open(dev, O_RDWR);
	if (fd == -1)
		goto error;

	if (tcgetattr(fd, &t))
		goto error;

	cfmakeraw(&t);

	cfsetispeed(&t, BAUDRATE);
	cfsetospeed(&t, BAUDRATE);

	if (tcsetattr(fd, TCSAFLUSH, &t))
		goto error;

	return (fd);

error:
	perror(dev);
	exit(1);
}

static void
mainloop()
{
	int nfd, cc;
	fd_set rset, wset;
	u_char c;

	file = -1;

	FD_ZERO(&rset);
	FD_ZERO(&wset);

	for (;;) {
		nfd = mdr;
		if (nfd < keybrd)
			nfd = keybrd;
		if (nfd < file)
			nfd = file;
		nfd++;

		FD_SET(mdr,    &rset);
		FD_SET(keybrd, &rset);

		if (file >= 0)
			FD_SET(mdr, &wset);

		if (select(nfd, &rset, (file >= 0) ? &wset : NULL, NULL, NULL) < 1)
			doh();

		/*
		 *  User input ?
		 */
		if (FD_ISSET(keybrd, &rset)) {

			if (read(keybrd, &c, 1) != 1)
				doh();

			if (c == '~') {
				write(screen, "\n\r", 2);
				hexbuf_cc = 0;
				file = open(hexfile, 0);
				if (file == -1) {
					perror(hexfile);
					write(screen, "\r", 1);
				}
				write(screen, "loading\n\r", 9);
			} else
				if (write(mdr, &c, 1) != 1)
					doh();
		}
		/*
		 *  MDR output ?
		 */
		if (FD_ISSET(mdr, &rset)) {

			if (read(mdr, &c, 1) != 1)
				doh();

			if (write(screen, &c, 1) != 1)
				doh();
		}
		if (file >= 0 && FD_ISSET(mdr, &wset)) {
			if (hexbuf_cc == 0) {
				hexbuf_cc = read(file, hexbuf, sizeof(hexbuf));
				if (hexbuf_cc <= 0) {
					if (hexbuf_cc == -1)
						perror(hexfile);
					else
						write(screen, "done\n", 5);
					write(screen, "\r", 1);
					close(file);
					file = -1;
					continue;
				}
				hexbuf_idx = 0;
			}
			cc = write(mdr, hexbuf + hexbuf_idx, hexbuf_cc);
			if (cc == -1) {
				perror(hexfile);
				write(screen, "\r", 1);
				close(file);
				file = -1;
				continue;
			}
			hexbuf_cc  -= cc;
			hexbuf_idx += cc;
		}
	}
}

main(int argc, char **argv)
{
	extern int optind, opterr;
	extern char *optarg;

	int c;
	struct termios t;

	while ((c = getopt(argc, argv, "d:f:v")) != -1)
		switch (c) {

		case 'v':
			debug++;
			break;

		case 'd':
			serport = optarg;
			break;

		case 'f':
			hexfile = optarg;
			break;

		default:
		usage:
			fprintf(stderr, "\
Usage: loader [-v] [-d /dev/serport] [-f hexfile]\n\
");
			exit(2);
		}

	if (optind != argc)
		goto usage;

	signal(SIGALRM, timeout);
	alarm(30);

	if (setjmp(outtahere)) {
		fprintf(stderr, "Timed out, sorry.\n");
		exit(1);
	}
	mdr = open_serial(serport);
	alarm(0);

	keybrd = 0;
	screen = 1;
	file   = -1;

	if (tcgetattr(keybrd, &savetty))
		doh();

	t = savetty;
	cfmakeraw(&t);

	if (tcsetattr(keybrd, TCSAFLUSH, &t))
		doh();

	mainloop();

	doh();
}

